'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _ = _interopDefault(require('lodash'));
var BN = _interopDefault(require('bn.js'));
var BigNumber = _interopDefault(require('bignumber.js'));
var WIF = _interopDefault(require('wif'));
var bs58 = _interopDefault(require('bs58'));
var crypto = _interopDefault(require('crypto'));
var elliptic = require('elliptic');
var scrypt = _interopDefault(require('scrypt-js'));
var xor = _interopDefault(require('buffer-xor'));

function _typeof(obj) {
  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function step(key, arg) {
        try {
          var info = gen[key](arg);
          var value = info.value;
        } catch (error) {
          reject(error);
          return;
        }

        if (info.done) {
          resolve(value);
        } else {
          Promise.resolve(value).then(_next, _throw);
        }
      }

      function _next(value) {
        step("next", value);
      }

      function _throw(err) {
        step("throw", err);
      }

      _next();
    });
  };
}

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _get(object, property, receiver) {
  if (object === null) object = Function.prototype;
  var desc = Object.getOwnPropertyDescriptor(object, property);

  if (desc === undefined) {
    var parent = Object.getPrototypeOf(object);

    if (parent === null) {
      return undefined;
    } else {
      return _get(parent, property, receiver);
    }
  } else if ("value" in desc) {
    return desc.value;
  } else {
    var getter = desc.get;

    if (getter === undefined) {
      return undefined;
    }

    return getter.call(receiver);
  }
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _sliceIterator(arr, i) {
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _slicedToArray(arr, i) {
  if (Array.isArray(arr)) {
    return arr;
  } else if (Symbol.iterator in Object(arr)) {
    return _sliceIterator(arr, i);
  } else {
    throw new TypeError("Invalid attempt to destructure non-iterable instance");
  }
}

function _toConsumableArray(arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
}

var BaseState = function BaseState(_ref) {
  var versionIn = _ref.version;
  _classCallCheck(this, BaseState);
  this.version = versionIn == null ? this.constructor.VERSION : versionIn;
};

Object.defineProperty(BaseState, "VERSION", {
  configurable: true,
  enumerable: true,
  writable: true,
  value: 0
});

var equals = function (clazz, equals) {
  var _this = this;

  return function (other) {
    return other != null && (_this === other || other instanceof clazz && equals(other));
  };
};

var lazy = function (getValue) {
  var value;
  return function () {
    if (value == null) {
      value = getValue();
    }

    return value;
  };
};

var lazyAsync = function (getValue) {
  var valuePromise;
  return function (input) {
    if (valuePromise == null) {
      valuePromise = getValue(input);
    }

    return valuePromise;
  };
};

var lazyOrValue = function (getValue) {
  var settings = typeof getValue === 'function' ? {
    type: 'lazy',
    getValue: getValue
  } : {
    type: 'evaluated',
    value: getValue
  };
  return function () {
    if (settings.type === 'lazy') {
      settings = {
        type: 'evaluated',
        value: settings.getValue()
      };
    }

    return settings.value;
  };
};

var USHORT_MAX_NUMBER = 65535;
var USHORT_MAX_NUMBER_PLUS_ONE = 65535 + 1;
var USHORT_MAX = new BN(USHORT_MAX_NUMBER);
var USHORT_MAX_PLUS_ONE = USHORT_MAX.addn(1);
var UINT_MAX_NUMBER = 4294967295;
var UINT_MAX = new BN(UINT_MAX_NUMBER);
var ZERO = new BN(0);
var ONE = new BN(1);
var NEGATIVE_ONE = new BN(-1);
var ONE_HUNDRED_MILLION = new BN(100000000);

var toSignedBuffer = function toSignedBuffer(value) {
  var buff = value.toArrayLike(Buffer, 'le');

  if (value.isNeg()) {
    return buff;
  }

  return value.toArrayLike(Buffer, 'le', buff.length + 1);
};

var fromSignedBuffer = function fromSignedBuffer(value) {
  return value.length === 0 ? ZERO : new BN(value, 'le').fromTwos(value.length * 8);
}; // TODO: This might be incorrect with the shenanigans we're doing above


var not = function not(value) {
  return value.notn(value.bitLength());
};

var getBoolean = function getBoolean(value) {
  return value.some(function (byte) {
    return byte !== 0;
  });
};

var booleanToBuffer = function booleanToBuffer(value) {
  return Buffer.from([value ? 1 : 0]);
};

var toASCII = function toASCII(bytes) {
  var result = '';

  for (var i = 0; i < bytes.length; i += 1) {
    result += String.fromCharCode(bytes.readUInt8(i));
  }

  return result;
};

var toUTF8 = function toUTF8(bytes) {
  return bytes.toString('utf8');
};

function keys(obj) {
  return Object.keys(obj);
}

function values(obj) {
  return Object.values(obj);
}

function entries(obj) {
  return Object.entries(obj);
}

var reverse = function reverse(src) {
  var out = Buffer.allocUnsafe(src.length);

  for (var i = 0, j = src.length - 1; i <= j; i += 1, j -= 1) {
    out[i] = src[j];
    out[j] = src[i];
  }

  return out;
};

var calculateClaimAmount = function () {
  var _ref2 = _asyncToGenerator(
  /*#__PURE__*/
  regeneratorRuntime.mark(function _callee2(_ref) {
    var coins, decrementInterval, generationAmount, getSystemFee, grouped, claimed;
    return regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            coins = _ref.coins, decrementInterval = _ref.decrementInterval, generationAmount = _ref.generationAmount, getSystemFee = _ref.getSystemFee;
            grouped = values(_.groupBy(coins, function (coin) {
              return "".concat(coin.startHeight, ":").concat(coin.endHeight);
            }));
            _context2.next = 4;
            return Promise.all(grouped.map(function () {
              var _ref3 = _asyncToGenerator(
              /*#__PURE__*/
              regeneratorRuntime.mark(function _callee(coinsGroup) {
                var _coinsGroup$, startHeight, endHeight, amount, ustart, istart, uend, iend, _ref4, _ref5, sysFeeEnd, sysFeeStart, totalValue;

                return regeneratorRuntime.wrap(function _callee$(_context) {
                  while (1) {
                    switch (_context.prev = _context.next) {
                      case 0:
                        _coinsGroup$ = coinsGroup[0], startHeight = _coinsGroup$.startHeight, endHeight = _coinsGroup$.endHeight;
                        amount = ZERO;
                        ustart = Math.floor(startHeight / decrementInterval);

                        if (ustart < generationAmount.length) {
                          istart = startHeight % decrementInterval;
                          uend = Math.floor(endHeight / decrementInterval);
                          iend = endHeight % decrementInterval;

                          if (uend >= generationAmount.length) {
                            uend = generationAmount.length;
                            iend = 0;
                          }

                          if (iend === 0) {
                            uend -= 1;
                            iend = decrementInterval;
                          }

                          while (ustart < uend) {
                            amount = amount.addn((decrementInterval - istart) * generationAmount[ustart]);
                            ustart += 1;
                            istart = 0;
                          }

                          amount = amount.addn((iend - istart) * generationAmount[ustart]);
                        }

                        _context.next = 6;
                        return Promise.all([getSystemFee(endHeight - 1), startHeight === 0 ? Promise.resolve(ZERO) : getSystemFee(startHeight - 1)]);

                      case 6:
                        _ref4 = _context.sent;
                        _ref5 = _slicedToArray(_ref4, 2);
                        sysFeeEnd = _ref5[0];
                        sysFeeStart = _ref5[1];
                        amount = amount.add(sysFeeEnd.sub(sysFeeStart).div(ONE_HUNDRED_MILLION));
                        totalValue = coinsGroup.reduce(function (acc, _ref6) {
                          var value = _ref6.value;
                          return acc.add(value);
                        }, ZERO);
                        return _context.abrupt("return", [totalValue, amount]);

                      case 13:
                      case "end":
                        return _context.stop();
                    }
                  }
                }, _callee, this);
              }));

              return function (_x2) {
                return _ref3.apply(this, arguments);
              };
            }()));

          case 4:
            claimed = _context2.sent;
            return _context2.abrupt("return", claimed.reduce(function (acc, _ref7) {
              var _ref8 = _slicedToArray(_ref7, 2),
                  value = _ref8[0],
                  amount = _ref8[1];

              return acc.add(value.div(ONE_HUNDRED_MILLION).mul(amount));
            }, ZERO));

          case 6:
          case "end":
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function calculateClaimAmount(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var randomUInt = function randomUInt() {
  return Math.floor(Math.random() * UINT_MAX_NUMBER);
};

var utils$1 = {
  FD: new BN(0xFD),
  FFFF: new BN(0xFFFF),
  FFFFFFFF: new BN(0xFFFFFFFF),
  ZERO: ZERO,
  ONE: ONE,
  NEGATIVE_ONE: NEGATIVE_ONE,
  INT_MAX_VALUE: new BN(2147483647),
  SATOSHI: ONE,
  NEGATIVE_SATOSHI: NEGATIVE_ONE,
  USHORT_MAX_NUMBER: USHORT_MAX_NUMBER,
  USHORT_MAX_NUMBER_PLUS_ONE: USHORT_MAX_NUMBER_PLUS_ONE,
  USHORT_MAX: USHORT_MAX,
  USHORT_MAX_PLUS_ONE: USHORT_MAX_PLUS_ONE,
  UINT_MAX_NUMBER: UINT_MAX_NUMBER,
  UINT_MAX: UINT_MAX,
  ONE_HUNDRED_MILLION: ONE_HUNDRED_MILLION,
  EIGHT: new BN(8),
  TEN: new BN(10),
  SIXTEEN: new BN(16),
  toSignedBuffer: toSignedBuffer,
  fromSignedBuffer: fromSignedBuffer,
  not: not,
  getBoolean: getBoolean,
  booleanToBuffer: booleanToBuffer,
  toASCII: toASCII,
  toUTF8: toUTF8,
  keys: keys,
  values: values,
  entries: entries,
  reverse: reverse,
  calculateClaimAmount: calculateClaimAmount,
  randomUInt: randomUInt
};

// TODO: Might be broken since we aren't using longs..
var weightedAverage = (function (input) {
  var sumWeight = 0;
  var sumValue = 0;
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = input[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var _value2 = _step.value;
      sumWeight += _value2.weight;
      sumValue += _value2.value * _value2.weight;
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return != null) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  if (sumValue === 0 || sumWeight === 0) {
    return 0;
  }

  return sumValue / sumWeight;
});

function weightedFilter(input, start, end, getValue) {
  var amount = input.reduce(function (acc, value) {
    return acc + getValue(value);
  }, 0);
  var sum = 0;
  var current = 0;
  var result = [];
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = input[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var _value2 = _step.value;

      if (current >= end) {
        break;
      }

      var weight = getValue(_value2);
      sum += weight;
      var old = current;
      current = sum / amount;

      if (current <= start) {
        // eslint-disable-next-line
        continue;
      }

      if (old < start) {
        if (current > end) {
          weight = (end - start) * amount;
        } else {
          weight = (current - start) * amount;
        }
      } else if (current > end) {
        weight = (end - old) * amount;
      }

      result.push([_value2, weight >= 0 ? Math.floor(weight) : Math.ceil(weight)]);
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return != null) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  return result;
}

var _fixBabelExtend = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var InvalidFormatError = _fixBabelExtend(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidFormatError, _Error);

  function InvalidFormatError() {
    _classCallCheck(this, InvalidFormatError);
    return _possibleConstructorReturn(this, (InvalidFormatError.__proto__ || Object.getPrototypeOf(InvalidFormatError)).call(this, 'Invalid format.'));
  }

  return InvalidFormatError;
}(Error));
var VerifyError = _fixBabelExtend(
/*#__PURE__*/
function (_Error2) {
  _inherits(VerifyError, _Error2);

  function VerifyError(reason) {
    _classCallCheck(this, VerifyError);
    return _possibleConstructorReturn(this, (VerifyError.__proto__ || Object.getPrototypeOf(VerifyError)).call(this, "Verification failed: ".concat(reason)));
  }

  return VerifyError;
}(Error));

var add0x = function add0x(value) {
  return "0x".concat(value);
};

var strip0x = function strip0x(value) {
  return value.startsWith('0x') ? value.substring(2) : value;
}; // eslint-disable-next-line


var UINT160_BUFFER_BYTES = 20;

var bufferToUInt160 = function bufferToUInt160(value) {
  if (value.length !== UINT160_BUFFER_BYTES) {
    throw new InvalidFormatError();
  }

  return value;
};

var uInt160ToHex = function uInt160ToHex(value) {
  return add0x(utils.reverse(value).toString('hex'));
};

var hexToUInt160 = function hexToUInt160(value) {
  return typeof value === 'string' ? bufferToUInt160(utils.reverse(Buffer.from(strip0x(value), 'hex'))) : value;
};

var uInt160ToBuffer = function uInt160ToBuffer(value) {
  return typeof value === 'string' ? hexToUInt160(value) : value;
};

var uInt160ToString = function uInt160ToString(value) {
  return typeof value === 'string' ? value : uInt160ToHex(value);
};

var stringToUInt160 = function stringToUInt160(value) {
  return hexToUInt160(value);
};

var uInt160Equal = function uInt160Equal(a, b) {
  return a.equals(b);
};

var isUInt160 = function isUInt160(value) {
  return value instanceof Buffer && value.length === UINT160_BUFFER_BYTES;
};

var asUInt160 = function asUInt160(value) {
  if (value instanceof Buffer) {
    return bufferToUInt160(value);
  }

  throw new InvalidFormatError();
};

var ZERO_UINT160 = Buffer.alloc(20, 0); // eslint-disable-next-line

var UINT256_BUFFER_BYTES = 32;
var ZERO_UINT256 = Buffer.alloc(UINT256_BUFFER_BYTES, 0);

var bufferToUInt256 = function bufferToUInt256(value) {
  if (value.length !== UINT256_BUFFER_BYTES) {
    throw new InvalidFormatError();
  }

  return value;
};

var uInt256ToHex = function uInt256ToHex(value) {
  return add0x(utils.reverse(value).toString('hex'));
};

var hexToUInt256 = function hexToUInt256(value) {
  return typeof value === 'string' ? bufferToUInt256(utils.reverse(Buffer.from(strip0x(value), 'hex'))) : value;
};

var uInt256ToBuffer = function uInt256ToBuffer(value) {
  return typeof value === 'string' ? hexToUInt256(value) : value;
};

var uInt256ToString = function uInt256ToString(value) {
  return typeof value === 'string' ? value : uInt256ToHex(value);
};

var stringToUInt256 = function stringToUInt256(value) {
  return hexToUInt256(value);
};

var uInt256Equal = function uInt256Equal(a, b) {
  return a.equals(b);
};

var toUInt32LE = function toUInt32LE(bytes) {
  return new BN(uInt256ToBuffer(bytes).slice(0, 4), 'le').toNumber();
}; // eslint-disable-next-line


// Encoded compressed ECPoint
var ECPOINT_BUFFER_BYTES = 33;
var ECPOINT_INFINITY_BYTE = 0x00;
var ECPOINT_INFINITY = Buffer.from([ECPOINT_INFINITY_BYTE]);

var bufferToECPoint = function bufferToECPoint(value) {
  if (value.length !== ECPOINT_BUFFER_BYTES) {
    if (value.length === 1 && value.equals(ECPOINT_INFINITY)) {
      return value;
    }

    throw new InvalidFormatError();
  }

  return value;
};

var ecPointToHex = function ecPointToHex(value) {
  return typeof value === 'string' ? value : value.toString('hex');
};

var hexToECPoint = function hexToECPoint(value) {
  return bufferToECPoint(typeof value === 'string' ? Buffer.from(value, 'hex') : value);
};

var ecPointToBuffer = function ecPointToBuffer(value) {
  return typeof value === 'string' ? hexToECPoint(value) : value;
};

var ecPointToString = function ecPointToString(value) {
  return typeof value === 'string' ? value : ecPointToHex(value);
};

var stringToECPoint = function stringToECPoint(value) {
  return hexToECPoint(value);
};

var ecPointEqual = function ecPointEqual(a, b) {
  return a.equals(b);
};

var ecPointCompare = function ecPointCompare(a, b) {
  var aHex = ecPointToHex(a);
  var bHex = ecPointToHex(b);

  if (aHex < bHex) {
    return -1;
  } else if (aHex > bHex) {
    return 1;
  }

  return 0;
};

var ecPointIsInfinity = function ecPointIsInfinity(value) {
  return value.equals(ECPOINT_INFINITY);
}; // eslint-disable-next-line


var PRIVATE_KEY_BUFFER_BYTES = 32;

var bufferToPrivateKey = function bufferToPrivateKey(value) {
  if (value.length !== PRIVATE_KEY_BUFFER_BYTES) {
    throw new InvalidFormatError();
  }

  return value;
};

var privateKeyToHex = function privateKeyToHex(value) {
  return typeof value === 'string' ? value : value.toString('hex');
};

var hexToPrivateKey = function hexToPrivateKey(value) {
  return bufferToPrivateKey(typeof value === 'string' ? Buffer.from(value, 'hex') : value);
};

var privateKeyToBuffer = function privateKeyToBuffer(value) {
  return typeof value === 'string' ? hexToPrivateKey(value) : value;
};

var privateKeyToString = function privateKeyToString(value) {
  return typeof value === 'string' ? value : ecPointToHex(value);
};

var stringToPrivateKey = function stringToPrivateKey(value) {
  return hexToPrivateKey(value);
};

var D = new BN(100000000);
var DBigNumber = new BigNumber(D.toString(10));

var fixed8FromDecimal = function fixed8FromDecimal(value) {
  if (typeof value === 'number') {
    return new BN(value).mul(D);
  }

  if (value instanceof BN) {
    return value.mul(D);
  }

  var valueBigNumber = value;

  if (typeof value === 'string') {
    valueBigNumber = new BigNumber(value);
  } // $FlowFixMe


  return new BN(valueBigNumber.times(DBigNumber).toString(), 10);
};

var fixed8ToDecimal = function fixed8ToDecimal(bn) {
  return new BigNumber(bn.toString(10)).div(DBigNumber);
};

var NEGATIVE_SATOSHI_FIXED8 = new BN(-1);
var ONE_HUNDRED_FIXED8 = fixed8FromDecimal(100);
var common = {
  D: D,
  NEO_ADDRESS_VERSION: 23,
  NEO_PRIVATE_KEY_VERSION: 0x80,
  ECPOINT_BUFFER_BYTES: ECPOINT_BUFFER_BYTES,
  ECPOINT_INFINITY: ECPOINT_INFINITY,
  ECPOINT_INFINITY_BYTE: ECPOINT_INFINITY_BYTE,
  PRIVATE_KEY_BUFFER_BYTES: PRIVATE_KEY_BUFFER_BYTES,
  UINT160_BUFFER_BYTES: UINT160_BUFFER_BYTES,
  UINT256_BUFFER_BYTES: UINT256_BUFFER_BYTES,
  ZERO_UINT160: ZERO_UINT160,
  ZERO_UINT256: ZERO_UINT256,
  NEGATIVE_SATOSHI_FIXED8: NEGATIVE_SATOSHI_FIXED8,
  ONE_HUNDRED_FIXED8: ONE_HUNDRED_FIXED8,
  uInt160ToBuffer: uInt160ToBuffer,
  bufferToUInt160: bufferToUInt160,
  uInt160ToHex: uInt160ToHex,
  hexToUInt160: hexToUInt160,
  uInt160ToString: uInt160ToString,
  stringToUInt160: stringToUInt160,
  uInt160Equal: uInt160Equal,
  isUInt160: isUInt160,
  asUInt160: asUInt160,
  uInt256ToBuffer: uInt256ToBuffer,
  bufferToUInt256: bufferToUInt256,
  uInt256ToHex: uInt256ToHex,
  hexToUInt256: hexToUInt256,
  uInt256ToString: uInt256ToString,
  stringToUInt256: stringToUInt256,
  uInt256Equal: uInt256Equal,
  toUInt32LE: toUInt32LE,
  ecPointToBuffer: ecPointToBuffer,
  bufferToECPoint: bufferToECPoint,
  ecPointToHex: ecPointToHex,
  hexToECPoint: hexToECPoint,
  ecPointToString: ecPointToString,
  ecPointCompare: ecPointCompare,
  stringToECPoint: stringToECPoint,
  ecPointEqual: ecPointEqual,
  ecPointIsInfinity: ecPointIsInfinity,
  privateKeyToHex: privateKeyToHex,
  hexToPrivateKey: hexToPrivateKey,
  privateKeyToBuffer: privateKeyToBuffer,
  bufferToPrivateKey: bufferToPrivateKey,
  privateKeyToString: privateKeyToString,
  stringToPrivateKey: stringToPrivateKey,
  fixed8FromDecimal: fixed8FromDecimal,
  fixed8ToDecimal: fixed8ToDecimal
};

var BinaryReader =
/*#__PURE__*/
function () {
  function BinaryReader(buffer, index) {
    _classCallCheck(this, BinaryReader);
    this.buffer = buffer;
    this.index = index || 0;
  }

  _createClass(BinaryReader, [{
    key: "hasMore",
    value: function hasMore() {
      return this.index < this.buffer.byteLength;
    }
  }, {
    key: "clone",
    value: function clone() {
      return new BinaryReader(this.buffer, this.index);
    }
  }, {
    key: "readBytes",
    value: function readBytes(numBytes) {
      var result = this.buffer.slice(this.index, this.index + numBytes);
      this.index += numBytes;
      return result;
    }
  }, {
    key: "readInt8",
    value: function readInt8() {
      var result = this.buffer.readInt8(this.index);
      this.index += 1;
      return result;
    }
  }, {
    key: "readUInt8",
    value: function readUInt8() {
      var result = this.buffer.readUInt8(this.index);
      this.index += 1;
      return result;
    }
  }, {
    key: "readBoolean",
    value: function readBoolean() {
      return this.readBytes(1)[0] !== 0;
    }
  }, {
    key: "readInt16LE",
    value: function readInt16LE() {
      var result = this.buffer.readInt16LE(this.index);
      this.index += 2;
      return result;
    }
  }, {
    key: "readUInt16LE",
    value: function readUInt16LE() {
      var result = this.buffer.readUInt16LE(this.index);
      this.index += 2;
      return result;
    }
  }, {
    key: "readInt32LE",
    value: function readInt32LE() {
      var result = this.buffer.readInt32LE(this.index);
      this.index += 4;
      return result;
    }
  }, {
    key: "readUInt32LE",
    value: function readUInt32LE() {
      var result = this.buffer.readUInt32LE(this.index);
      this.index += 4;
      return result;
    }
  }, {
    key: "readUInt64LE",
    value: function readUInt64LE() {
      return new BN(this.readBytes(8), 'le');
    }
  }, {
    key: "readInt64LE",
    value: function readInt64LE() {
      var buffer = this.readBytes(8);
      return new BN(buffer, 'le').fromTwos(buffer.length * 8);
    } // NEO specific

  }, {
    key: "readUInt160",
    value: function readUInt160() {
      return common.bufferToUInt160(this.readBytes(common.UINT160_BUFFER_BYTES));
    }
  }, {
    key: "readUInt256",
    value: function readUInt256() {
      return common.bufferToUInt256(this.readBytes(common.UINT256_BUFFER_BYTES));
    }
  }, {
    key: "readFixed8",
    value: function readFixed8() {
      return this.readInt64LE();
    }
  }, {
    key: "readFixedString",
    value: function readFixedString(length) {
      var values = _.takeWhile([].concat(_toConsumableArray(this.readBytes(length))), function (value) {
        return value !== 0;
      });

      return Buffer.from(values).toString('utf8');
    }
  }, {
    key: "readArray",
    value: function readArray(read) {
      var max = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0x10000000;
      var count = this.readVarUIntLE(new BN(max)).toNumber();
      var result = [];

      for (var i = 0; i < count; i += 1) {
        result.push(read());
      }

      return result;
    }
  }, {
    key: "readObject",
    value: function readObject(read) {
      var max = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0x10000000;
      var count = this.readVarUIntLE(new BN(max)).toNumber();
      var result = {};

      for (var i = 0; i < count; i += 1) {
        var _read = read(),
            _key = _read.key,
            _value = _read.value;

        result[_key] = _value;
      }

      return result;
    }
  }, {
    key: "readVarBytesLE",
    value: function readVarBytesLE() {
      var max = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : Number.MAX_SAFE_INTEGER;
      return this.readBytes(this.readVarUIntLE(new BN(max)).toNumber());
    }
  }, {
    key: "readVarUIntLE",
    value: function readVarUIntLE() {
      var max = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : new BN('18446744073709551615', 10);
      var fb = this.readUInt8();
      var value = new BN(fb);

      if (fb === 0xFD) {
        value = new BN(this.readUInt16LE());
      } else if (fb === 0xFE) {
        value = new BN(this.readUInt32LE());
      } else if (fb === 0xFF) {
        value = this.readUInt64LE();
      }

      if (value.gt(max)) {
        throw new InvalidFormatError();
      }

      return value;
    }
  }, {
    key: "readVarString",
    value: function readVarString() {
      var max = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 0x7fffffc7;
      return this.readVarBytesLE(max).toString('utf8');
    }
  }, {
    key: "readECPoint",
    value: function readECPoint() {
      var firstByte = this.readBytes(1);

      if (firstByte[0] === common.ECPOINT_INFINITY_BYTE) {
        return common.ECPOINT_INFINITY;
      }

      return common.bufferToECPoint(Buffer.concat([firstByte, this.readBytes(common.ECPOINT_BUFFER_BYTES - 1)]));
    }
  }, {
    key: "remaining",
    get: function get() {
      return this.buffer.length - this.index;
    }
  }, {
    key: "remainingBuffer",
    get: function get() {
      return this.buffer.slice(this.index);
    }
  }]);
  return BinaryReader;
}();

var BinaryWriter =
/*#__PURE__*/
function () {
  function BinaryWriter() {
    _classCallCheck(this, BinaryWriter);
    this.buffer = [];
  }

  _createClass(BinaryWriter, [{
    key: "toBuffer",
    value: function toBuffer() {
      return Buffer.concat(this.buffer);
    }
  }, {
    key: "writeBytes",
    value: function writeBytes(value) {
      this.buffer.push(value);
      return this;
    }
  }, {
    key: "writeUInt8",
    value: function writeUInt8(value) {
      var buffer = Buffer.allocUnsafe(1);
      buffer.writeUInt8(value, 0);
      return this.writeBytes(buffer);
    }
  }, {
    key: "writeInt16LE",
    value: function writeInt16LE(value) {
      var buffer = Buffer.allocUnsafe(2);
      buffer.writeInt16LE(value, 0);
      return this.writeBytes(buffer);
    }
  }, {
    key: "writeUInt16LE",
    value: function writeUInt16LE(value) {
      var buffer = Buffer.allocUnsafe(2);
      buffer.writeUInt16LE(value, 0);
      return this.writeBytes(buffer);
    }
  }, {
    key: "writeInt32LE",
    value: function writeInt32LE(value) {
      var buffer = Buffer.allocUnsafe(4);
      buffer.writeInt32LE(value, 0);
      return this.writeBytes(buffer);
    }
  }, {
    key: "writeUInt32LE",
    value: function writeUInt32LE(value) {
      var buffer = Buffer.allocUnsafe(4);
      buffer.writeUInt32LE(value, 0);
      return this.writeBytes(buffer);
    }
  }, {
    key: "writeInt64LE",
    value: function writeInt64LE(value) {
      return this.writeBytes(value.toTwos(8 * 8).toArrayLike(Buffer, 'le', 8));
    }
  }, {
    key: "writeUInt64LE",
    value: function writeUInt64LE(value) {
      return this.writeBytes(value.toArrayLike(Buffer, 'le', 8));
    }
  }, {
    key: "writeBoolean",
    value: function writeBoolean(value) {
      this.writeBytes(Buffer.from([value ? 1 : 0]));
      return this;
    } // Special methods that don't fit the LE mold and/or are specific to NEO.

  }, {
    key: "writeUInt160",
    value: function writeUInt160(hash) {
      return this.writeBytes(common.uInt160ToBuffer(hash));
    }
  }, {
    key: "writeUInt256",
    value: function writeUInt256(hash) {
      return this.writeBytes(common.uInt256ToBuffer(hash));
    }
  }, {
    key: "writeFixed8",
    value: function writeFixed8(value) {
      return this.writeInt64LE(value);
    }
  }, {
    key: "writeFixedString",
    value: function writeFixedString(value, length) {
      if (value.length > length) {
        throw new InvalidFormatError();
      }

      var buffer = Buffer.from(value, 'utf8');

      if (buffer.length > length) {
        throw new InvalidFormatError();
      }

      this.writeBytes(buffer);

      if (buffer.length < length) {
        this.writeBytes(Buffer.alloc(length - buffer.length, 0));
      }

      return this;
    }
  }, {
    key: "writeArray",
    value: function writeArray(values, write) {
      this.writeVarUIntLE(values.length);
      values.forEach(function (value) {
        return write(value);
      });
      return this;
    }
  }, {
    key: "writeObject",
    value: function writeObject(value, write) {
      var entries = utils$1.entries(value);
      this.writeVarUIntLE(entries.length);
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = entries[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var _ref3 = _step.value;

          var _ref2 = _slicedToArray(_ref3, 2);

          var _key2 = _ref2[0];
          var _val = _ref2[1];
          write(_key2, _val);
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return != null) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }

      return this;
    }
  }, {
    key: "writeVarBytesLE",
    value: function writeVarBytesLE(value) {
      this.writeVarUIntLE(value.length);
      return this.writeBytes(value);
    }
  }, {
    key: "writeVarUIntLE",
    value: function writeVarUIntLE(valueIn) {
      var value = new BN(valueIn);

      if (value.lt(utils$1.ZERO)) {
        throw new InvalidFormatError();
      }

      if (value.lt(utils$1.FD)) {
        this.writeUInt8(value.toNumber());
      } else if (value.lte(utils$1.FFFF)) {
        this.writeUInt8(0xfd);
        this.writeUInt16LE(value.toNumber());
      } else if (value.lte(utils$1.FFFFFFFF)) {
        this.writeUInt8(0xfe);
        this.writeUInt32LE(value.toNumber());
      } else {
        this.writeUInt8(0xff);
        this.writeUInt64LE(value);
      }

      return this;
    }
  }, {
    key: "writeVarString",
    value: function writeVarString(value, max) {
      var buffer = Buffer.from(value, 'utf8');

      if (max != null) {
        buffer = buffer.slice(0, max);
      }

      return this.writeVarBytesLE(buffer);
    }
  }, {
    key: "writeECPoint",
    value: function writeECPoint(value) {
      if (common.ecPointIsInfinity(value)) {
        return this.writeBytes(Buffer.from([common.ECPOINT_INFINITY_BYTE]));
      }

      return this.writeBytes(common.ecPointToBuffer(value));
    }
  }]);
  return BinaryWriter;
}();

var sizeOfUInt8 = 1;
var sizeOfBoolean = 1;
var sizeOfUInt16LE = 2;
var sizeOfUInt32LE = 4;
var sizeOfUInt64LE = 8;
var sizeOfUInt160 = common.UINT160_BUFFER_BYTES;
var sizeOfUInt256 = common.UINT256_BUFFER_BYTES;

var sizeOfECPoint = function sizeOfECPoint(ecPoint) {
  if (common.ecPointIsInfinity(ecPoint)) {
    return 1;
  }

  return common.ECPOINT_BUFFER_BYTES;
};

var sizeOfFixed8 = 8;

var sizeOfVarUIntLE = function sizeOfVarUIntLE(valueIn) {
  var value = new BN(valueIn);

  if (value.lt(utils$1.ZERO)) {
    throw new InvalidFormatError();
  }

  if (value.lt(utils$1.FD)) {
    return sizeOfUInt8;
  } else if (value.lte(utils$1.FFFF)) {
    return sizeOfUInt8 + sizeOfUInt16LE;
  } else if (value.lte(utils$1.FFFFFFFF)) {
    return sizeOfUInt8 + sizeOfUInt32LE;
  }

  return sizeOfUInt8 + sizeOfUInt64LE;
};

var sizeOfVarBytesLE = function sizeOfVarBytesLE(value) {
  return sizeOfVarUIntLE(value.length) + value.length;
};

var sizeOfVarString = function sizeOfVarString(value) {
  return sizeOfVarBytesLE(Buffer.from(value, 'utf8'));
};

var sizeOfFixedString = function sizeOfFixedString(length) {
  return length;
};

function sizeOfArray(values, sizeOf) {
  return values.reduce(function (acc, value) {
    return acc + sizeOf(value);
  }, sizeOfVarUIntLE(values.length));
}

function sizeOfObject(obj, sizeOf) {
  var entries = utils$1.entries(obj);
  return entries.reduce(function (acc, _ref) {
    var _ref2 = _slicedToArray(_ref, 2),
        key = _ref2[0],
        value = _ref2[1];

    return acc + sizeOf(key, value);
  }, entries.length);
}


var IOHelper = {
  sizeOfUInt8: sizeOfUInt8,
  sizeOfBoolean: sizeOfBoolean,
  sizeOfUInt16LE: sizeOfUInt16LE,
  sizeOfUInt32LE: sizeOfUInt32LE,
  sizeOfUInt64LE: sizeOfUInt64LE,
  sizeOfVarBytesLE: sizeOfVarBytesLE,
  sizeOfVarString: sizeOfVarString,
  sizeOfFixedString: sizeOfFixedString,
  sizeOfArray: sizeOfArray,
  sizeOfObject: sizeOfObject,
  sizeOfUInt160: sizeOfUInt160,
  sizeOfUInt256: sizeOfUInt256,
  sizeOfECPoint: sizeOfECPoint,
  sizeOfFixed8: sizeOfFixed8
};

var JSONHelper = {
  writeUInt64LE: function writeUInt64LE(value) {
    return value.toString(16, 16);
  },
  readUInt64LE: function readUInt64LE(value) {
    return new BN(value, 16);
  },
  writeUInt160: function writeUInt160(value) {
    return common.uInt160ToString(value);
  },
  readUInt160: function readUInt160(hash) {
    return common.stringToUInt160(hash);
  },
  writeUInt256: function writeUInt256(value) {
    return common.uInt256ToString(value);
  },
  readUInt256: function readUInt256(hash) {
    return common.stringToUInt256(hash);
  },
  writeFixed8: function writeFixed8(value) {
    return common.fixed8ToDecimal(value).toString();
  },
  readFixed8: function readFixed8(value) {
    return common.fixed8FromDecimal(new BigNumber(value));
  },
  writeECPoint: function writeECPoint(value) {
    return common.ecPointToString(value);
  },
  readECPoint: function readECPoint(value) {
    return common.stringToECPoint(value);
  },
  writeBuffer: function writeBuffer(value) {
    return value.toString('hex');
  },
  readBuffer: function readBuffer(value) {
    return Buffer.from(value, 'hex');
  }
};

var _fixBabelExtend$2 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var OPCODE_PAIRS = [[0x00, 'PUSH0'], [0x01, 'PUSHBYTES1']].concat(_.range(2, 74).map(function (idx) {
  return [idx, "PUSHBYTES".concat(idx)];
})).concat([[0x4B, 'PUSHBYTES75'], [0x4C, 'PUSHDATA1'], [0x4D, 'PUSHDATA2'], [0x4E, 'PUSHDATA4'], [0x4F, 'PUSHM1'], [0x51, 'PUSH1'], [0x52, 'PUSH2'], [0x53, 'PUSH3'], [0x54, 'PUSH4'], [0x55, 'PUSH5'], [0x56, 'PUSH6'], [0x57, 'PUSH7'], [0x58, 'PUSH8'], [0x59, 'PUSH9'], [0x5A, 'PUSH10'], [0x5B, 'PUSH11'], [0x5C, 'PUSH12'], [0x5D, 'PUSH13'], [0x5E, 'PUSH14'], [0x5F, 'PUSH15'], [0x60, 'PUSH16'], [0x61, 'NOP'], [0x62, 'JMP'], [0x63, 'JMPIF'], [0x64, 'JMPIFNOT'], [0x65, 'CALL'], [0x66, 'RET'], [0x67, 'APPCALL'], [0x68, 'SYSCALL'], [0x69, 'TAILCALL'], [0x6A, 'DUPFROMALTSTACK'], [0x6B, 'TOALTSTACK'], [0x6C, 'FROMALTSTACK'], [0x6D, 'XDROP'], [0x72, 'XSWAP'], [0x73, 'XTUCK'], [0x74, 'DEPTH'], [0x75, 'DROP'], [0x76, 'DUP'], [0x77, 'NIP'], [0x78, 'OVER'], [0x79, 'PICK'], [0x7A, 'ROLL'], [0x7B, 'ROT'], [0x7C, 'SWAP'], [0x7D, 'TUCK'], [0x7E, 'CAT'], [0x7F, 'SUBSTR'], [0x80, 'LEFT'], [0x81, 'RIGHT'], [0x82, 'SIZE'], [0x83, 'INVERT'], [0x84, 'AND'], [0x85, 'OR'], [0x86, 'XOR'], [0x87, 'EQUAL'], [0x88, 'OP_EQUALVERIFY'], [0x89, 'OP_RESERVED1'], [0x8A, 'OP_RESERVED2'], [0x8B, 'INC'], [0x8C, 'DEC'], [0x8D, 'SIGN'], [0x8F, 'NEGATE'], [0x90, 'ABS'], [0x91, 'NOT'], [0x92, 'NZ'], [0x93, 'ADD'], [0x94, 'SUB'], [0x95, 'MUL'], [0x96, 'DIV'], [0x97, 'MOD'], [0x98, 'SHL'], [0x99, 'SHR'], [0x9A, 'BOOLAND'], [0x9B, 'BOOLOR'], [0x9C, 'NUMEQUAL'], [0x9E, 'NUMNOTEQUAL'], [0x9F, 'LT'], [0xA0, 'GT'], [0xA1, 'LTE'], [0xA2, 'GTE'], [0xA3, 'MIN'], [0xA4, 'MAX'], [0xA5, 'WITHIN'], [0xA7, 'SHA1'], [0xA8, 'SHA256'], [0xA9, 'HASH160'], [0xAA, 'HASH256'], [0xAC, 'CHECKSIG'], [0xAE, 'CHECKMULTISIG'], [0xC0, 'ARRAYSIZE'], [0xC1, 'PACK'], [0xC2, 'UNPACK'], [0xC3, 'PICKITEM'], [0xC4, 'SETITEM'], [0xC5, 'NEWARRAY'], [0xC6, 'NEWSTRUCT'], [0xF0, 'THROW'], [0xF1, 'THROWIFNOT']]);
var BYTECODE_TO_OPCODE = _.fromPairs(OPCODE_PAIRS);
var OPCODE_TO_BYTECODE = _.mapValues(_.invert(BYTECODE_TO_OPCODE), function (val) {
  return parseInt(val, 10);
});

var InvalidSysCallNameError = _fixBabelExtend$2(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidSysCallNameError, _Error);

  function InvalidSysCallNameError(value) {
    var _this;

    _classCallCheck(this, InvalidSysCallNameError);
    _this = _possibleConstructorReturn(this, (InvalidSysCallNameError.__proto__ || Object.getPrototypeOf(InvalidSysCallNameError)).call(this, "Expected sys call name, found: ".concat(value)));
    _this.value = value;
    return _this;
  }

  return InvalidSysCallNameError;
}(Error));

var _fixBabelExtend$1 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var UnknownOpError = _fixBabelExtend$1(
/*#__PURE__*/
function (_Error) {
  _inherits(UnknownOpError, _Error);

  function UnknownOpError(byteCode) {
    var _this;

    _classCallCheck(this, UnknownOpError);
    _this = _possibleConstructorReturn(this, (UnknownOpError.__proto__ || Object.getPrototypeOf(UnknownOpError)).call(this, "Unknown op: ".concat(byteCode)));
    _this.byteCode = byteCode;
    return _this;
  }

  return UnknownOpError;
}(Error));
var InvalidParamError = _fixBabelExtend$1(
/*#__PURE__*/
function (_Error2) {
  _inherits(InvalidParamError, _Error2);

  function InvalidParamError() {
    _classCallCheck(this, InvalidParamError);
    return _possibleConstructorReturn(this, (InvalidParamError.__proto__ || Object.getPrototypeOf(InvalidParamError)).call(this, 'Invalid Param'));
  }

  return InvalidParamError;
}(Error));

var BYTECODE_TO_BYTECODE_BUFFER = _.fromPairs(Object.values(OPCODE_TO_BYTECODE).map(function (byteCode) {
  return [byteCode, Buffer.from([byteCode])];
}));

var ScriptBuilder =
/*#__PURE__*/
function () {
  function ScriptBuilder() {
    _classCallCheck(this, ScriptBuilder);
    this.buffers = [];
  }

  _createClass(ScriptBuilder, [{
    key: "emitPush",
    value: function emitPush(value) {
      if (value.length <= OPCODE_TO_BYTECODE.PUSHBYTES75) {
        this.emitOpByte(value.length, value);
      } else if (value.length < 0x100) {
        this.emitOp('PUSHDATA1');
        this.emitUInt8(value.length);
        this.emit(value);
      } else if (value.length < 0x10000) {
        this.emitOp('PUSHDATA2');
        this.emitUInt16LE(value.length);
        this.emit(value); // TODO: Check this condition if (data.Length < 0x100000000L)
      } else {
        this.emitOp('PUSHDATA4');
        this.emitUInt32LE(value.length);
        this.emit(value);
      }

      return this;
    }
  }, {
    key: "emitUInt8",
    value: function emitUInt8(value) {
      var buff = Buffer.allocUnsafe(1);
      buff.writeUInt8(value, 0);
      return this.emit(buff);
    }
  }, {
    key: "emitUInt16LE",
    value: function emitUInt16LE(value) {
      var buff = Buffer.allocUnsafe(2);
      buff.writeUInt16LE(value, 0);
      return this.emit(buff);
    }
  }, {
    key: "emitUInt32LE",
    value: function emitUInt32LE(value) {
      var buff = Buffer.allocUnsafe(4);
      buff.writeUInt32LE(value, 0);
      return this.emit(buff);
    }
  }, {
    key: "emitPushInt",
    value: function emitPushInt(valueIn) {
      var value = new BN(valueIn);

      if (value.eq(utils$1.NEGATIVE_ONE)) {
        return this.emitOp('PUSHM1');
      } else if (value.eq(utils$1.ZERO)) {
        return this.emitOp('PUSH0');
      } else if (value.gt(utils$1.ZERO) && value.lt(utils$1.SIXTEEN)) {
        return this.emitOpByte(OPCODE_TO_BYTECODE.PUSH1 - 1 + value.toNumber());
      }

      return this.emitPush(utils$1.toSignedBuffer(value));
    }
  }, {
    key: "emitPushUInt160",
    value: function emitPushUInt160(value) {
      return this.emitPush(common.uInt160ToBuffer(value));
    }
  }, {
    key: "emitPushECPoint",
    value: function emitPushECPoint(ecPoint) {
      return this.emitPush(common.ecPointToBuffer(ecPoint));
    }
  }, {
    key: "emitPushString",
    value: function emitPushString(value) {
      return this.emitPush(Buffer.from(value, 'utf8'));
    }
  }, {
    key: "emitOp",
    value: function emitOp(op, buffer) {
      this.emitOpByte(OPCODE_TO_BYTECODE[op], buffer);
      return this;
    }
  }, {
    key: "emitPushParam",
    value: function emitPushParam(param) {
      if (common.isUInt160(param)) {
        return this.emitPushUInt160(common.asUInt160(param));
      } else if (typeof param === 'number' || param instanceof BN) {
        return this.emitPushInt(param);
      } else if (typeof param === 'string') {
        return this.emitPushString(param);
      }

      throw new InvalidParamError();
    }
  }, {
    key: "emitAppCall",
    value: function emitAppCall(scriptHash, operation) {
      for (var i = (arguments.length <= 2 ? 0 : arguments.length - 2) - 1; i >= 0; i -= 1) {
        this.emitPushParam(i + 2 < 2 || arguments.length <= i + 2 ? undefined : arguments[i + 2]);
      }

      this.emitPushParam(arguments.length <= 2 ? 0 : arguments.length - 2);
      this.emitOp('PACK');
      this.emitPushParam(operation);
      return this.emitOp('APPCALL', common.uInt160ToBuffer(scriptHash));
    }
  }, {
    key: "emitOpByte",
    value: function emitOpByte(byteCodeIn, buffer) {
      var byteCode = "".concat(byteCodeIn == null ? '' : byteCodeIn);
      var byteCodeBuffer = BYTECODE_TO_BYTECODE_BUFFER[byteCode];

      if (byteCodeBuffer == null) {
        throw new UnknownOpError(byteCode);
      }

      this.emit(byteCodeBuffer);
      this.emit(buffer);
      return this;
    }
  }, {
    key: "emit",
    value: function emit(buffer) {
      if (buffer != null) {
        this.buffers.push(buffer);
      }

      return this;
    }
  }, {
    key: "build",
    value: function build() {
      return Buffer.concat(this.buffers);
    }
  }]);
  return ScriptBuilder;
}();

var utils = _extends({}, utils$1, {
  equals: equals,
  lazy: lazy,
  lazyAsync: lazyAsync,
  lazyOrValue: lazyOrValue,
  weightedAverage: weightedAverage,
  weightedFilter: weightedFilter
});

var createSerializeWire = function createSerializeWire(serializeWireBase) {
  return function () {
    var writer = new BinaryWriter();
    serializeWireBase(writer);
    return writer.toBuffer();
  };
};
function createDeserializeWire(deserializeWireBase) {
  return function (options) {
    return deserializeWireBase({
      context: options.context,
      reader: new BinaryReader(options.buffer)
    });
  };
}

var Account =
/*#__PURE__*/
function (_BaseState) {
  _inherits(Account, _BaseState);

  function Account(_ref) {
    var _this;

    var version = _ref.version,
        hash = _ref.hash,
        isFrozen = _ref.isFrozen,
        votes = _ref.votes,
        balances = _ref.balances;
    _classCallCheck(this, Account);
    _this = _possibleConstructorReturn(this, (Account.__proto__ || Object.getPrototypeOf(Account)).call(this, {
      version: version
    }));
    Object.defineProperty(_this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(Account, function (other) {
        return common.uInt160Equal(_this.hash, other.hash);
      })
    });
    Object.defineProperty(_this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(_this.serializeWireBase.bind(_this))
    });
    _this.hash = hash;
    _this.hashHex = common.uInt160ToHex(hash);
    _this.isFrozen = isFrozen || false;
    _this.votes = votes || [];
    _this.balances = balances || {};
    return _this;
  }

  _createClass(Account, [{
    key: "isDeletable",
    value: function isDeletable() {
      var balances = utils.values(this.balances);
      return !this.isFrozen && this.votes.length === 0 && (balances.length === 0 || balances.every(function (value) {
        return value.lte(utils.ZERO);
      }));
    }
  }, {
    key: "update",
    value: function update(_ref2) {
      var isFrozen = _ref2.isFrozen,
          votes = _ref2.votes,
          balances = _ref2.balances;
      return new Account({
        hash: this.hash,
        isFrozen: isFrozen == null ? this.isFrozen : isFrozen,
        votes: votes == null ? this.votes : votes,
        balances: balances == null ? this.balances : balances
      });
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt8(this.version);
      writer.writeUInt160(this.hash);
      writer.writeBoolean(this.isFrozen);
      writer.writeArray(this.votes, function (vote) {
        writer.writeECPoint(vote);
      });

      var balances = _.pickBy(this.balances, function (value) {
        return value.gt(utils.ZERO);
      });

      writer.writeObject(balances, function (key, value) {
        writer.writeUInt256(key);
        writer.writeFixed8(value);
      });
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        version: this.version,
        script_hash: JSONHelper.writeUInt160(this.hash),
        frozen: this.isFrozen,
        votes: this.votes.map(function (vote) {
          return JSONHelper.writeECPoint(vote);
        }),
        balances: utils.entries(this.balances).map(function (_ref3) {
          var _ref4 = _slicedToArray(_ref3, 2),
              asset = _ref4[0],
              value = _ref4[1];

          return {
            asset: JSONHelper.writeUInt256(asset),
            value: JSONHelper.writeFixed8(value)
          };
        })
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref5) {
      var reader = _ref5.reader;
      var version = reader.readUInt8();
      var hash = reader.readUInt160();
      var isFrozen = reader.readBoolean();
      var votes = reader.readArray(function () {
        return reader.readECPoint();
      });
      var balances = reader.readObject(function () {
        var key = common.uInt256ToHex(reader.readUInt256());
        var value = reader.readFixed8();
        return {
          key: key,
          value: value
        };
      });
      return new this({
        version: version,
        hash: hash,
        isFrozen: isFrozen,
        votes: votes,
        balances: balances
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return Account;
}(BaseState);

var _fixBabelExtend$4 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var ASSET_TYPE = {
  CREDIT_FLAG: 0x40,
  DUTY_FLAG: 0x80,
  GOVERNING_TOKEN: 0x00,
  UTILITY_TOKEN: 0x01,
  CURRENCY: 0x08,
  SHARE: 0x90,
  INVOICE: 0x98,
  TOKEN: 0x60
};
var hasFlag = function hasFlag(assetType, flag // eslint-disable-next-line
) {
  return (assetType & flag) === flag;
};
var InvalidAssetTypeError = _fixBabelExtend$4(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidAssetTypeError, _Error);

  function InvalidAssetTypeError(assetType) {
    var _this;

    _classCallCheck(this, InvalidAssetTypeError);
    _this = _possibleConstructorReturn(this, (InvalidAssetTypeError.__proto__ || Object.getPrototypeOf(InvalidAssetTypeError)).call(this, "Expected asset type, found: ".concat(assetType.toString(16))));
    _this.assetType = assetType;
    return _this;
  }

  return InvalidAssetTypeError;
}(Error));
var assertAssetType = function assertAssetType(assetType) {
  switch (assetType) {
    case ASSET_TYPE.CREDIT_FLAG:
      return ASSET_TYPE.CREDIT_FLAG;

    case ASSET_TYPE.DUTY_FLAG:
      return ASSET_TYPE.DUTY_FLAG;

    case ASSET_TYPE.GOVERNING_TOKEN:
      return ASSET_TYPE.GOVERNING_TOKEN;

    case ASSET_TYPE.UTILITY_TOKEN:
      return ASSET_TYPE.UTILITY_TOKEN;

    case ASSET_TYPE.CURRENCY:
      return ASSET_TYPE.CURRENCY;

    case ASSET_TYPE.SHARE:
      return ASSET_TYPE.SHARE;

    case ASSET_TYPE.INVOICE:
      return ASSET_TYPE.INVOICE;

    case ASSET_TYPE.TOKEN:
      return ASSET_TYPE.TOKEN;

    default:
      throw new InvalidAssetTypeError(assetType);
  }
};
var toJSONAssetType = function toJSONAssetType(type) {
  switch (type) {
    case 0x40:
      return 'CreditFlag';

    case 0x80:
      return 'DutyFlag';

    case 0x00:
      return 'GoverningToken';

    case 0x01:
      return 'UtilityToken';

    case 0x08:
      return 'Currency';

    case 0x90:
      return 'Share';

    case 0x98:
      return 'Invoice';

    case 0x60:
      return 'Token';

    default:
      // eslint-disable-next-line
      
      throw new InvalidAssetTypeError(type);
  }
};

var Witness =
/*#__PURE__*/
function () {
  function Witness(_ref) {
    var _this = this;

    var invocation = _ref.invocation,
        verification = _ref.verification;
    _classCallCheck(this, Witness);
    Object.defineProperty(this, "__size", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.lazy(function () {
        return IOHelper.sizeOfVarBytesLE(_this.invocation) + IOHelper.sizeOfVarBytesLE(_this.verification);
      })
    });
    Object.defineProperty(this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(Witness, function (other) {
        return _this.invocation.equals(other.invocation) && _this.verification.equals(other.verification);
      })
    });
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.invocation = invocation;
    this.verification = verification;
  }

  _createClass(Witness, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeVarBytesLE(this.invocation);
      writer.writeVarBytesLE(this.verification);
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        invocation: JSONHelper.writeBuffer(this.invocation),
        verification: JSONHelper.writeBuffer(this.verification)
      };
    }
  }, {
    key: "size",
    get: function get() {
      return this.__size();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref2) {
      var reader = _ref2.reader;
      var invocation = reader.readVarBytesLE(utils.USHORT_MAX_NUMBER_PLUS_ONE);
      var verification = reader.readVarBytesLE(utils.USHORT_MAX_NUMBER_PLUS_ONE);
      return new this({
        invocation: invocation,
        verification: verification
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return Witness;
}();

var _fixBabelExtend$5 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var ec$1 = new elliptic.ec('p256');
var Base58CheckError = _fixBabelExtend$5(
/*#__PURE__*/
function (_Error) {
  _inherits(Base58CheckError, _Error);

  function Base58CheckError() {
    _classCallCheck(this, Base58CheckError);
    return _possibleConstructorReturn(this, (Base58CheckError.__proto__ || Object.getPrototypeOf(Base58CheckError)).call(this, 'Base58 Check Decode Error.'));
  }

  return Base58CheckError;
}(Error));
var InvalidAddressError = _fixBabelExtend$5(
/*#__PURE__*/
function (_Error2) {
  _inherits(InvalidAddressError, _Error2);

  function InvalidAddressError(address) {
    var _this;

    _classCallCheck(this, InvalidAddressError);
    _this = _possibleConstructorReturn(this, (InvalidAddressError.__proto__ || Object.getPrototypeOf(InvalidAddressError)).call(this, "Invalid Address: ".concat(address)));
    _this.address = address;
    return _this;
  }

  return InvalidAddressError;
}(Error));

var sha1 = function sha1(value) {
  return crypto.createHash('sha1').update(value).digest();
};

var sha256 = function sha256(value) {
  return crypto.createHash('sha256').update(value).digest();
};

var rmd160 = function rmd160(value) {
  return crypto.createHash('rmd160').update(value).digest();
};

var hash160 = function hash160(value) {
  return common.bufferToUInt160(rmd160(sha256(value)));
};

var hash256 = function hash256(value) {
  return common.bufferToUInt256(sha256(sha256(value)));
}; // TODO: This + verify should probably handle DER format signatures as well


var sign = function sign(_ref) {
  var message = _ref.message,
      privateKey = _ref.privateKey;
  var sig = ec$1.sign(sha256(message), common.privateKeyToBuffer(privateKey));
  return Buffer.concat([sig.r.toArrayLike(Buffer, 'be', 32), sig.s.toArrayLike(Buffer, 'be', 32)]);
};

var InvalidSignatureError = _fixBabelExtend$5(
/*#__PURE__*/
function (_Error3) {
  _inherits(InvalidSignatureError, _Error3);

  function InvalidSignatureError() {
    _classCallCheck(this, InvalidSignatureError);
    return _possibleConstructorReturn(this, (InvalidSignatureError.__proto__ || Object.getPrototypeOf(InvalidSignatureError)).call(this, 'Invalid Signature'));
  }

  return InvalidSignatureError;
}(Error));

var verify = function verify(_ref2) {
  var message = _ref2.message,
      signature = _ref2.signature,
      publicKey = _ref2.publicKey;

  if (signature.length !== 64) {
    throw new InvalidSignatureError();
  }

  var r = signature.slice(0, 32);
  var s = signature.slice(32);
  return ec$1.verify(sha256(message), {
    r: r,
    s: s
  }, publicKey);
};

var InvalidPrivateKeyError = _fixBabelExtend$5(
/*#__PURE__*/
function (_Error4) {
  _inherits(InvalidPrivateKeyError, _Error4);

  function InvalidPrivateKeyError() {
    _classCallCheck(this, InvalidPrivateKeyError);
    return _possibleConstructorReturn(this, (InvalidPrivateKeyError.__proto__ || Object.getPrototypeOf(InvalidPrivateKeyError)).call(this, 'Invalid Private Key'));
  }

  return InvalidPrivateKeyError;
}(Error));

var toECPointFromKeyPair = function toECPointFromKeyPair(pair) {
  return common.bufferToECPoint(Buffer.from(pair.getPublic(true, 'hex'), 'hex'));
};

var getPublicKey = function getPublicKey(privateKey) {
  var key = ec$1.keyFromPrivate(common.privateKeyToBuffer(privateKey));
  key.getPublic(true, 'hex');

  var _key$validate = key.validate(),
      result = _key$validate.result;

  if (!result) {
    throw new InvalidPrivateKeyError();
  }

  return toECPointFromKeyPair(key);
};

var createKeyPair = function createKeyPair() {
  var key = ec$1.genKeyPair();
  return {
    privateKey: common.bufferToPrivateKey(key.getPrivate().toArrayLike(Buffer, 'be')),
    publicKey: toECPointFromKeyPair(key)
  };
};

var createPrivateKey = function createPrivateKey() {
  return common.bufferToPrivateKey(crypto.randomBytes(32));
};

var toScriptHash = function toScriptHash(value) {
  return hash160(value);
}; // Takes various formats and converts to standard ECPoint


var toECPoint = function toECPoint(publicKey) {
  return toECPointFromKeyPair(ec$1.keyFromPublic(publicKey));
};

var isInfinity = function isInfinity(ecPoint) {
  return ec$1.keyFromPublic(ecPoint).getPublic().isInfinity();
};

var base58Checksum = function base58Checksum(buffer) {
  return common.uInt256ToBuffer(hash256(buffer)).slice(0, 4);
};

var base58CheckEncode = function base58CheckEncode(buffer) {
  var checksum = base58Checksum(buffer);
  return bs58.encode(Buffer.concat([buffer, checksum]));
};

var base58CheckDecode = function base58CheckDecode(value) {
  var buffer = bs58.decode(value);
  var payload = buffer.slice(0, -4);
  var checksum = buffer.slice(-4);
  var payloadChecksum = base58Checksum(payload);

  if (!checksum.equals(payloadChecksum)) {
    throw new Base58CheckError();
  }

  return payload;
};

var scriptHashToAddress = function scriptHashToAddress(_ref3) {
  var addressVersion = _ref3.addressVersion,
      scriptHash = _ref3.scriptHash;
  var buffer = Buffer.allocUnsafe(21);
  buffer[0] = addressVersion;
  common.uInt160ToBuffer(scriptHash).copy(buffer, 1);
  return base58CheckEncode(buffer);
};

var addressToScriptHash = function addressToScriptHash(_ref4) {
  var addressVersion = _ref4.addressVersion,
      address = _ref4.address;
  var decodedAddress = base58CheckDecode(address);

  if (decodedAddress.length !== 21 || decodedAddress[0] !== addressVersion) {
    throw new InvalidAddressError(address);
  }

  return common.bufferToUInt160(decodedAddress.slice(1));
};

var createInvocationScript = function createInvocationScript(message, privateKey) {
  var builder = new ScriptBuilder();
  builder.emitPush(sign({
    message: message,
    privateKey: privateKey
  }));
  return builder.build();
};

var createVerificationScript = function createVerificationScript(publicKey) {
  var builder = new ScriptBuilder();
  builder.emitPushECPoint(publicKey);
  builder.emitOp('CHECKSIG');
  return builder.build();
};

var createWitness = function createWitness(message, privateKey) {
  var verification = createVerificationScript(getPublicKey(privateKey));
  var invocation = createInvocationScript(message, privateKey);
  return new Witness({
    verification: verification,
    invocation: invocation
  });
};

var getVerificationScriptHash = function getVerificationScriptHash(publicKey) {
  return toScriptHash(createVerificationScript(publicKey));
};

var compareKeys = function compareKeys(a, b) {
  var aPublic = a.getPublic();
  var bPublic = b.getPublic();
  var result = aPublic.getX().cmp(bPublic.getX());

  if (result !== 0) {
    return result;
  }

  return aPublic.getY().cmp(bPublic.getY());
};

var sortKeys = function sortKeys(publicKeys) {
  return publicKeys.map(function (publicKey) {
    return ec$1.keyFromPublic(publicKey);
  }).sort(compareKeys).map(function (keyPair) {
    return toECPointFromKeyPair(keyPair);
  });
};

var createMultiSignatureRedeemScript = function createMultiSignatureRedeemScript(m, publicKeys) {
  if (!(m >= 1 && m <= publicKeys.length && publicKeys.length <= 1024)) {
    // TODO: Better error
    throw new Error();
  }

  var builder = new ScriptBuilder();
  builder.emitPushInt(Math.floor(m));
  var publicKeysSorted = sortKeys(publicKeys);
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = publicKeysSorted[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var _ecPoint = _step.value;
      builder.emitPushECPoint(_ecPoint);
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return != null) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  builder.emitPushInt(publicKeysSorted.length);
  builder.emitOp('CHECKMULTISIG');
  return builder.build();
};

var getConsensusAddress = function getConsensusAddress(validators) {
  return toScriptHash(createMultiSignatureRedeemScript(validators.length - (validators.length - 1) / 3, validators));
};

var wifToPrivateKey = function wifToPrivateKey(wif, privateKeyVersion) {
  var privateKeyDecoded = base58CheckDecode(wif);

  if (privateKeyDecoded.length !== 34 || privateKeyDecoded[0] !== privateKeyVersion || privateKeyDecoded[33] !== 0x01) {
    throw new InvalidFormatError();
  }

  return common.bufferToPrivateKey(privateKeyDecoded.slice(1, 33));
};

var privateKeyToWif = function privateKeyToWif(privateKey, privateKeyVersion) {
  return WIF.encode(privateKeyVersion, common.privateKeyToBuffer(privateKey), true);
};

var privateKeyToScriptHash = function privateKeyToScriptHash(privateKey) {
  return getVerificationScriptHash(getPublicKey(privateKey));
};

var privateKeyToAddress = function privateKeyToAddress(_ref5) {
  var addressVersion = _ref5.addressVersion,
      privateKey = _ref5.privateKey;
  return scriptHashToAddress({
    addressVersion: addressVersion,
    scriptHash: privateKeyToScriptHash(privateKey)
  });
};

var NEP2_KDFPARAMS = {
  N: 16384,
  r: 8,
  p: 8,
  dklen: 64
};
var NEP2_ZERO = 0x01;
var NEP2_ONE = 0x42;
var NEP2_TWO = 0xe0;
var NEP2_CIPHER = 'aes-256-ecb';

var getNEP2Derived = function getNEP2Derived(_ref6) {
  var password = _ref6.password,
      salt = _ref6.salt;
  return new Promise(function (resolve, reject) {
    return scrypt(Buffer.from(password, 'utf8'), salt, NEP2_KDFPARAMS.N, NEP2_KDFPARAMS.r, NEP2_KDFPARAMS.p, NEP2_KDFPARAMS.dklen, function (error, progress, key) {
      if (error != null) {
        reject(error);
      } else if (key) {
        resolve(Buffer.from(key));
      }
    });
  });
};

var getNEP2Salt = function getNEP2Salt(_ref7) {
  var addressVersion = _ref7.addressVersion,
      privateKey = _ref7.privateKey;
  var address = privateKeyToAddress({
    addressVersion: addressVersion,
    privateKey: privateKey
  });
  return common.uInt256ToBuffer(hash256(Buffer.from(address, 'latin1'))).slice(0, 4);
};

var encryptNEP2 = function () {
  var _ref9 = _asyncToGenerator(
  /*#__PURE__*/
  regeneratorRuntime.mark(function _callee(_ref8) {
    var addressVersion, password, privateKey, salt, derived, derived1, derived2, cipher, ciphertext, result;
    return regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            addressVersion = _ref8.addressVersion, password = _ref8.password, privateKey = _ref8.privateKey;
            salt = getNEP2Salt({
              addressVersion: addressVersion,
              privateKey: privateKey
            });
            _context.next = 4;
            return getNEP2Derived({
              password: password,
              salt: salt
            });

          case 4:
            derived = _context.sent;
            derived1 = derived.slice(0, 32);
            derived2 = derived.slice(32, 64);
            cipher = crypto.createCipheriv(NEP2_CIPHER, derived2, Buffer.alloc(0, 0));
            cipher.setAutoPadding(false);
            cipher.end(xor(privateKey, derived1));
            ciphertext = cipher.final();
            result = Buffer.alloc(7 + 32, 0);
            result.writeUInt8(0x01, NEP2_ZERO);
            result.writeUInt8(0x42, NEP2_ONE);
            result.writeUInt8(0xe0, NEP2_TWO);
            salt.copy(result, 3);
            ciphertext.copy(result, 7);
            return _context.abrupt("return", base58CheckEncode(result));

          case 18:
          case "end":
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function encryptNEP2(_x) {
    return _ref9.apply(this, arguments);
  };
}();

var decryptNEP2 = function () {
  var _ref11 = _asyncToGenerator(
  /*#__PURE__*/
  regeneratorRuntime.mark(function _callee2(_ref10) {
    var addressVersion, encryptedKey, password, decoded, salt, derived, derived1, derived2, decipher, plainText, privateKey, addressSalt;
    return regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            addressVersion = _ref10.addressVersion, encryptedKey = _ref10.encryptedKey, password = _ref10.password;
            decoded = base58CheckDecode(encryptedKey);

            if (!(decoded.length !== 39 || decoded.readUInt8(0) !== NEP2_ZERO || decoded.readUInt8(1) !== NEP2_ONE || decoded.readUInt8(2) !== NEP2_TWO)) {
              _context2.next = 4;
              break;
            }

            throw new Error('Invalid NEP2 format.');

          case 4:
            salt = decoded.slice(3, 7);
            _context2.next = 7;
            return getNEP2Derived({
              password: password,
              salt: salt
            });

          case 7:
            derived = _context2.sent;
            derived1 = derived.slice(0, 32);
            derived2 = derived.slice(32, 64);
            decipher = crypto.createDecipheriv(NEP2_CIPHER, derived2, Buffer.alloc(0, 0));
            decipher.setAutoPadding(false);
            decipher.end(decoded.slice(7, 7 + 32));
            plainText = decipher.final();
            privateKey = xor(plainText, derived1);
            addressSalt = getNEP2Salt({
              addressVersion: addressVersion,
              privateKey: privateKey
            });

            if (salt.equals(addressSalt)) {
              _context2.next = 18;
              break;
            }

            throw new Error('Wrong passphrase.');

          case 18:
            return _context2.abrupt("return", common.bufferToPrivateKey(privateKey));

          case 19:
          case "end":
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function decryptNEP2(_x2) {
    return _ref11.apply(this, arguments);
  };
}();

var crypto$2 = {
  sha1: sha1,
  sha256: sha256,
  hash160: hash160,
  hash256: hash256,
  sign: sign,
  verify: verify,
  getPublicKey: getPublicKey,
  toScriptHash: toScriptHash,
  toECPoint: toECPoint,
  isInfinity: isInfinity,
  createKeyPair: createKeyPair,
  scriptHashToAddress: scriptHashToAddress,
  addressToScriptHash: addressToScriptHash,
  createInvocationScript: createInvocationScript,
  createVerificationScript: createVerificationScript,
  createWitness: createWitness,
  getVerificationScriptHash: getVerificationScriptHash,
  createMultiSignatureRedeemScript: createMultiSignatureRedeemScript,
  getConsensusAddress: getConsensusAddress,
  privateKeyToWif: privateKeyToWif,
  wifToPrivateKey: wifToPrivateKey,
  privateKeyToScriptHash: privateKeyToScriptHash,
  privateKeyToAddress: privateKeyToAddress,
  encryptNEP2: encryptNEP2,
  decryptNEP2: decryptNEP2,
  createPrivateKey: createPrivateKey
};

var _fixBabelExtend$6 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var InvalidMerkleTreeException = _fixBabelExtend$6(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidMerkleTreeException, _Error);

  function InvalidMerkleTreeException() {
    _classCallCheck(this, InvalidMerkleTreeException);
    return _possibleConstructorReturn(this, (InvalidMerkleTreeException.__proto__ || Object.getPrototypeOf(InvalidMerkleTreeException)).call(this, 'Invalid Merkle tree.'));
  }

  return InvalidMerkleTreeException;
}(Error));

var MerkleTreeNode =
/*#__PURE__*/
function () {
  function MerkleTreeNode(_ref) {
    var hash = _ref.hash,
        leftChild = _ref.leftChild,
        rightChild = _ref.rightChild,
        parent = _ref.parent;
    _classCallCheck(this, MerkleTreeNode);
    this.hash = hash;
    this.leftChild = leftChild;
    this.rightChild = rightChild;
    this.parent = parent;
  }

  _createClass(MerkleTreeNode, [{
    key: "clone",
    value: function clone(parent) {
      var self = new MerkleTreeNode({
        hash: this.hash,
        parent: parent
      });

      if (this.leftChild != null) {
        self.leftChild = this.leftChild.clone(self);
      }

      if (this.rightChild != null) {
        self.rightChild = this.rightChild.clone(self);
      }

      return self;
    }
  }]);
  return MerkleTreeNode;
}();

var build = function build(leavesIn) {
  var leaves = leavesIn;

  if (leaves.length === 0) {
    throw new InvalidMerkleTreeException();
  }

  if (leaves.length === 1) {
    return leaves[0];
  }

  var parents = [];
  var length = Math.floor((leaves.length + 1) / 2);

  for (var i = 0; i < length; i += 1) {
    var leftChild = leaves[i * 2];
    var rightChild = void 0;

    if (i * 2 + 1 === leaves.length) {
      rightChild = leftChild;
    } else {
      rightChild = leaves[i * 2 + 1];
    }

    var node = new MerkleTreeNode({
      hash: crypto$2.hash256(Buffer.concat([common.uInt256ToBuffer(leftChild.hash), common.uInt256ToBuffer(rightChild.hash)])),
      leftChild: leftChild,
      rightChild: rightChild
    });
    parents[i] = node;
    leaves[i * 2].parent = node;

    if (i * 2 + 1 !== leaves.length) {
      leaves[i * 2 + 1].parent = node;
    }
  }

  return build(parents);
};

var _trim = function _trim(node, index, depth, flags) {
  if (depth === 1 || node.leftChild == null) {
    return;
  }

  if (depth === 2) {
    if (!flags[index * 2] && !flags[index * 2 + 1]) {
      // eslint-disable-next-line
      node.leftChild = null; // eslint-disable-next-line

      node.rightChild = null;
    }
  } else {
    var leftChild = node.leftChild,
        rightChild = node.rightChild;

    if (leftChild != null && rightChild != null) {
      _trim(leftChild, index * 2, depth - 1, flags);

      _trim(rightChild, index * 2 + 1, depth - 1, flags);

      if (node.leftChild != null && node.rightChild != null && node.leftChild.leftChild == null && node.rightChild.rightChild == null) {
        // eslint-disable-next-line
        node.leftChild = null; // eslint-disable-next-line

        node.rightChild = null;
      }
    }
  }
};

var _depthFirstSearch = function _depthFirstSearch(node, hashes) {
  var leftChild = node.leftChild,
      rightChild = node.rightChild;

  if (leftChild == null || rightChild == null) {
    hashes.push(node.hash);
  } else {
    _depthFirstSearch(leftChild, hashes);

    _depthFirstSearch(rightChild, hashes);
  }
};

var _depthFirstSearch2 = function _depthFirstSearch2(node) {
  var hashes = [];

  _depthFirstSearch(node, hashes);

  return hashes;
};

var MerkleTree =
/*#__PURE__*/
function () {
  function MerkleTree(hashesOrNode) {
    _classCallCheck(this, MerkleTree);
    this.root = Array.isArray(hashesOrNode) ? build(hashesOrNode.map(function (hash) {
      return new MerkleTreeNode({
        hash: hash
      });
    })) : hashesOrNode;
    this.depth = 1;

    for (var node = this.root; node.leftChild != null; node = node.leftChild) {
      this.depth += 1;
    }
  }

  _createClass(MerkleTree, [{
    key: "trim",
    value: function trim(flags) {
      var result = this.root.clone();

      _trim(result, 0, this.depth, flags);

      return new MerkleTree(result);
    }
  }, {
    key: "depthFirstSearch",
    value: function depthFirstSearch() {
      return _depthFirstSearch2(this.root);
    }
  }, {
    key: "toHashArray",
    value: function toHashArray() {
      return this.depthFirstSearch();
    }
  }], [{
    key: "computeRoot",
    value: function computeRoot(hashes) {
      var tree = new this(hashes);
      return tree.root.hash;
    }
  }]);
  return MerkleTree;
}();

var _fixBabelExtend$3 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var InvalidAssetError = _fixBabelExtend$3(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidAssetError, _Error);

  function InvalidAssetError() {
    _classCallCheck(this, InvalidAssetError);
    return _possibleConstructorReturn(this, (InvalidAssetError.__proto__ || Object.getPrototypeOf(InvalidAssetError)).apply(this, arguments));
  }

  return InvalidAssetError;
}(Error));

var Asset =
/*#__PURE__*/
function (_BaseState) {
  _inherits(Asset, _BaseState);

  function Asset(_ref) {
    var _this;

    var version = _ref.version,
        hash = _ref.hash,
        type = _ref.type,
        name = _ref.name,
        amount = _ref.amount,
        available = _ref.available,
        precision = _ref.precision,
        feeMode = _ref.feeMode,
        fee = _ref.fee,
        feeAddress = _ref.feeAddress,
        owner = _ref.owner,
        admin = _ref.admin,
        issuer = _ref.issuer,
        expiration = _ref.expiration,
        isFrozen = _ref.isFrozen;
    _classCallCheck(this, Asset);
    _this = _possibleConstructorReturn(this, (Asset.__proto__ || Object.getPrototypeOf(Asset)).call(this, {
      version: version
    })); // eslint-disable-next-line

    Object.defineProperty(_this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(Asset, function (other) {
        return common.uInt256Equal(_this.hash, other.hash);
      })
    });
    Object.defineProperty(_this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(_this.serializeWireBase.bind(_this))
    });
    verifyAsset({
      name: name,
      type: type,
      amount: amount,
      precision: precision
    });
    _this.hash = hash;
    _this.hashHex = common.uInt256ToHex(hash);
    _this.type = type;
    _this.name = name;
    _this.amount = amount;
    _this.available = available || utils.ZERO;
    _this.precision = precision;
    _this.feeMode = feeMode || 0;
    _this.fee = fee || utils.ZERO;
    _this.feeAddress = feeAddress || common.ZERO_UINT160;
    _this.owner = owner;
    _this.admin = admin;
    _this.issuer = issuer;
    _this.expiration = expiration;
    _this.isFrozen = isFrozen || false;
    return _this;
  }

  _createClass(Asset, [{
    key: "update",
    value: function update(_ref2) {
      var available = _ref2.available,
          expiration = _ref2.expiration,
          isFrozen = _ref2.isFrozen;
      return new Asset({
        hash: this.hash,
        type: this.type,
        name: this.name,
        amount: this.amount,
        precision: this.precision,
        fee: this.fee,
        feeAddress: this.feeAddress,
        owner: this.owner,
        admin: this.admin,
        issuer: this.issuer,
        available: available == null ? this.available : available,
        expiration: expiration == null ? this.expiration : expiration,
        isFrozen: isFrozen == null ? this.isFrozen : isFrozen
      });
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt8(this.version);
      writer.writeUInt256(this.hash);
      writer.writeUInt8(this.type);
      writer.writeVarString(this.name);
      writer.writeFixed8(this.amount);
      writer.writeFixed8(this.available);
      writer.writeUInt8(this.precision);
      writer.writeUInt8(this.feeMode);
      writer.writeFixed8(this.fee);
      writer.writeUInt160(this.feeAddress);
      writer.writeECPoint(this.owner);
      writer.writeUInt160(this.admin);
      writer.writeUInt160(this.issuer);
      writer.writeUInt32LE(this.expiration);
      writer.writeBoolean(this.isFrozen);
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      var name = this.name || '';

      try {
        name = JSON.parse(name);
      } catch (error) {// eslint-disable-next-line
      }

      return {
        version: this.version,
        id: JSONHelper.writeUInt256(this.hash),
        type: toJSONAssetType(this.type),
        name: name,
        amount: JSONHelper.writeFixed8(this.amount),
        available: JSONHelper.writeFixed8(this.available),
        precision: this.precision,
        owner: JSONHelper.writeECPoint(this.owner),
        admin: crypto$2.scriptHashToAddress({
          addressVersion: context.addressVersion,
          scriptHash: this.admin
        }),
        issuer: crypto$2.scriptHashToAddress({
          addressVersion: context.addressVersion,
          scriptHash: this.issuer
        }),
        expiration: this.expiration,
        frozen: this.isFrozen
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref3) {
      var reader = _ref3.reader;
      var version = reader.readUInt8();
      var hash = reader.readUInt256();
      var type = assertAssetType(reader.readUInt8());
      var name = reader.readVarString();
      var amount = reader.readFixed8();
      var available = reader.readFixed8();
      var precision = reader.readUInt8();
      reader.readUInt8(); // FeeMode

      var fee = reader.readFixed8();
      var feeAddress = reader.readUInt160();
      var owner = reader.readECPoint();
      var admin = reader.readUInt160();
      var issuer = reader.readUInt160();
      var expiration = reader.readUInt32LE();
      var isFrozen = reader.readBoolean();
      return new this({
        version: version,
        hash: hash,
        type: type,
        name: name,
        amount: amount,
        available: available,
        precision: precision,
        fee: fee,
        feeAddress: feeAddress,
        owner: owner,
        admin: admin,
        issuer: issuer,
        expiration: expiration,
        isFrozen: isFrozen
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return Asset;
}(BaseState);

var verifyAsset = function verifyAsset(_ref4) {
  var name = _ref4.name,
      type = _ref4.type,
      amount = _ref4.amount,
      precision = _ref4.precision;

  if (type === ASSET_TYPE.CREDIT_FLAG || type === ASSET_TYPE.DUTY_FLAG) {
    throw new InvalidAssetError('Asset type cannot be CREDIT_FLAG or DUTY_FLAG');
  }

  var nameBuffer = Buffer.from(name, 'utf8');

  if (nameBuffer.length > 1024) {
    throw new InvalidAssetError('Name too long');
  }

  if (amount.lte(utils.ZERO) && !amount.eq(common.NEGATIVE_SATOSHI_FIXED8)) {
    throw new InvalidAssetError('Amount must be greater than 0');
  }

  if (type === ASSET_TYPE.INVOICE && !amount.eq(common.NEGATIVE_SATOSHI_FIXED8)) {
    throw new InvalidAssetError('Invoice assets must have unlimited amount.');
  }

  if (precision > 8) {
    throw new InvalidAssetError('Max precision is 8.');
  }

  if (!amount.eq(utils.NEGATIVE_SATOSHI) && !amount.mod(utils.TEN.pow(utils.EIGHT.subn(precision))).eq(utils.ZERO)) {
    throw new InvalidAssetError('Invalid precision for amount.');
  }
};

var BlockBase =
/*#__PURE__*/
function () {
  function BlockBase(_ref) {
    var _this = this;

    var version = _ref.version,
        previousHash = _ref.previousHash,
        merkleRoot = _ref.merkleRoot,
        timestamp = _ref.timestamp,
        index = _ref.index,
        consensusData = _ref.consensusData,
        nextConsensus = _ref.nextConsensus,
        script = _ref.script,
        hash = _ref.hash;
    _classCallCheck(this, BlockBase);
    Object.defineProperty(this, "_hashHex", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.lazy(function () {
        return common.uInt256ToHex(_this.hash);
      })
    });
    Object.defineProperty(this, "_message", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.lazy(function () {
        return _this.serializeUnsigned();
      })
    });
    Object.defineProperty(this, "__thisSize", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.lazy(function () {
        return IOHelper.sizeOfUInt32LE + IOHelper.sizeOfUInt256 + IOHelper.sizeOfUInt256 + IOHelper.sizeOfUInt32LE + IOHelper.sizeOfUInt32LE + IOHelper.sizeOfUInt64LE + IOHelper.sizeOfUInt160 + IOHelper.sizeOfUInt8 + _this.script.size;
      })
    });
    Object.defineProperty(this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(this.constructor, function (other) {
        return common.uInt256Equal(_this.hash, other.hash);
      })
    });
    Object.defineProperty(this, "getScriptHashesForVerifying", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.lazyAsync(function () {
        var _ref3 = _asyncToGenerator(
        /*#__PURE__*/
        regeneratorRuntime.mark(function _callee(_ref2) {
          var getHeader, previousHeader;
          return regeneratorRuntime.wrap(function _callee$(_context) {
            while (1) {
              switch (_context.prev = _context.next) {
                case 0:
                  getHeader = _ref2.getHeader;

                  if (!(_this.index === 0)) {
                    _context.next = 3;
                    break;
                  }

                  return _context.abrupt("return", new Set([common.uInt160ToHex(crypto$2.toScriptHash(_this.script.verification))]));

                case 3:
                  _context.next = 5;
                  return getHeader({
                    hashOrIndex: _this.previousHash
                  });

                case 5:
                  previousHeader = _context.sent;
                  return _context.abrupt("return", new Set([common.uInt160ToHex(previousHeader.nextConsensus)]));

                case 7:
                case "end":
                  return _context.stop();
              }
            }
          }, _callee, this);
        }));

        return function (_x) {
          return _ref3.apply(this, arguments);
        };
      }())
    });
    Object.defineProperty(this, "serializeUnsigned", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(this.serializeUnsignedBase.bind(this))
    });
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.version = version || 0;
    this.previousHash = previousHash;
    this.merkleRoot = merkleRoot;
    this.timestamp = timestamp;
    this.index = index;
    this.consensusData = consensusData;
    this.nextConsensus = nextConsensus;
    this.script = script;
    var hashIn = hash;
    this._hash = hashIn == null ? utils.lazy(function () {
      return crypto$2.hash256(_this.message);
    }) : function () {
      return hashIn;
    };
  }

  _createClass(BlockBase, [{
    key: "serializeUnsignedBase",
    value: function serializeUnsignedBase(writer) {
      writer.writeUInt32LE(this.version);
      writer.writeUInt256(this.previousHash);
      writer.writeUInt256(this.merkleRoot);
      writer.writeUInt32LE(this.timestamp);
      writer.writeUInt32LE(this.index);
      writer.writeUInt64LE(this.consensusData);
      writer.writeUInt160(this.nextConsensus);
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      this.serializeUnsignedBase(writer);
      writer.writeUInt8(1);
      this.script.serializeWireBase(writer);
    }
  }, {
    key: "serializeBlockBaseJSON",
    value: function serializeBlockBaseJSON(context) {
      return {
        version: this.version,
        hash: JSONHelper.writeUInt256(this.hash),
        size: this.size,
        previousblockhash: JSONHelper.writeUInt256(this.previousHash),
        merkleroot: JSONHelper.writeUInt256(this.merkleRoot),
        time: this.timestamp,
        index: this.index,
        nonce: JSONHelper.writeUInt64LE(this.consensusData),
        nextconsensus: crypto$2.scriptHashToAddress({
          addressVersion: context.addressVersion,
          scriptHash: this.nextConsensus
        }),
        script: this.script.serializeJSON(context),
        // TODO: Not sure if we should bother with this...
        confirmations: 0
      };
    }
  }, {
    key: "hash",
    get: function get() {
      return this._hash();
    }
  }, {
    key: "hashHex",
    get: function get() {
      return this._hashHex();
    }
  }, {
    key: "message",
    get: function get() {
      return this._message();
    }
  }, {
    key: "size",
    get: function get() {
      return this.__thisSize();
    }
  }], [{
    key: "deserializeBlockBaseWireBase",
    value: function deserializeBlockBaseWireBase(options) {
      var reader = options.reader;
      var version = reader.readUInt32LE();
      var previousHash = reader.readUInt256();
      var merkleRoot = reader.readUInt256();
      var timestamp = reader.readUInt32LE();
      var index = reader.readUInt32LE();
      var consensusData = reader.readUInt64LE();
      var nextConsensus = reader.readUInt160();

      if (reader.readUInt8() !== 1) {
        throw new InvalidFormatError();
      }

      var script = Witness.deserializeWireBase(options);
      return {
        version: version,
        previousHash: previousHash,
        merkleRoot: merkleRoot,
        timestamp: timestamp,
        index: index,
        consensusData: consensusData,
        nextConsensus: nextConsensus,
        script: script
      };
    }
  }]);
  return BlockBase;
}();

/* @fl/*  */
var Header =
/*#__PURE__*/
function (_BlockBase) {
  _inherits(Header, _BlockBase);

  function Header(add) {
    var _this;

    _classCallCheck(this, Header);
    _this = _possibleConstructorReturn(this, (Header.__proto__ || Object.getPrototypeOf(Header)).call(this, add));
    _this.__headerSize = utils.lazy(function () {
      return _get(Header.prototype.__proto__ || Object.getPrototypeOf(Header.prototype), "size", _this) + IOHelper.sizeOfUInt8;
    });
    return _this;
  }

  _createClass(Header, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(Header.prototype.__proto__ || Object.getPrototypeOf(Header.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeUInt8(0);
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function () {
      var _ref = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee(context) {
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                return _context.abrupt("return", _get(Header.prototype.__proto__ || Object.getPrototypeOf(Header.prototype), "serializeBlockBaseJSON", this).call(this, context));

              case 1:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function serializeJSON(_x) {
        return _ref.apply(this, arguments);
      }

      return serializeJSON;
    }()
  }, {
    key: "size",
    get: function get() {
      return this.__headerSize();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      var blockBase = _get(Header.__proto__ || Object.getPrototypeOf(Header), "deserializeBlockBaseWireBase", this).call(this, options);

      if (reader.readUInt8() !== 0) {
        throw new InvalidFormatError();
      }

      return new this({
        version: blockBase.version,
        previousHash: blockBase.previousHash,
        merkleRoot: blockBase.merkleRoot,
        timestamp: blockBase.timestamp,
        index: blockBase.index,
        consensusData: blockBase.consensusData,
        nextConsensus: blockBase.nextConsensus,
        script: blockBase.script
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return Header;
}(BlockBase);

var _fixBabelExtend$7 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var SCRIPT_CONTAINER_TYPE = {
  TRANSACTION: 0x00,
  BLOCK: 0x01
};
var InvalidScriptContainerTypeError = _fixBabelExtend$7(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidScriptContainerTypeError, _Error);

  function InvalidScriptContainerTypeError(value) {
    var _this;

    _classCallCheck(this, InvalidScriptContainerTypeError);
    _this = _possibleConstructorReturn(this, (InvalidScriptContainerTypeError.__proto__ || Object.getPrototypeOf(InvalidScriptContainerTypeError)).call(this, "Expected script container type, found: ".concat(value.toString(16))));
    _this.value = value;
    return _this;
  }

  return InvalidScriptContainerTypeError;
}(Error));

var _fixBabelExtend$8 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var ATTRIBUTE_USAGE = {
  CONTRACTHASH: 0x00,
  ECDH02: 0x02,
  ECDH03: 0x03,
  SCRIPT: 0x20,
  VOTE: 0x30,
  DESCRIPTIONURL: 0x81,
  DESCRIPTION: 0x90,
  HASH1: 0xa1,
  HASH2: 0xa2,
  HASH3: 0xa3,
  HASH4: 0xa4,
  HASH5: 0xa5,
  HASH6: 0xa6,
  HASH7: 0xa7,
  HASH8: 0xa8,
  HASH9: 0xa9,
  HASH10: 0xaa,
  HASH11: 0xab,
  HASH12: 0xac,
  HASH13: 0xad,
  HASH14: 0xae,
  HASH15: 0xaf,
  REMARK: 0xf0,
  REMARK1: 0xf1,
  REMARK2: 0xf2,
  REMARK3: 0xf3,
  REMARK4: 0xf4,
  REMARK5: 0xf5,
  REMARK6: 0xf6,
  REMARK7: 0xf7,
  REMARK8: 0xf8,
  REMARK9: 0xf9,
  REMARK10: 0xfa,
  REMARK11: 0xfb,
  REMARK12: 0xfc,
  REMARK13: 0xfd,
  REMARK14: 0xfe,
  REMARK15: 0xff
};
var InvalidAttributeUsageError = _fixBabelExtend$8(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidAttributeUsageError, _Error);

  function InvalidAttributeUsageError(transactionAttributeUsage) {
    var _this;

    _classCallCheck(this, InvalidAttributeUsageError);
    _this = _possibleConstructorReturn(this, (InvalidAttributeUsageError.__proto__ || Object.getPrototypeOf(InvalidAttributeUsageError)).call(this, "Expected transaction attribute usage, " + "found: ".concat(transactionAttributeUsage.toString(16))));
    _this.transactionAttributeUsage = transactionAttributeUsage;
    return _this;
  }

  return InvalidAttributeUsageError;
}(Error));
var assertAttributeUsage = function assertAttributeUsage(value) {
  switch (value) {
    case 0x00:
      // ContractHash
      return 0x00;

    case 0x02:
      // ECDH02
      return 0x02;

    case 0x03:
      // ECDH03
      return 0x03;

    case 0x20:
      // Script
      return 0x20;

    case 0x30:
      // Vote
      return 0x30;

    case 0x81:
      // DescriptionUrl
      return 0x81;

    case 0x90:
      // Description
      return 0x90;

    case 0xa1:
      // Hash1
      return 0xa1;

    case 0xa2:
      // Hash2
      return 0xa2;

    case 0xa3:
      // Hash3
      return 0xa3;

    case 0xa4:
      // Hash4
      return 0xa4;

    case 0xa5:
      // Hash5
      return 0xa5;

    case 0xa6:
      // Hash6
      return 0xa6;

    case 0xa7:
      // Hash7
      return 0xa7;

    case 0xa8:
      // Hash8
      return 0xa8;

    case 0xa9:
      // Hash9
      return 0xa9;

    case 0xaa:
      // Hash10
      return 0xaa;

    case 0xab:
      // Hash11
      return 0xab;

    case 0xac:
      // Hash12
      return 0xac;

    case 0xad:
      // Hash13
      return 0xad;

    case 0xae:
      // Hash14
      return 0xae;

    case 0xaf:
      // Hash15
      return 0xaf;

    case 0xf0:
      // Remark
      return 0xf0;

    case 0xf1:
      // Remark1
      return 0xf1;

    case 0xf2:
      // Remark2
      return 0xf2;

    case 0xf3:
      // Remark3
      return 0xf3;

    case 0xf4:
      // Remark4
      return 0xf4;

    case 0xf5:
      // Remark5
      return 0xf5;

    case 0xf6:
      // Remark6
      return 0xf6;

    case 0xf7:
      // Remark7
      return 0xf7;

    case 0xf8:
      // Remark8
      return 0xf8;

    case 0xf9:
      // Remark9
      return 0xf9;

    case 0xfa:
      // Remark10
      return 0xfa;

    case 0xfb:
      // Remark11
      return 0xfb;

    case 0xfc:
      // Remark12
      return 0xfc;

    case 0xfd:
      // Remark13
      return 0xfd;

    case 0xfe:
      // Remark14
      return 0xfe;

    case 0xff:
      // Remark15
      return 0xff;

    default:
      throw new InvalidAttributeUsageError(value);
  }
};
var InvalidJSONAttributeUsageError = _fixBabelExtend$8(
/*#__PURE__*/
function (_Error2) {
  _inherits(InvalidJSONAttributeUsageError, _Error2);

  function InvalidJSONAttributeUsageError(transactionAttributeUsage) {
    var _this2;

    _classCallCheck(this, InvalidJSONAttributeUsageError);
    _this2 = _possibleConstructorReturn(this, (InvalidJSONAttributeUsageError.__proto__ || Object.getPrototypeOf(InvalidJSONAttributeUsageError)).call(this, "Expected transaction attribute usage, " + "found: ".concat(transactionAttributeUsage)));
    _this2.transactionAttributeUsage = transactionAttributeUsage;
    return _this2;
  }

  return InvalidJSONAttributeUsageError;
}(Error));
var toJSONAttributeUsage = function toJSONAttributeUsage(usage) {
  switch (usage) {
    case 0x00:
      return 'ContractHash';

    case 0x02:
      return 'ECDH02';

    case 0x03:
      return 'ECDH03';

    case 0x20:
      return 'Script';

    case 0x30:
      return 'Vote';

    case 0x81:
      return 'DescriptionUrl';

    case 0x90:
      return 'Description';

    case 0xa1:
      return 'Hash1';

    case 0xa2:
      return 'Hash2';

    case 0xa3:
      return 'Hash3';

    case 0xa4:
      return 'Hash4';

    case 0xa5:
      return 'Hash5';

    case 0xa6:
      return 'Hash6';

    case 0xa7:
      return 'Hash7';

    case 0xa8:
      return 'Hash8';

    case 0xa9:
      return 'Hash9';

    case 0xaa:
      return 'Hash10';

    case 0xab:
      return 'Hash11';

    case 0xac:
      return 'Hash12';

    case 0xad:
      return 'Hash13';

    case 0xae:
      return 'Hash14';

    case 0xaf:
      return 'Hash15';

    case 0xf0:
      return 'Remark';

    case 0xf1:
      return 'Remark1';

    case 0xf2:
      return 'Remark2';

    case 0xf3:
      return 'Remark3';

    case 0xf4:
      return 'Remark4';

    case 0xf5:
      return 'Remark5';

    case 0xf6:
      return 'Remark6';

    case 0xf7:
      return 'Remark7';

    case 0xf8:
      return 'Remark8';

    case 0xf9:
      return 'Remark9';

    case 0xfa:
      return 'Remark10';

    case 0xfb:
      return 'Remark11';

    case 0xfc:
      return 'Remark12';

    case 0xfd:
      return 'Remark13';

    case 0xfe:
      return 'Remark14';

    case 0xff:
      return 'Remark15';

    default:
      // eslint-disable-next-line
      
      throw new InvalidAttributeUsageError(usage);
  }
};

var AttributeBase =
/*#__PURE__*/
function () {
  function AttributeBase() {
    _classCallCheck(this, AttributeBase);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(this.serializeWireBase.bind(this))
    });
  }

  _createClass(AttributeBase, [{
    key: "equals",
    value: function equals(other) {
      return this === other;
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt8(this.usage);
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      throw new Error('Not Implemented');
    }
  }], [{
    key: "deserializeAttributeWireBase",
    // eslint-disable-next-line
    value: function deserializeAttributeWireBase(_ref) {
      var reader = _ref.reader;
      var usage = reader.readUInt8();
      return {
        usage: usage
      };
    } // eslint-disable-next-line

  }, {
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      throw new Error('Not Implemented');
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return AttributeBase;
}();

var BufferAttribute =
/*#__PURE__*/
function (_AttributeBase) {
  _inherits(BufferAttribute, _AttributeBase);

  function BufferAttribute(_ref) {
    var _this;

    var usage = _ref.usage,
        value = _ref.value;
    _classCallCheck(this, BufferAttribute);
    _this = _possibleConstructorReturn(this, (BufferAttribute.__proto__ || Object.getPrototypeOf(BufferAttribute)).call(this));
    Object.defineProperty(_this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(BufferAttribute, function (other) {
        return _this.usage === other.usage && _this.value.equals(other.value);
      })
    });
    _this.usage = usage;
    _this.value = value;

    if (_this.usage === 0x81) {
      _this.size = IOHelper.sizeOfUInt8 + IOHelper.sizeOfUInt8 + _this.value.length;
    } else {
      _this.size = IOHelper.sizeOfUInt8 + IOHelper.sizeOfVarBytesLE(_this.value);
    }

    return _this;
  }

  _createClass(BufferAttribute, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(BufferAttribute.prototype.__proto__ || Object.getPrototypeOf(BufferAttribute.prototype), "serializeWireBase", this).call(this, writer);

      if (this.usage === 0x81) {
        writer.writeUInt8(this.value.length);
        writer.writeBytes(this.value);
      } else {
        writer.writeVarBytesLE(this.value);
      }
    } // eslint-disable-next-line

  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        usage: toJSONAttributeUsage(this.usage),
        data: JSONHelper.writeBuffer(this.value)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;

      var _rollupPluginBabelHel = _get(BufferAttribute.__proto__ || Object.getPrototypeOf(BufferAttribute), "deserializeAttributeWireBase", this).call(this, options),
          usage = _rollupPluginBabelHel.usage;

      if (!(usage === 0x81 || usage === 0x90 || usage === 0xf0 || usage === 0xf1 || usage === 0xf2 || usage === 0xf3 || usage === 0xf4 || usage === 0xf5 || usage === 0xf6 || usage === 0xf7 || usage === 0xf8 || usage === 0xf9 || usage === 0xfa || usage === 0xfb || usage === 0xfc || usage === 0xfd || usage === 0xfe || usage === 0xff)) {
        throw new InvalidFormatError();
      }

      var value;

      if (usage === 0x81) {
        value = reader.readBytes(reader.readUInt8());
      } else {
        value = reader.readVarBytesLE();
      }

      return new this({
        usage: usage,
        value: value
      });
    }
  }]);
  return BufferAttribute;
}(AttributeBase);

var ECPointAttribute =
/*#__PURE__*/
function (_AttributeBase) {
  _inherits(ECPointAttribute, _AttributeBase);

  function ECPointAttribute(_ref) {
    var _this;

    var usage = _ref.usage,
        value = _ref.value;
    _classCallCheck(this, ECPointAttribute);
    _this = _possibleConstructorReturn(this, (ECPointAttribute.__proto__ || Object.getPrototypeOf(ECPointAttribute)).call(this));
    Object.defineProperty(_this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(ECPointAttribute, function (other) {
        return _this.usage === other.usage && common.ecPointEqual(_this.value, other.value);
      })
    });
    _this.usage = usage;
    _this.value = value;
    _this.size = IOHelper.sizeOfUInt8 + IOHelper.sizeOfECPoint(_this.value);
    return _this;
  }

  _createClass(ECPointAttribute, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(ECPointAttribute.prototype.__proto__ || Object.getPrototypeOf(ECPointAttribute.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeBytes(common.ecPointToBuffer(this.value).slice(1));
    } // eslint-disable-next-line

  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        usage: toJSONAttributeUsage(this.usage),
        data: JSONHelper.writeECPoint(this.value)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;

      var _rollupPluginBabelHel = _get(ECPointAttribute.__proto__ || Object.getPrototypeOf(ECPointAttribute), "deserializeAttributeWireBase", this).call(this, options),
          usage = _rollupPluginBabelHel.usage;

      if (!(usage === 0x02 || usage === 0x03)) {
        throw new InvalidFormatError();
      }

      var value = common.bufferToECPoint(Buffer.concat([Buffer.from([usage]), reader.readBytes(32)]));
      return new this({
        usage: usage,
        value: value
      });
    }
  }]);
  return ECPointAttribute;
}(AttributeBase);

var UInt160Attribute =
/*#__PURE__*/
function (_AttributeBase) {
  _inherits(UInt160Attribute, _AttributeBase);

  function UInt160Attribute(_ref) {
    var _this;

    var usage = _ref.usage,
        value = _ref.value;
    _classCallCheck(this, UInt160Attribute);
    _this = _possibleConstructorReturn(this, (UInt160Attribute.__proto__ || Object.getPrototypeOf(UInt160Attribute)).call(this));
    Object.defineProperty(_this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(UInt160Attribute, function (other) {
        return _this.usage === other.usage && common.uInt160Equal(_this.value, other.value);
      })
    });
    _this.usage = usage;
    _this.value = value;
    _this.size = IOHelper.sizeOfUInt8 + IOHelper.sizeOfUInt160;
    return _this;
  }

  _createClass(UInt160Attribute, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(UInt160Attribute.prototype.__proto__ || Object.getPrototypeOf(UInt160Attribute.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeUInt160(this.value);
    } // eslint-disable-next-line

  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        usage: toJSONAttributeUsage(this.usage),
        data: JSONHelper.writeUInt160(this.value)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;

      var _rollupPluginBabelHel = _get(UInt160Attribute.__proto__ || Object.getPrototypeOf(UInt160Attribute), "deserializeAttributeWireBase", this).call(this, options),
          usage = _rollupPluginBabelHel.usage;

      if (usage !== 0x20) {
        throw new InvalidFormatError();
      }

      var value = reader.readUInt160();
      return new this({
        usage: usage,
        value: value
      });
    }
  }]);
  return UInt160Attribute;
}(AttributeBase);

var UInt256Attribute =
/*#__PURE__*/
function (_AttributeBase) {
  _inherits(UInt256Attribute, _AttributeBase);

  function UInt256Attribute(_ref) {
    var _this;

    var usage = _ref.usage,
        value = _ref.value;
    _classCallCheck(this, UInt256Attribute);
    _this = _possibleConstructorReturn(this, (UInt256Attribute.__proto__ || Object.getPrototypeOf(UInt256Attribute)).call(this));
    Object.defineProperty(_this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(UInt256Attribute, function (other) {
        return _this.usage === other.usage && common.uInt256Equal(_this.value, other.value);
      })
    });
    _this.usage = usage;
    _this.value = value;
    _this.size = IOHelper.sizeOfUInt8 + IOHelper.sizeOfUInt256;
    return _this;
  }

  _createClass(UInt256Attribute, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(UInt256Attribute.prototype.__proto__ || Object.getPrototypeOf(UInt256Attribute.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeUInt256(this.value);
    } // eslint-disable-next-line

  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        usage: toJSONAttributeUsage(this.usage),
        data: JSONHelper.writeUInt256(this.value)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;

      var _rollupPluginBabelHel = _get(UInt256Attribute.__proto__ || Object.getPrototypeOf(UInt256Attribute), "deserializeAttributeWireBase", this).call(this, options),
          usage = _rollupPluginBabelHel.usage;

      if (!(usage === 0x00 || usage === 0x30 || usage === 0xa1 || usage === 0xa2 || usage === 0xa3 || usage === 0xa4 || usage === 0xa5 || usage === 0xa6 || usage === 0xa7 || usage === 0xa8 || usage === 0xa9 || usage === 0xaa || usage === 0xab || usage === 0xac || usage === 0xad || usage === 0xae || usage === 0xaf)) {
        throw new InvalidFormatError();
      }

      var value = reader.readUInt256();
      return new this({
        usage: usage,
        value: value
      });
    }
  }]);
  return UInt256Attribute;
}(AttributeBase);

var deserializeWireBase = function deserializeWireBase(options) {
  var reader = options.reader;
  var usage = assertAttributeUsage(reader.clone().readUInt8());

  switch (usage) {
    case 0x20:
      return UInt160Attribute.deserializeWireBase(options);

    case 0x00:
    case 0x30:
    case 0xa1:
    case 0xa2:
    case 0xa3:
    case 0xa4:
    case 0xa5:
    case 0xa6:
    case 0xa7:
    case 0xa8:
    case 0xa9:
    case 0xaa:
    case 0xab:
    case 0xac:
    case 0xad:
    case 0xae:
    case 0xaf:
      return UInt256Attribute.deserializeWireBase(options);

    case 0x02:
    case 0x03:
      return ECPointAttribute.deserializeWireBase(options);

    case 0x81:
    case 0x90:
    case 0xf0:
    case 0xf1:
    case 0xf2:
    case 0xf3:
    case 0xf4:
    case 0xf5:
    case 0xf6:
    case 0xf7:
    case 0xf8:
    case 0xf9:
    case 0xfa:
    case 0xfb:
    case 0xfc:
    case 0xfd:
    case 0xfe:
    case 0xff:
      return BufferAttribute.deserializeWireBase(options);

    default:
      // eslint-disable-next-line
      
      throw new InvalidAttributeUsageError(usage);
  }
};

var _fixBabelExtend$9 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var TRANSACTION_TYPE = {
  MINER: 0x00,
  ISSUE: 0x01,
  CLAIM: 0x02,
  ENROLLMENT: 0x20,
  REGISTER: 0x40,
  CONTRACT: 0x80,
  PUBLISH: 0xd0,
  INVOCATION: 0xd1
};
var InvalidTransactionTypeError = _fixBabelExtend$9(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidTransactionTypeError, _Error);

  function InvalidTransactionTypeError(transactionType) {
    var _this;

    _classCallCheck(this, InvalidTransactionTypeError);
    _this = _possibleConstructorReturn(this, (InvalidTransactionTypeError.__proto__ || Object.getPrototypeOf(InvalidTransactionTypeError)).call(this, "Expected transaction type, found: ".concat(transactionType.toString(16))));
    _this.transactionType = transactionType;
    return _this;
  }

  return InvalidTransactionTypeError;
}(Error));
var assertTransactionType = function assertTransactionType(value) {
  switch (value) {
    case 0x00:
      // Miner
      return 0x00;

    case 0x01:
      // Issue
      return 0x01;

    case 0x02:
      // Claim
      return 0x02;

    case 0x20:
      // Enrollment
      return 0x20;

    case 0x40:
      // Register
      return 0x40;

    case 0x80:
      // Contract
      return 0x80;

    case 0xd0:
      // Publish
      return 0xd0;

    case 0xd1:
      // Invocation
      return 0xd1;

    default:
      throw new InvalidTransactionTypeError(value);
  }
};

var Input =
/*#__PURE__*/
function () {
  function Input(_ref) {
    var _this = this;

    var hash = _ref.hash,
        index = _ref.index;
    _classCallCheck(this, Input);
    Object.defineProperty(this, "size", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: IOHelper.sizeOfUInt256 + IOHelper.sizeOfUInt16LE
    });
    Object.defineProperty(this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(Input, function (other) {
        return common.uInt256Equal(_this.hash, other.hash) && other.index === _this.index;
      })
    });
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.hash = hash;
    this.index = index;
  }

  _createClass(Input, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt256(this.hash);
      writer.writeUInt16LE(this.index);
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        txid: JSONHelper.writeUInt256(this.hash),
        vout: this.index
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref2) {
      var reader = _ref2.reader;
      var hash = reader.readUInt256();
      var index = reader.readUInt16LE();
      return new this({
        hash: hash,
        index: index
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return Input;
}();

var Output =
/*#__PURE__*/
function () {
  function Output(_ref) {
    var _this = this;

    var asset = _ref.asset,
        value = _ref.value,
        address = _ref.address;
    _classCallCheck(this, Output);
    Object.defineProperty(this, "size", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: IOHelper.sizeOfUInt256 + IOHelper.sizeOfUInt16LE + IOHelper.sizeOfUInt160
    });
    Object.defineProperty(this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(Output, function (other) {
        return common.uInt256Equal(_this.asset, other.asset) && _this.value.eq(other.value) && common.uInt160Equal(_this.address, other.address);
      })
    });
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.asset = asset;
    this.value = value;
    this.address = address;
  }

  _createClass(Output, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt256(this.asset);
      writer.writeFixed8(this.value);
      writer.writeUInt160(this.address);
    }
  }, {
    key: "serializeJSON",
    // TODO: Maybe store index on Output directly? And add it when deserializing
    //       from transaction
    // eslint-disable-next-line
    value: function serializeJSON(context, index) {
      return {
        n: index,
        asset: JSONHelper.writeUInt256(this.asset),
        value: JSONHelper.writeFixed8(this.value),
        address: crypto$2.scriptHashToAddress({
          addressVersion: context.addressVersion,
          scriptHash: this.address
        })
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref2) {
      var reader = _ref2.reader;
      var asset = reader.readUInt256();
      var value = reader.readFixed8();
      var address = reader.readUInt160();
      return new this({
        asset: asset,
        value: value,
        address: address
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return Output;
}();

var hasDuplicateInputs = function hasDuplicateInputs(inputs) {
  for (var i = 1; i < inputs.length; i += 1) {
    for (var j = 0; j < i; j += 1) {
      if (inputs[i].equals(inputs[j])) {
        return true;
      }
    }
  }

  return false;
};
var hasIntersectingInputs = function hasIntersectingInputs(a, b) {
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = a[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var _inputA = _step.value;
      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        for (var _iterator2 = b[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          var _inputB = _step2.value;

          if (_inputA.equals(_inputB)) {
            return true;
          }
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return != null) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return != null) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  return false;
};

var getUtilityValue = function getUtilityValue(_ref) {
  var outputs = _ref.outputs,
      utilityToken = _ref.utilityToken;
  return outputs.filter(function (output) {
    return common.uInt256Equal(output.asset, utilityToken.hash);
  }).reduce(function (acc, output) {
    return acc.add(output.value);
  }, utils.ZERO);
};

var MAX_TRANSACTION_ATTRIBUTES = 16;

var TransactionBase =
/*#__PURE__*/
function () {
  function TransactionBase(_ref2) {
    var _this = this;

    var version = _ref2.version,
        type = _ref2.type,
        attributes = _ref2.attributes,
        inputs = _ref2.inputs,
        outputs = _ref2.outputs,
        scripts = _ref2.scripts,
        hash = _ref2.hash;
    _classCallCheck(this, TransactionBase);

    _initialiseProps.call(this);

    this.version = version == null ? this.constructor.VERSION : version;
    this.type = type;
    this.attributes = attributes || [];
    this.inputs = inputs || [];
    this.outputs = outputs || [];
    this.scripts = scripts || [];
    var hashIn = hash;
    this._hash = hashIn == null ? utils.lazy(function () {
      return crypto$2.hash256(_this.message);
    }) : function () {
      return hashIn;
    };

    if (this.attributes.length > MAX_TRANSACTION_ATTRIBUTES) {
      throw new InvalidFormatError();
    }
  }

  _createClass(TransactionBase, [{
    key: "clone",
    // eslint-disable-next-line
    value: function clone(scripts) {
      throw new Error('Not Implemented');
    }
  }, {
    key: "sign",
    value: function sign(key) {
      return this.clone(this.scripts.concat([crypto$2.createWitness(this.serializeUnsigned(), key)]));
    }
  }, {
    key: "serializeExclusiveBase",
    // eslint-disable-next-line
    value: function serializeExclusiveBase(writer) {}
  }, {
    key: "serializeUnsignedBase",
    value: function serializeUnsignedBase(writer) {
      writer.writeUInt8(this.type);
      writer.writeUInt8(this.version);
      this.serializeExclusiveBase(writer);
      writer.writeArray(this.attributes, function (attribute) {
        attribute.serializeWireBase(writer);
      });
      writer.writeArray(this.inputs, function (input) {
        input.serializeWireBase(writer);
      });
      writer.writeArray(this.outputs, function (output) {
        output.serializeWireBase(writer);
      });
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      this.serializeUnsignedBase(writer);
      writer.writeArray(this.scripts, function (script) {
        script.serializeWireBase(writer);
      });
    }
  }, {
    key: "serializeTransactionBaseJSON",
    value: function () {
      var _ref3 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee(context) {
        var networkFee;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return this.getNetworkFee(context.feeContext);

              case 2:
                networkFee = _context.sent;
                return _context.abrupt("return", {
                  txid: common.uInt256ToString(this.hashHex),
                  size: this.size,
                  version: this.version,
                  attributes: this.attributes.map(function (attribute) {
                    return attribute.serializeJSON(context);
                  }),
                  vin: this.inputs.map(function (input) {
                    return input.serializeJSON(context);
                  }),
                  vout: this.outputs.map(function (output, index) {
                    return output.serializeJSON(context, index);
                  }),
                  scripts: this.scripts.map(function (script) {
                    return script.serializeJSON(context);
                  }),
                  sys_fee: JSONHelper.writeFixed8(this.getSystemFee(context.feeContext)),
                  net_fee: JSONHelper.writeFixed8(networkFee)
                });

              case 4:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function serializeTransactionBaseJSON(_x) {
        return _ref3.apply(this, arguments);
      }

      return serializeTransactionBaseJSON;
    }() // eslint-disable-next-line

  }, {
    key: "serializeJSON",
    value: function () {
      var _ref4 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee2(context) {
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                throw new Error('Not Implemented');

              case 1:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function serializeJSON(_x2) {
        return _ref4.apply(this, arguments);
      }

      return serializeJSON;
    }()
  }, {
    key: "getNetworkFee",
    value: function () {
      var _ref5 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee3(context) {
        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                return _context3.abrupt("return", this._networkFee(context));

              case 1:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function getNetworkFee(_x3) {
        return _ref5.apply(this, arguments);
      }

      return getNetworkFee;
    }()
  }, {
    key: "getSystemFee",
    value: function getSystemFee(_ref6) {
      var fees = _ref6.fees;
      return fees[this.type] || utils.ZERO;
    }
  }, {
    key: "getReferences",
    value: function getReferences(options) {
      return this.__getReferences(options);
    }
  }, {
    key: "getTransactionResults",
    value: function getTransactionResults(options) {
      return this.__getTransactionResults(options);
    }
  }, {
    key: "getScriptHashesForVerifying",
    value: function () {
      var _ref7 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee4(options) {
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                return _context4.abrupt("return", this.__baseGetScriptHashesForVerifying(options));

              case 1:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function getScriptHashesForVerifying(_x4) {
        return _ref7.apply(this, arguments);
      }

      return getScriptHashesForVerifying;
    }()
  }, {
    key: "verify",
    value: function () {
      var _ref8 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee5(options) {
        var _this2 = this;

        var _options$memPool, memPool;

        return regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _options$memPool = options.memPool, memPool = _options$memPool === void 0 ? [] : _options$memPool;

                if (!hasDuplicateInputs(this.inputs)) {
                  _context5.next = 3;
                  break;
                }

                throw new VerifyError('Duplicate inputs');

              case 3:
                if (!memPool.some(function (tx) {
                  return !tx.equals(_this2) && tx.inputs.some(function (input) {
                    return _this2.inputs.some(function (thisInput) {
                      return input.equals(thisInput);
                    });
                  });
                })) {
                  _context5.next = 5;
                  break;
                }

                throw new VerifyError('Input already exists in mempool');

              case 5:
                if (!(this.attributes.filter(function (attribute) {
                  return attribute.usage === ATTRIBUTE_USAGE.ECDH02 || attribute.usage === ATTRIBUTE_USAGE.ECDH03;
                }).length > 1)) {
                  _context5.next = 7;
                  break;
                }

                throw new VerifyError('Too many ECDH attributes.');

              case 7:
                _context5.next = 9;
                return Promise.all([this._verifyDoubleSpend(options), this._verifyOutputs(options), this._verifyTransactionResults(options), this._verifyScripts(options)]);

              case 9:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function verify(_x5) {
        return _ref8.apply(this, arguments);
      }

      return verify;
    }()
  }, {
    key: "_verifyDoubleSpend",
    value: function () {
      var _ref10 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee6(_ref9) {
        var isSpent, isDoubleSpend;
        return regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                isSpent = _ref9.isSpent;
                _context6.next = 3;
                return Promise.all(this.inputs.map(function (input) {
                  return isSpent(input);
                }));

              case 3:
                isDoubleSpend = _context6.sent;

                if (!isDoubleSpend.some(function (value) {
                  return value;
                })) {
                  _context6.next = 6;
                  break;
                }

                throw new VerifyError('Transaction is a double spend');

              case 6:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function _verifyDoubleSpend(_x6) {
        return _ref10.apply(this, arguments);
      }

      return _verifyDoubleSpend;
    }()
  }, {
    key: "_verifyOutputs",
    value: function () {
      var _ref12 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee8(_ref11) {
        var getAsset, currentHeight, outputsGrouped, hasInvalidOutputs;
        return regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                getAsset = _ref11.getAsset, currentHeight = _ref11.currentHeight;
                outputsGrouped = utils.entries(_.groupBy(this.outputs, function (output) {
                  return common.uInt256ToHex(output.asset);
                }));
                _context8.next = 4;
                return Promise.all(outputsGrouped.map(function () {
                  var _ref14 = _asyncToGenerator(
                  /*#__PURE__*/
                  regeneratorRuntime.mark(function _callee7(_ref13) {
                    var _ref15, assetHex, outputs, asset;

                    return regeneratorRuntime.wrap(function _callee7$(_context7) {
                      while (1) {
                        switch (_context7.prev = _context7.next) {
                          case 0:
                            _ref15 = _slicedToArray(_ref13, 2), assetHex = _ref15[0], outputs = _ref15[1];
                            _context7.next = 3;
                            return getAsset({
                              hash: common.hexToUInt256(assetHex)
                            });

                          case 3:
                            asset = _context7.sent;

                            if (!(asset.expiration <= currentHeight + 1 && asset.type !== ASSET_TYPE.GOVERNING_TOKEN && asset.type !== ASSET_TYPE.UTILITY_TOKEN)) {
                              _context7.next = 6;
                              break;
                            }

                            return _context7.abrupt("return", true);

                          case 6:
                            return _context7.abrupt("return", outputs.some(function (output) {
                              return !output.value.mod(utils.TEN.pow(utils.EIGHT.subn(asset.precision))).eq(utils.ZERO);
                            }));

                          case 7:
                          case "end":
                            return _context7.stop();
                        }
                      }
                    }, _callee7, this);
                  }));

                  return function (_x8) {
                    return _ref14.apply(this, arguments);
                  };
                }()));

              case 4:
                hasInvalidOutputs = _context8.sent;

                if (!hasInvalidOutputs.some(function (value) {
                  return value;
                })) {
                  _context8.next = 7;
                  break;
                }

                throw new VerifyError('Transaction has invalid output');

              case 7:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee8, this);
      }));

      function _verifyOutputs(_x7) {
        return _ref12.apply(this, arguments);
      }

      return _verifyOutputs;
    }()
  }, {
    key: "_verifyTransactionResults",
    value: function () {
      var _ref17 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee9(_ref16) {
        var getOutput, utilityToken, governingToken, fees, results, resultsDestroy, feeContext, systemFee, resultsIssue;
        return regeneratorRuntime.wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                getOutput = _ref16.getOutput, utilityToken = _ref16.utilityToken, governingToken = _ref16.governingToken, fees = _ref16.fees;
                _context9.next = 3;
                return this.getTransactionResults({
                  getOutput: getOutput
                });

              case 3:
                results = _context9.sent;
                resultsDestroy = utils.entries(results).filter( // eslint-disable-next-line
                function (_ref18) {
                  var _ref19 = _slicedToArray(_ref18, 2),
                      _$$1 = _ref19[0],
                      value = _ref19[1];

                  return value.gt(utils.ZERO);
                });

                if (!(resultsDestroy.length > 1 || resultsDestroy.length === 1 && !common.uInt256Equal(common.hexToUInt256(resultsDestroy[0][0]), utilityToken.hash))) {
                  _context9.next = 7;
                  break;
                }

                throw new VerifyError('Invalid destroyed output.');

              case 7:
                feeContext = {
                  getOutput: getOutput,
                  governingToken: governingToken,
                  utilityToken: utilityToken,
                  fees: fees
                };
                systemFee = this.getSystemFee(feeContext);

                if (!(systemFee.gt(utils.ZERO) && (resultsDestroy.length === 0 || resultsDestroy[0][1].lt(systemFee)))) {
                  _context9.next = 11;
                  break;
                }

                throw new VerifyError('Not enough output value for system fee.');

              case 11:
                resultsIssue = utils.entries(results).filter( // eslint-disable-next-line
                function (_ref20) {
                  var _ref21 = _slicedToArray(_ref20, 2),
                      _$$1 = _ref21[0],
                      value = _ref21[1];

                  return value.lt(utils.ZERO);
                });
                _context9.t0 = this.type;
                _context9.next = _context9.t0 === TRANSACTION_TYPE.MINER ? 15 : _context9.t0 === TRANSACTION_TYPE.CLAIM ? 15 : _context9.t0 === TRANSACTION_TYPE.ISSUE ? 18 : 21;
                break;

              case 15:
                if (!resultsIssue.some( // eslint-disable-next-line
                function (_ref22) {
                  var _ref23 = _slicedToArray(_ref22, 2),
                      assetHex = _ref23[0],
                      _$$1 = _ref23[1];

                  return !common.uInt256Equal(common.hexToUInt256(assetHex), utilityToken.hash);
                })) {
                  _context9.next = 17;
                  break;
                }

                throw new VerifyError('Invalid miner/claim result');

              case 17:
                return _context9.abrupt("break", 24);

              case 18:
                if (! // eslint-disable-next-line
                resultsIssue.some(function (_ref24) {
                  var _ref25 = _slicedToArray(_ref24, 2),
                      assetHex = _ref25[0],
                      _$$1 = _ref25[1];

                  return common.uInt256Equal(common.hexToUInt256(assetHex), utilityToken.hash);
                })) {
                  _context9.next = 20;
                  break;
                }

                throw new VerifyError('Invalid issue result');

              case 20:
                return _context9.abrupt("break", 24);

              case 21:
                if (!(resultsIssue.length > 0)) {
                  _context9.next = 23;
                  break;
                }

                throw new VerifyError('Invalid results.');

              case 23:
                return _context9.abrupt("break", 24);

              case 24:
              case "end":
                return _context9.stop();
            }
          }
        }, _callee9, this);
      }));

      function _verifyTransactionResults(_x9) {
        return _ref17.apply(this, arguments);
      }

      return _verifyTransactionResults;
    }()
  }, {
    key: "_verifyScripts",
    value: function () {
      var _ref27 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee10(_ref26) {
        var _this3 = this;

        var getAsset, getOutput, verifyScript, hashesSet, hashes;
        return regeneratorRuntime.wrap(function _callee10$(_context10) {
          while (1) {
            switch (_context10.prev = _context10.next) {
              case 0:
                getAsset = _ref26.getAsset, getOutput = _ref26.getOutput, verifyScript = _ref26.verifyScript;
                _context10.next = 3;
                return this.getScriptHashesForVerifying({
                  getAsset: getAsset,
                  getOutput: getOutput
                });

              case 3:
                hashesSet = _context10.sent;

                if (!(hashesSet.size !== this.scripts.length)) {
                  _context10.next = 6;
                  break;
                }

                throw new VerifyError('Invalid witnesses');

              case 6:
                hashes = [].concat(_toConsumableArray(hashesSet)).sort().map(function (value) {
                  return common.hexToUInt160(value);
                });
                _context10.next = 9;
                return Promise.all(_.zip(hashes, this.scripts).map(function (_ref28) {
                  var _ref29 = _slicedToArray(_ref28, 2),
                      hash = _ref29[0],
                      witness = _ref29[1];

                  return verifyScript({
                    scriptContainer: {
                      type: SCRIPT_CONTAINER_TYPE.TRANSACTION,
                      value: _this3
                    },
                    hash: hash,
                    witness: witness
                  });
                }));

              case 9:
              case "end":
                return _context10.stop();
            }
          }
        }, _callee10, this);
      }));

      function _verifyScripts(_x10) {
        return _ref27.apply(this, arguments);
      }

      return _verifyScripts;
    }()
  }, {
    key: "hash",
    get: function get() {
      return this._hash();
    }
  }, {
    key: "hashHex",
    get: function get() {
      return this._hashHex();
    }
  }, {
    key: "size",
    get: function get() {
      return this._size();
    }
  }, {
    key: "message",
    get: function get() {
      return this._message();
    }
  }], [{
    key: "deserializeTransactionBaseStartWireBase",
    value: function deserializeTransactionBaseStartWireBase(_ref30) {
      var reader = _ref30.reader;
      var type = assertTransactionType(reader.readUInt8());
      var version = reader.readUInt8();
      return {
        type: type,
        version: version
      };
    }
  }, {
    key: "deserializeTransactionBaseEndWireBase",
    value: function deserializeTransactionBaseEndWireBase(options) {
      var reader = options.reader;
      var attributes = reader.readArray(function () {
        return deserializeWireBase(options);
      }, MAX_TRANSACTION_ATTRIBUTES);
      var inputs = reader.readArray(function () {
        return Input.deserializeWireBase(options);
      });
      var outputs = reader.readArray(function () {
        return Output.deserializeWireBase(options);
      }, utils.USHORT_MAX_NUMBER + 1);
      var scripts = reader.readArray(function () {
        return Witness.deserializeWireBase(options);
      });
      return {
        attributes: attributes,
        inputs: inputs,
        outputs: outputs,
        scripts: scripts
      };
    } // eslint-disable-next-line

  }, {
    key: "deserializeWireBase",
    value: function deserializeWireBase$$1(options) {
      throw new Error('Not Implemented');
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return TransactionBase;
}();

Object.defineProperty(TransactionBase, "VERSION", {
  configurable: true,
  enumerable: true,
  writable: true,
  value: 0
});

var _initialiseProps = function _initialiseProps() {
  var _this4 = this;

  Object.defineProperty(this, "_hashHex", {
    configurable: true,
    enumerable: true,
    writable: true,
    value: utils.lazy(function () {
      return common.uInt256ToHex(_this4.hash);
    })
  });
  Object.defineProperty(this, "_size", {
    configurable: true,
    enumerable: true,
    writable: true,
    value: utils.lazy(function () {
      return IOHelper.sizeOfUInt8 + IOHelper.sizeOfArray(_this4.attributes, function (attribute) {
        return attribute.size;
      }) + IOHelper.sizeOfArray(_this4.inputs, function (input) {
        return input.size;
      }) + IOHelper.sizeOfArray(_this4.outputs, function (output) {
        return output.size;
      }) + IOHelper.sizeOfArray(_this4.scripts, function (script) {
        return script.size;
      });
    })
  });
  Object.defineProperty(this, "_message", {
    configurable: true,
    enumerable: true,
    writable: true,
    value: utils.lazy(function () {
      return _this4.serializeUnsigned();
    })
  });
  Object.defineProperty(this, "equals", {
    configurable: true,
    enumerable: true,
    writable: true,
    value: utils.equals(this.constructor, function (other) {
      return common.uInt256Equal(_this4.hash, other.hash);
    })
  });
  Object.defineProperty(this, "serializeWire", {
    configurable: true,
    enumerable: true,
    writable: true,
    value: createSerializeWire(this.serializeWireBase.bind(this))
  });
  Object.defineProperty(this, "serializeUnsigned", {
    configurable: true,
    enumerable: true,
    writable: true,
    value: createSerializeWire(this.serializeUnsignedBase.bind(this))
  });
  Object.defineProperty(this, "_networkFee", {
    configurable: true,
    enumerable: true,
    writable: true,
    value: utils.lazyAsync(function () {
      var _ref31 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee11(context) {
        var getOutput, utilityToken, outputsForInputs, inputValue, outputValue;
        return regeneratorRuntime.wrap(function _callee11$(_context11) {
          while (1) {
            switch (_context11.prev = _context11.next) {
              case 0:
                getOutput = context.getOutput, utilityToken = context.utilityToken;
                _context11.next = 3;
                return Promise.all(_this4.inputs.map(function (input) {
                  return getOutput(input);
                }));

              case 3:
                outputsForInputs = _context11.sent;
                inputValue = getUtilityValue({
                  outputs: outputsForInputs,
                  utilityToken: utilityToken
                });
                outputValue = getUtilityValue({
                  outputs: _this4.outputs,
                  utilityToken: utilityToken
                });
                return _context11.abrupt("return", inputValue.sub(outputValue).sub(_this4.getSystemFee(context)));

              case 7:
              case "end":
                return _context11.stop();
            }
          }
        }, _callee11, this);
      }));

      return function (_x11) {
        return _ref31.apply(this, arguments);
      };
    }())
  });
  Object.defineProperty(this, "__getReferences", {
    configurable: true,
    enumerable: true,
    writable: true,
    value: utils.lazyAsync(function () {
      var _ref33 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee12(_ref32) {
        var getOutput;
        return regeneratorRuntime.wrap(function _callee12$(_context12) {
          while (1) {
            switch (_context12.prev = _context12.next) {
              case 0:
                getOutput = _ref32.getOutput;
                return _context12.abrupt("return", Promise.all(_this4.inputs.map(function (input) {
                  return getOutput(input);
                })));

              case 2:
              case "end":
                return _context12.stop();
            }
          }
        }, _callee12, this);
      }));

      return function (_x12) {
        return _ref33.apply(this, arguments);
      };
    }())
  });
  Object.defineProperty(this, "__getTransactionResults", {
    configurable: true,
    enumerable: true,
    writable: true,
    value: utils.lazyAsync(function () {
      var _ref35 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee13(_ref34) {
        var getOutput, inputOutputs, results, addOutputs;
        return regeneratorRuntime.wrap(function _callee13$(_context13) {
          while (1) {
            switch (_context13.prev = _context13.next) {
              case 0:
                getOutput = _ref34.getOutput;
                _context13.next = 3;
                return _this4.getReferences({
                  getOutput: getOutput
                });

              case 3:
                inputOutputs = _context13.sent;
                results = {};

                addOutputs = function addOutputs(outputs, negative) {
                  var _iteratorNormalCompletion = true;
                  var _didIteratorError = false;
                  var _iteratorError = undefined;

                  try {
                    for (var _iterator = outputs[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                      var _output = _step.value;

                      var _key2 = common.uInt256ToHex(_output.asset);

                      if (results[_key2] == null) {
                        results[_key2] = utils.ZERO;
                      }

                      results[_key2] = results[_key2].add(negative ? _output.value.neg() : _output.value);
                    }
                  } catch (err) {
                    _didIteratorError = true;
                    _iteratorError = err;
                  } finally {
                    try {
                      if (!_iteratorNormalCompletion && _iterator.return != null) {
                        _iterator.return();
                      }
                    } finally {
                      if (_didIteratorError) {
                        throw _iteratorError;
                      }
                    }
                  }
                };

                addOutputs(inputOutputs);
                addOutputs(_this4.outputs, true);
                return _context13.abrupt("return", _.pickBy(results, function (value) {
                  return !value.eq(utils.ZERO);
                }));

              case 9:
              case "end":
                return _context13.stop();
            }
          }
        }, _callee13, this);
      }));

      return function (_x13) {
        return _ref35.apply(this, arguments);
      };
    }())
  });
  Object.defineProperty(this, "__baseGetScriptHashesForVerifying", {
    configurable: true,
    enumerable: true,
    writable: true,
    value: utils.lazyAsync(function () {
      var _ref37 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee16(_ref36) {
        var getOutput, getAsset, _ref38, _ref39, inputHashes, outputHashes, attributeHashes;

        return regeneratorRuntime.wrap(function _callee16$(_context16) {
          while (1) {
            switch (_context16.prev = _context16.next) {
              case 0:
                getOutput = _ref36.getOutput, getAsset = _ref36.getAsset;
                _context16.next = 3;
                return Promise.all([Promise.all(_this4.inputs.map(function () {
                  var _ref40 = _asyncToGenerator(
                  /*#__PURE__*/
                  regeneratorRuntime.mark(function _callee14(input) {
                    var output;
                    return regeneratorRuntime.wrap(function _callee14$(_context14) {
                      while (1) {
                        switch (_context14.prev = _context14.next) {
                          case 0:
                            _context14.next = 2;
                            return getOutput(input);

                          case 2:
                            output = _context14.sent;
                            return _context14.abrupt("return", common.uInt160ToHex(output.address));

                          case 4:
                          case "end":
                            return _context14.stop();
                        }
                      }
                    }, _callee14, this);
                  }));

                  return function (_x15) {
                    return _ref40.apply(this, arguments);
                  };
                }())), Promise.all(_this4.outputs.map(function () {
                  var _ref41 = _asyncToGenerator(
                  /*#__PURE__*/
                  regeneratorRuntime.mark(function _callee15(output) {
                    var asset;
                    return regeneratorRuntime.wrap(function _callee15$(_context15) {
                      while (1) {
                        switch (_context15.prev = _context15.next) {
                          case 0:
                            _context15.next = 2;
                            return getAsset({
                              hash: output.asset
                            });

                          case 2:
                            asset = _context15.sent;

                            if (!hasFlag(asset.type, ASSET_TYPE.DUTY_FLAG)) {
                              _context15.next = 5;
                              break;
                            }

                            return _context15.abrupt("return", common.uInt160ToHex(output.address));

                          case 5:
                            return _context15.abrupt("return", null);

                          case 6:
                          case "end":
                            return _context15.stop();
                        }
                      }
                    }, _callee15, this);
                  }));

                  return function (_x16) {
                    return _ref41.apply(this, arguments);
                  };
                }())).then(function (hashes) {
                  return hashes.filter(Boolean);
                })]);

              case 3:
                _ref38 = _context16.sent;
                _ref39 = _slicedToArray(_ref38, 2);
                inputHashes = _ref39[0];
                outputHashes = _ref39[1];
                attributeHashes = _this4.attributes.map(function (attribute) {
                  return attribute instanceof UInt160Attribute && attribute.usage === ATTRIBUTE_USAGE.SCRIPT ? common.uInt160ToHex(attribute.value) : null;
                }).filter(Boolean);
                return _context16.abrupt("return", new Set([].concat(_toConsumableArray(inputHashes), _toConsumableArray(outputHashes), _toConsumableArray(attributeHashes))));

              case 9:
              case "end":
                return _context16.stop();
            }
          }
        }, _callee16, this);
      }));

      return function (_x14) {
        return _ref37.apply(this, arguments);
      };
    }())
  });
};

var MinerTransaction =
/*#__PURE__*/
function (_TransactionBase) {
  _inherits(MinerTransaction, _TransactionBase);

  function MinerTransaction(_ref) {
    var _this;

    var version = _ref.version,
        attributes = _ref.attributes,
        inputs = _ref.inputs,
        outputs = _ref.outputs,
        scripts = _ref.scripts,
        hash = _ref.hash,
        nonce = _ref.nonce;
    _classCallCheck(this, MinerTransaction);
    _this = _possibleConstructorReturn(this, (MinerTransaction.__proto__ || Object.getPrototypeOf(MinerTransaction)).call(this, {
      version: version,
      type: TRANSACTION_TYPE.MINER,
      attributes: attributes,
      inputs: inputs,
      outputs: outputs,
      scripts: scripts,
      hash: hash
    }));
    _this.nonce = nonce;

    if (_this.version !== 0) {
      throw new InvalidFormatError();
    }

    _this.__size = utils.lazy(function () {
      return _get(MinerTransaction.prototype.__proto__ || Object.getPrototypeOf(MinerTransaction.prototype), "size", _this) + IOHelper.sizeOfUInt8 + IOHelper.sizeOfUInt32LE;
    });
    return _this;
  }

  _createClass(MinerTransaction, [{
    key: "clone",
    value: function clone(scripts) {
      return new this.constructor({
        version: this.version,
        attributes: this.attributes,
        inputs: this.inputs,
        outputs: this.outputs,
        scripts: scripts,
        nonce: this.nonce
      });
    }
  }, {
    key: "serializeExclusiveBase",
    value: function serializeExclusiveBase(writer) {
      writer.writeUInt32LE(this.nonce);
    } // eslint-disable-next-line

  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function () {
      var _ref2 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee(context) {
        var transactionBaseJSON;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return _get(MinerTransaction.prototype.__proto__ || Object.getPrototypeOf(MinerTransaction.prototype), "serializeTransactionBaseJSON", this).call(this, context);

              case 2:
                transactionBaseJSON = _context.sent;
                return _context.abrupt("return", {
                  type: 'MinerTransaction',
                  txid: transactionBaseJSON.txid,
                  size: transactionBaseJSON.size,
                  version: transactionBaseJSON.version,
                  attributes: transactionBaseJSON.attributes,
                  vin: transactionBaseJSON.vin,
                  vout: transactionBaseJSON.vout,
                  scripts: transactionBaseJSON.scripts,
                  sys_fee: transactionBaseJSON.sys_fee,
                  net_fee: transactionBaseJSON.net_fee,
                  nonce: this.nonce
                });

              case 4:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function serializeJSON(_x) {
        return _ref2.apply(this, arguments);
      }

      return serializeJSON;
    }() // eslint-disable-next-line

  }, {
    key: "getNetworkFee",
    value: function () {
      var _ref3 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee2(context) {
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                return _context2.abrupt("return", utils.ZERO);

              case 1:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function getNetworkFee(_x2) {
        return _ref3.apply(this, arguments);
      }

      return getNetworkFee;
    }()
  }, {
    key: "verify",
    value: function () {
      var _ref4 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee3(options) {
        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                _context3.next = 2;
                return Promise.all([_get(MinerTransaction.prototype.__proto__ || Object.getPrototypeOf(MinerTransaction.prototype), "verify", this).call(this, options), this._verify(options)]);

              case 2:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function verify(_x3) {
        return _ref4.apply(this, arguments);
      }

      return verify;
    }()
  }, {
    key: "_verify",
    value: function () {
      var _ref5 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee4(options) {
        var getOutput, utilityToken, results, resultsIssue;
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                getOutput = options.getOutput, utilityToken = options.utilityToken;
                _context4.next = 3;
                return this.getTransactionResults({
                  getOutput: getOutput
                });

              case 3:
                results = _context4.sent;
                resultsIssue = utils.entries(results).filter( // eslint-disable-next-line
                function (_ref6) {
                  var _ref7 = _slicedToArray(_ref6, 2),
                      _$$1 = _ref7[0],
                      value = _ref7[1];

                  return value.lt(utils.ZERO);
                }); // eslint-disable-next-line

                if (!resultsIssue.some(function (_ref8) {
                  var _ref9 = _slicedToArray(_ref8, 2),
                      assetHex = _ref9[0],
                      _$$1 = _ref9[1];

                  return !common.uInt256Equal(common.hexToUInt256(assetHex), utilityToken.hash);
                })) {
                  _context4.next = 7;
                  break;
                }

                throw new VerifyError('Invalid miner result');

              case 7:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function _verify(_x4) {
        return _ref5.apply(this, arguments);
      }

      return _verify;
    }()
  }, {
    key: "size",
    get: function get() {
      return this.__size();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;

      var _rollupPluginBabelHel = _get(MinerTransaction.__proto__ || Object.getPrototypeOf(MinerTransaction), "deserializeTransactionBaseStartWireBase", this).call(this, options),
          type = _rollupPluginBabelHel.type,
          version = _rollupPluginBabelHel.version;

      if (type !== TRANSACTION_TYPE.MINER) {
        throw new InvalidFormatError();
      }

      var nonce = reader.readUInt32LE();

      var _rollupPluginBabelHel2 = _get(MinerTransaction.__proto__ || Object.getPrototypeOf(MinerTransaction), "deserializeTransactionBaseEndWireBase", this).call(this, options),
          attributes = _rollupPluginBabelHel2.attributes,
          inputs = _rollupPluginBabelHel2.inputs,
          outputs = _rollupPluginBabelHel2.outputs,
          scripts = _rollupPluginBabelHel2.scripts;

      return new this({
        version: version,
        attributes: attributes,
        inputs: inputs,
        outputs: outputs,
        scripts: scripts,
        nonce: nonce
      });
    }
  }]);
  return MinerTransaction;
}(TransactionBase);

var IssueTransaction =
/*#__PURE__*/
function (_TransactionBase) {
  _inherits(IssueTransaction, _TransactionBase);

  // TODO: How is version set?
  // static VERSION = 1;
  function IssueTransaction(_ref) {
    var _this;

    var version = _ref.version,
        attributes = _ref.attributes,
        inputs = _ref.inputs,
        outputs = _ref.outputs,
        scripts = _ref.scripts,
        hash = _ref.hash;
    _classCallCheck(this, IssueTransaction);
    _this = _possibleConstructorReturn(this, (IssueTransaction.__proto__ || Object.getPrototypeOf(IssueTransaction)).call(this, {
      version: version,
      type: TRANSACTION_TYPE.ISSUE,
      attributes: attributes,
      inputs: inputs,
      outputs: outputs,
      scripts: scripts,
      hash: hash
    }));

    if (_this.version > 1) {
      throw new InvalidFormatError();
    }

    _this.__size = utils.lazy(function () {
      return _get(IssueTransaction.prototype.__proto__ || Object.getPrototypeOf(IssueTransaction.prototype), "size", _this) + IOHelper.sizeOfUInt8;
    });
    _this.__issueGetScriptHashesForVerifying = utils.lazyAsync(function () {
      var _ref2 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee2(options) {
        var getOutput, getAsset, _ref3, _ref4, hashes, issuerHashes;

        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                getOutput = options.getOutput, getAsset = options.getAsset;
                _context2.next = 3;
                return Promise.all([_get(IssueTransaction.prototype.__proto__ || Object.getPrototypeOf(IssueTransaction.prototype), "getScriptHashesForVerifying", _this).call(_this, options), _this.getTransactionResults({
                  getOutput: getOutput
                }).then(function (results) {
                  return Promise.all(utils.entries(results) // eslint-disable-next-line
                  .filter(function (_ref5) {
                    var _ref6 = _slicedToArray(_ref5, 2),
                        _$$1 = _ref6[0],
                        value = _ref6[1];

                    return value.lt(utils.ZERO);
                  }) // eslint-disable-next-line
                  .map(function () {
                    var _ref8 = _asyncToGenerator(
                    /*#__PURE__*/
                    regeneratorRuntime.mark(function _callee(_ref7) {
                      var _ref9, hash, _$$1, asset;

                      return regeneratorRuntime.wrap(function _callee$(_context) {
                        while (1) {
                          switch (_context.prev = _context.next) {
                            case 0:
                              _ref9 = _slicedToArray(_ref7, 2), hash = _ref9[0], _$$1 = _ref9[1];
                              _context.next = 3;
                              return getAsset({
                                hash: common.hexToUInt256(hash)
                              });

                            case 3:
                              asset = _context.sent;
                              return _context.abrupt("return", common.uInt160ToHex(asset.issuer));

                            case 5:
                            case "end":
                              return _context.stop();
                          }
                        }
                      }, _callee, this);
                    }));

                    return function (_x2) {
                      return _ref8.apply(this, arguments);
                    };
                  }()));
                })]);

              case 3:
                _ref3 = _context2.sent;
                _ref4 = _slicedToArray(_ref3, 2);
                hashes = _ref4[0];
                issuerHashes = _ref4[1];
                return _context2.abrupt("return", new Set([].concat(_toConsumableArray(hashes), _toConsumableArray(issuerHashes))));

              case 8:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      return function (_x) {
        return _ref2.apply(this, arguments);
      };
    }());
    return _this;
  }

  _createClass(IssueTransaction, [{
    key: "clone",
    value: function clone(scripts) {
      return new this.constructor({
        version: this.version,
        attributes: this.attributes,
        inputs: this.inputs,
        outputs: this.outputs,
        scripts: scripts
      });
    } // eslint-disable-next-line

  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function () {
      var _ref10 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee3(context) {
        var transactionBaseJSON;
        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                _context3.next = 2;
                return _get(IssueTransaction.prototype.__proto__ || Object.getPrototypeOf(IssueTransaction.prototype), "serializeTransactionBaseJSON", this).call(this, context);

              case 2:
                transactionBaseJSON = _context3.sent;
                return _context3.abrupt("return", {
                  type: 'IssueTransaction',
                  txid: transactionBaseJSON.txid,
                  size: transactionBaseJSON.size,
                  version: transactionBaseJSON.version,
                  attributes: transactionBaseJSON.attributes,
                  vin: transactionBaseJSON.vin,
                  vout: transactionBaseJSON.vout,
                  scripts: transactionBaseJSON.scripts,
                  sys_fee: transactionBaseJSON.sys_fee,
                  net_fee: transactionBaseJSON.net_fee
                });

              case 4:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function serializeJSON(_x3) {
        return _ref10.apply(this, arguments);
      }

      return serializeJSON;
    }()
  }, {
    key: "getSystemFee",
    value: function getSystemFee(context) {
      if (this.version >= 1) {
        return utils.ZERO;
      }

      var governingToken = context.governingToken,
          utilityToken = context.utilityToken;

      if (this.outputs.every(function (output) {
        return common.uInt256Equal(output.asset, governingToken.hash) || common.uInt256Equal(output.asset, utilityToken.hash);
      })) {
        return utils.ZERO;
      }

      return _get(IssueTransaction.prototype.__proto__ || Object.getPrototypeOf(IssueTransaction.prototype), "getSystemFee", this).call(this, context);
    }
  }, {
    key: "getScriptHashesForVerifying",
    value: function () {
      var _ref11 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee4(options) {
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                return _context4.abrupt("return", this.__issueGetScriptHashesForVerifying(options));

              case 1:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function getScriptHashesForVerifying(_x4) {
        return _ref11.apply(this, arguments);
      }

      return getScriptHashesForVerifying;
    }()
  }, {
    key: "verify",
    value: function () {
      var _ref12 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee5(options) {
        return regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _context5.next = 2;
                return Promise.all([_get(IssueTransaction.prototype.__proto__ || Object.getPrototypeOf(IssueTransaction.prototype), "verify", this).call(this, options), this._verify(options)]);

              case 2:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function verify(_x5) {
        return _ref12.apply(this, arguments);
      }

      return verify;
    }()
  }, {
    key: "_verify",
    value: function () {
      var _ref13 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee7(options) {
        var _this2 = this;

        var getOutput, getAsset, _options$memPool, memPool, results;

        return regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                getOutput = options.getOutput, getAsset = options.getAsset, _options$memPool = options.memPool, memPool = _options$memPool === void 0 ? [] : _options$memPool;
                _context7.next = 3;
                return this.getTransactionResults({
                  getOutput: getOutput
                });

              case 3:
                results = _context7.sent;
                _context7.next = 6;
                return Promise.all(utils.entries(results).map(function () {
                  var _ref15 = _asyncToGenerator(
                  /*#__PURE__*/
                  regeneratorRuntime.mark(function _callee6(_ref14) {
                    var _ref16, assetHex, value, hash, asset, issued;

                    return regeneratorRuntime.wrap(function _callee6$(_context6) {
                      while (1) {
                        switch (_context6.prev = _context6.next) {
                          case 0:
                            _ref16 = _slicedToArray(_ref14, 2), assetHex = _ref16[0], value = _ref16[1];
                            hash = common.hexToUInt256(assetHex);
                            _context6.next = 4;
                            return getAsset({
                              hash: hash
                            });

                          case 4:
                            asset = _context6.sent;

                            if (!asset.amount.lt(utils.ZERO)) {
                              _context6.next = 7;
                              break;
                            }

                            return _context6.abrupt("return");

                          case 7:
                            issued = asset.available.add(memPool.filter(function (transaction) {
                              return transaction !== _this2;
                            }).reduce(function (acc, transaction) {
                              return transaction.outputs.filter(function (output) {
                                return common.uInt256Equal(hash, output.asset);
                              }).reduce(function (innerAcc, output) {
                                return innerAcc.add(output.value);
                              }, acc);
                            }, utils.ZERO));

                            if (!asset.amount.sub(issued).lt(value.neg())) {
                              _context6.next = 10;
                              break;
                            }

                            throw new VerifyError('Invalid issue amount');

                          case 10:
                          case "end":
                            return _context6.stop();
                        }
                      }
                    }, _callee6, this);
                  }));

                  return function (_x7) {
                    return _ref15.apply(this, arguments);
                  };
                }()));

              case 6:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function _verify(_x6) {
        return _ref13.apply(this, arguments);
      }

      return _verify;
    }()
  }, {
    key: "size",
    get: function get() {
      return this.__size();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var _rollupPluginBabelHel = _get(IssueTransaction.__proto__ || Object.getPrototypeOf(IssueTransaction), "deserializeTransactionBaseStartWireBase", this).call(this, options),
          type = _rollupPluginBabelHel.type,
          version = _rollupPluginBabelHel.version;

      if (type !== TRANSACTION_TYPE.ISSUE) {
        throw new InvalidFormatError();
      }

      var _rollupPluginBabelHel2 = _get(IssueTransaction.__proto__ || Object.getPrototypeOf(IssueTransaction), "deserializeTransactionBaseEndWireBase", this).call(this, options),
          attributes = _rollupPluginBabelHel2.attributes,
          inputs = _rollupPluginBabelHel2.inputs,
          outputs = _rollupPluginBabelHel2.outputs,
          scripts = _rollupPluginBabelHel2.scripts;

      return new this({
        version: version,
        attributes: attributes,
        inputs: inputs,
        outputs: outputs,
        scripts: scripts
      });
    }
  }]);
  return IssueTransaction;
}(TransactionBase);

var ClaimTransaction =
/*#__PURE__*/
function (_TransactionBase) {
  _inherits(ClaimTransaction, _TransactionBase);

  function ClaimTransaction(_ref) {
    var _this;

    var version = _ref.version,
        attributes = _ref.attributes,
        inputs = _ref.inputs,
        outputs = _ref.outputs,
        scripts = _ref.scripts,
        hash = _ref.hash,
        claims = _ref.claims;
    _classCallCheck(this, ClaimTransaction);
    _this = _possibleConstructorReturn(this, (ClaimTransaction.__proto__ || Object.getPrototypeOf(ClaimTransaction)).call(this, {
      version: version,
      type: TRANSACTION_TYPE.CLAIM,
      attributes: attributes,
      inputs: inputs,
      outputs: outputs,
      scripts: scripts,
      hash: hash
    }));
    _this.claims = claims;

    if (_this.version !== 0) {
      throw new InvalidFormatError();
    }

    if (_this.claims.length === 0) {
      throw new InvalidFormatError();
    }

    _this.__size = utils.lazy(function () {
      return _get(ClaimTransaction.prototype.__proto__ || Object.getPrototypeOf(ClaimTransaction.prototype), "size", _this) + IOHelper.sizeOfUInt8 + IOHelper.sizeOfArray(_this.claims, function (claim) {
        return claim.size;
      });
    });
    _this.__claimGetScriptHashesForVerifying = utils.lazyAsync(function () {
      var _ref2 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee2(options) {
        var getOutput, _ref3, _ref4, hashesSet, hashes;

        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                getOutput = options.getOutput;
                _context2.next = 3;
                return Promise.all([_get(ClaimTransaction.prototype.__proto__ || Object.getPrototypeOf(ClaimTransaction.prototype), "getScriptHashesForVerifying", _this).call(_this, options), Promise.all(_this.claims.map(function () {
                  var _ref5 = _asyncToGenerator(
                  /*#__PURE__*/
                  regeneratorRuntime.mark(function _callee(claim) {
                    var output;
                    return regeneratorRuntime.wrap(function _callee$(_context) {
                      while (1) {
                        switch (_context.prev = _context.next) {
                          case 0:
                            _context.next = 2;
                            return getOutput(claim);

                          case 2:
                            output = _context.sent;
                            return _context.abrupt("return", common.uInt160ToHex(output.address));

                          case 4:
                          case "end":
                            return _context.stop();
                        }
                      }
                    }, _callee, this);
                  }));

                  return function (_x2) {
                    return _ref5.apply(this, arguments);
                  };
                }()))]);

              case 3:
                _ref3 = _context2.sent;
                _ref4 = _slicedToArray(_ref3, 2);
                hashesSet = _ref4[0];
                hashes = _ref4[1];
                return _context2.abrupt("return", new Set([].concat(_toConsumableArray(hashesSet), _toConsumableArray(hashes))));

              case 8:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      return function (_x) {
        return _ref2.apply(this, arguments);
      };
    }());
    return _this;
  }

  _createClass(ClaimTransaction, [{
    key: "clone",
    value: function clone(scripts) {
      return new this.constructor({
        version: this.version,
        attributes: this.attributes,
        inputs: this.inputs,
        outputs: this.outputs,
        scripts: scripts,
        claims: this.claims
      });
    }
  }, {
    key: "serializeExclusiveBase",
    value: function serializeExclusiveBase(writer) {
      writer.writeArray(this.claims, function (claim) {
        claim.serializeWireBase(writer);
      });
    }
  }, {
    key: "getNetworkFee",
    // eslint-disable-next-line
    value: function () {
      var _ref6 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee3(context) {
        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                return _context3.abrupt("return", utils.ZERO);

              case 1:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function getNetworkFee(_x3) {
        return _ref6.apply(this, arguments);
      }

      return getNetworkFee;
    }()
  }, {
    key: "getScriptHashesForVerifying",
    value: function () {
      var _ref7 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee4(options) {
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                return _context4.abrupt("return", this.__claimGetScriptHashesForVerifying(options));

              case 1:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function getScriptHashesForVerifying(_x4) {
        return _ref7.apply(this, arguments);
      }

      return getScriptHashesForVerifying;
    }()
  }, {
    key: "verify",
    value: function () {
      var _ref8 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee5(options) {
        return regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                _context5.next = 2;
                return Promise.all([_get(ClaimTransaction.prototype.__proto__ || Object.getPrototypeOf(ClaimTransaction.prototype), "verify", this).call(this, options), this._verify(options)]);

              case 2:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function verify(_x5) {
        return _ref8.apply(this, arguments);
      }

      return verify;
    }()
  }, {
    key: "_verify",
    value: function () {
      var _ref9 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee6(options) {
        var _this2 = this;

        var calculateClaimAmount, getOutput, utilityToken, _options$memPool, memPool, _ref10, _ref11, results, claimAmount, result;

        return regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                calculateClaimAmount = options.calculateClaimAmount, getOutput = options.getOutput, utilityToken = options.utilityToken, _options$memPool = options.memPool, memPool = _options$memPool === void 0 ? [] : _options$memPool;

                if (!hasDuplicateInputs(this.claims)) {
                  _context6.next = 3;
                  break;
                }

                throw new VerifyError('Duplicate claims');

              case 3:
                if (!memPool.some(function (transaction) {
                  return transaction instanceof ClaimTransaction && transaction.type === TRANSACTION_TYPE.CLAIM && hasIntersectingInputs(_this2.claims, transaction.claims);
                })) {
                  _context6.next = 5;
                  break;
                }

                throw new VerifyError('Dupliate claims in mempool');

              case 5:
                _context6.next = 7;
                return Promise.all([this.getTransactionResults({
                  getOutput: getOutput
                }), calculateClaimAmount(this.claims).catch(function (error) {
                  throw new VerifyError("Invalid claims: ".concat(error.message));
                })]);

              case 7:
                _ref10 = _context6.sent;
                _ref11 = _slicedToArray(_ref10, 2);
                results = _ref11[0];
                claimAmount = _ref11[1];
                result = utils.entries(results).find( // eslint-disable-next-line
                function (_ref12) {
                  var _ref13 = _slicedToArray(_ref12, 2),
                      assetHex = _ref13[0],
                      _$$1 = _ref13[1];

                  return common.uInt256Equal(common.hexToUInt256(assetHex), utilityToken.hash);
                });

                if (!(result == null || result[1].gt(utils.ZERO))) {
                  _context6.next = 14;
                  break;
                }

                throw new VerifyError('Invalid claim value');

              case 14:
                if (claimAmount.eq(result[1].neg())) {
                  _context6.next = 16;
                  break;
                }

                throw new VerifyError('Invalid claim value');

              case 16:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function _verify(_x6) {
        return _ref9.apply(this, arguments);
      }

      return _verify;
    }() // eslint-disable-next-line

  }, {
    key: "serializeJSON",
    value: function () {
      var _ref14 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee7(context) {
        var transactionBaseJSON;
        return regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                _context7.next = 2;
                return _get(ClaimTransaction.prototype.__proto__ || Object.getPrototypeOf(ClaimTransaction.prototype), "serializeTransactionBaseJSON", this).call(this, context);

              case 2:
                transactionBaseJSON = _context7.sent;
                return _context7.abrupt("return", {
                  type: 'ClaimTransaction',
                  txid: transactionBaseJSON.txid,
                  size: transactionBaseJSON.size,
                  version: transactionBaseJSON.version,
                  attributes: transactionBaseJSON.attributes,
                  vin: transactionBaseJSON.vin,
                  vout: transactionBaseJSON.vout,
                  scripts: transactionBaseJSON.scripts,
                  sys_fee: transactionBaseJSON.sys_fee,
                  net_fee: transactionBaseJSON.net_fee,
                  claims: this.claims.map(function (claim) {
                    return claim.serializeJSON(context);
                  })
                });

              case 4:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function serializeJSON(_x7) {
        return _ref14.apply(this, arguments);
      }

      return serializeJSON;
    }()
  }, {
    key: "size",
    get: function get() {
      return this.__size();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;

      var _rollupPluginBabelHel = _get(ClaimTransaction.__proto__ || Object.getPrototypeOf(ClaimTransaction), "deserializeTransactionBaseStartWireBase", this).call(this, options),
          type = _rollupPluginBabelHel.type,
          version = _rollupPluginBabelHel.version;

      if (type !== TRANSACTION_TYPE.CLAIM) {
        throw new InvalidFormatError();
      }

      var claims = reader.readArray(function () {
        return Input.deserializeWireBase(options);
      });

      var _rollupPluginBabelHel2 = _get(ClaimTransaction.__proto__ || Object.getPrototypeOf(ClaimTransaction), "deserializeTransactionBaseEndWireBase", this).call(this, options),
          attributes = _rollupPluginBabelHel2.attributes,
          inputs = _rollupPluginBabelHel2.inputs,
          outputs = _rollupPluginBabelHel2.outputs,
          scripts = _rollupPluginBabelHel2.scripts;

      return new this({
        version: version,
        claims: claims,
        attributes: attributes,
        inputs: inputs,
        outputs: outputs,
        scripts: scripts
      });
    }
  }]);
  return ClaimTransaction;
}(TransactionBase);

var EnrollmentTransaction =
/*#__PURE__*/
function (_TransactionBase) {
  _inherits(EnrollmentTransaction, _TransactionBase);

  function EnrollmentTransaction(_ref) {
    var _this;

    var version = _ref.version,
        attributes = _ref.attributes,
        inputs = _ref.inputs,
        outputs = _ref.outputs,
        scripts = _ref.scripts,
        hash = _ref.hash,
        publicKey = _ref.publicKey;
    _classCallCheck(this, EnrollmentTransaction);
    _this = _possibleConstructorReturn(this, (EnrollmentTransaction.__proto__ || Object.getPrototypeOf(EnrollmentTransaction)).call(this, {
      version: version,
      type: TRANSACTION_TYPE.ENROLLMENT,
      attributes: attributes,
      inputs: inputs,
      outputs: outputs,
      scripts: scripts,
      hash: hash
    }));
    _this.publicKey = publicKey;

    if (_this.version !== 0) {
      throw new InvalidFormatError();
    }

    _this.__size = utils.lazy(function () {
      return _get(EnrollmentTransaction.prototype.__proto__ || Object.getPrototypeOf(EnrollmentTransaction.prototype), "size", _this) + IOHelper.sizeOfUInt8 + IOHelper.sizeOfECPoint(_this.publicKey);
    });
    _this.__enrollmentGetScriptHashesForVerifying = utils.lazyAsync(function () {
      var _ref2 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee(options) {
        var hashes;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return _get(EnrollmentTransaction.prototype.__proto__ || Object.getPrototypeOf(EnrollmentTransaction.prototype), "getScriptHashesForVerifying", _this).call(_this, options);

              case 2:
                hashes = _context.sent;
                return _context.abrupt("return", new Set([].concat(_toConsumableArray(hashes), [common.uInt160ToHex(crypto$2.getVerificationScriptHash(_this.publicKey))])));

              case 4:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      return function (_x) {
        return _ref2.apply(this, arguments);
      };
    }());
    return _this;
  }

  _createClass(EnrollmentTransaction, [{
    key: "clone",
    value: function clone(scripts) {
      return new this.constructor({
        version: this.version,
        attributes: this.attributes,
        inputs: this.inputs,
        outputs: this.outputs,
        scripts: scripts,
        publicKey: this.publicKey
      });
    }
  }, {
    key: "serializeExclusiveBase",
    value: function serializeExclusiveBase(writer) {
      writer.writeECPoint(this.publicKey);
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function () {
      var _ref3 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee2(context) {
        var transactionBaseJSON;
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.next = 2;
                return _get(EnrollmentTransaction.prototype.__proto__ || Object.getPrototypeOf(EnrollmentTransaction.prototype), "serializeTransactionBaseJSON", this).call(this, context);

              case 2:
                transactionBaseJSON = _context2.sent;
                return _context2.abrupt("return", {
                  type: 'EnrollmentTransaction',
                  txid: transactionBaseJSON.txid,
                  size: transactionBaseJSON.size,
                  version: transactionBaseJSON.version,
                  attributes: transactionBaseJSON.attributes,
                  vin: transactionBaseJSON.vin,
                  vout: transactionBaseJSON.vout,
                  scripts: transactionBaseJSON.scripts,
                  sys_fee: transactionBaseJSON.sys_fee,
                  net_fee: transactionBaseJSON.net_fee,
                  pubkey: JSONHelper.writeECPoint(this.publicKey)
                });

              case 4:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function serializeJSON(_x2) {
        return _ref3.apply(this, arguments);
      }

      return serializeJSON;
    }()
  }, {
    key: "getScriptHashesForVerifying",
    value: function () {
      var _ref4 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee3(options) {
        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                return _context3.abrupt("return", this.__enrollmentGetScriptHashesForVerifying(options));

              case 1:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function getScriptHashesForVerifying(_x3) {
        return _ref4.apply(this, arguments);
      }

      return getScriptHashesForVerifying;
    }() // eslint-disable-next-line

  }, {
    key: "verify",
    value: function () {
      var _ref5 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee4(options) {
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                throw new VerifyError('Enrollment transactions are obsolete');

              case 1:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function verify(_x4) {
        return _ref5.apply(this, arguments);
      }

      return verify;
    }()
  }, {
    key: "size",
    get: function get() {
      return this.__size();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;

      var _rollupPluginBabelHel = _get(EnrollmentTransaction.__proto__ || Object.getPrototypeOf(EnrollmentTransaction), "deserializeTransactionBaseStartWireBase", this).call(this, options),
          type = _rollupPluginBabelHel.type,
          version = _rollupPluginBabelHel.version;

      if (type !== TRANSACTION_TYPE.ENROLLMENT) {
        throw new InvalidFormatError();
      }

      var publicKey = reader.readECPoint();

      var _rollupPluginBabelHel2 = _get(EnrollmentTransaction.__proto__ || Object.getPrototypeOf(EnrollmentTransaction), "deserializeTransactionBaseEndWireBase", this).call(this, options),
          attributes = _rollupPluginBabelHel2.attributes,
          inputs = _rollupPluginBabelHel2.inputs,
          outputs = _rollupPluginBabelHel2.outputs,
          scripts = _rollupPluginBabelHel2.scripts;

      return new this({
        version: version,
        publicKey: publicKey,
        attributes: attributes,
        inputs: inputs,
        outputs: outputs,
        scripts: scripts
      });
    }
  }]);
  return EnrollmentTransaction;
}(TransactionBase);

var RegisterTransaction =
/*#__PURE__*/
function (_TransactionBase) {
  _inherits(RegisterTransaction, _TransactionBase);

  function RegisterTransaction(_ref) {
    var _this;

    var version = _ref.version,
        attributes = _ref.attributes,
        inputs = _ref.inputs,
        outputs = _ref.outputs,
        scripts = _ref.scripts,
        hash = _ref.hash,
        asset = _ref.asset;
    _classCallCheck(this, RegisterTransaction);
    _this = _possibleConstructorReturn(this, (RegisterTransaction.__proto__ || Object.getPrototypeOf(RegisterTransaction)).call(this, {
      version: version,
      type: TRANSACTION_TYPE.REGISTER,
      attributes: attributes,
      inputs: inputs,
      outputs: outputs,
      scripts: scripts,
      hash: hash
    }));
    _this.asset = asset;

    if (_this.version !== 0) {
      throw new InvalidFormatError();
    }

    if (common.ecPointIsInfinity(asset.owner) && asset.type !== ASSET_TYPE.GOVERNING_TOKEN && asset.type !== ASSET_TYPE.UTILITY_TOKEN) {
      throw new InvalidFormatError();
    }

    _this.__size = utils.lazy(function () {
      return _get(RegisterTransaction.prototype.__proto__ || Object.getPrototypeOf(RegisterTransaction.prototype), "size", _this) + IOHelper.sizeOfUInt8 + IOHelper.sizeOfUInt8 + IOHelper.sizeOfVarString(_this.asset.name) + IOHelper.sizeOfFixed8 + IOHelper.sizeOfUInt8 + IOHelper.sizeOfECPoint(_this.asset.owner) + IOHelper.sizeOfUInt160;
    });
    _this.__registerGetScriptHashesForVerifying = utils.lazyAsync(function () {
      var _ref2 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee(options) {
        var hashes, scriptHash;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return _get(RegisterTransaction.prototype.__proto__ || Object.getPrototypeOf(RegisterTransaction.prototype), "getScriptHashesForVerifying", _this).call(_this, options);

              case 2:
                hashes = _context.sent;
                scriptHash = common.uInt160ToHex(crypto$2.getVerificationScriptHash(_this.asset.owner));
                return _context.abrupt("return", new Set([].concat(_toConsumableArray(hashes), [scriptHash])));

              case 5:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      return function (_x) {
        return _ref2.apply(this, arguments);
      };
    }());
    return _this;
  }

  _createClass(RegisterTransaction, [{
    key: "clone",
    value: function clone(scripts) {
      return new this.constructor({
        version: this.version,
        attributes: this.attributes,
        inputs: this.inputs,
        outputs: this.outputs,
        scripts: scripts,
        hash: this.hash,
        asset: this.asset
      });
    }
  }, {
    key: "serializeExclusiveBase",
    value: function serializeExclusiveBase(writer) {
      writer.writeUInt8(this.asset.type);
      writer.writeVarString(this.asset.name);
      writer.writeFixed8(this.asset.amount);
      writer.writeUInt8(this.asset.precision);
      writer.writeECPoint(this.asset.owner);
      writer.writeUInt160(this.asset.admin);
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function () {
      var _ref3 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee2(context) {
        var transactionBaseJSON, name;
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.next = 2;
                return _get(RegisterTransaction.prototype.__proto__ || Object.getPrototypeOf(RegisterTransaction.prototype), "serializeTransactionBaseJSON", this).call(this, context);

              case 2:
                transactionBaseJSON = _context2.sent;
                name = this.asset.name;

                try {
                  name = JSON.parse(name);
                } catch (error) {// eslint-disable-next-line
                }

                return _context2.abrupt("return", {
                  type: 'RegisterTransaction',
                  txid: transactionBaseJSON.txid,
                  size: transactionBaseJSON.size,
                  version: transactionBaseJSON.version,
                  attributes: transactionBaseJSON.attributes,
                  vin: transactionBaseJSON.vin,
                  vout: transactionBaseJSON.vout,
                  scripts: transactionBaseJSON.scripts,
                  sys_fee: transactionBaseJSON.sys_fee,
                  net_fee: transactionBaseJSON.net_fee,
                  asset: {
                    type: toJSONAssetType(this.asset.type),
                    name: name,
                    amount: JSONHelper.writeFixed8(this.asset.amount),
                    precision: this.asset.precision,
                    owner: JSONHelper.writeECPoint(this.asset.owner),
                    admin: crypto$2.scriptHashToAddress({
                      addressVersion: context.addressVersion,
                      scriptHash: this.asset.admin
                    })
                  }
                });

              case 6:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function serializeJSON(_x2) {
        return _ref3.apply(this, arguments);
      }

      return serializeJSON;
    }()
  }, {
    key: "getSystemFee",
    value: function getSystemFee(context) {
      if (this.asset.type === ASSET_TYPE.GOVERNING_TOKEN || this.asset.type === ASSET_TYPE.UTILITY_TOKEN) {
        return utils.ZERO;
      }

      return _get(RegisterTransaction.prototype.__proto__ || Object.getPrototypeOf(RegisterTransaction.prototype), "getSystemFee", this).call(this, context);
    }
  }, {
    key: "getScriptHashesForVerifying",
    value: function () {
      var _ref4 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee3(options) {
        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                return _context3.abrupt("return", this.__registerGetScriptHashesForVerifying(options));

              case 1:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function getScriptHashesForVerifying(_x3) {
        return _ref4.apply(this, arguments);
      }

      return getScriptHashesForVerifying;
    }() // eslint-disable-next-line

  }, {
    key: "verify",
    value: function () {
      var _ref5 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee4(options) {
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                throw new VerifyError('Enrollment transactions are obsolete');

              case 1:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function verify(_x4) {
        return _ref5.apply(this, arguments);
      }

      return verify;
    }()
  }, {
    key: "size",
    get: function get() {
      return this.__size();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;

      var _rollupPluginBabelHel = _get(RegisterTransaction.__proto__ || Object.getPrototypeOf(RegisterTransaction), "deserializeTransactionBaseStartWireBase", this).call(this, options),
          type = _rollupPluginBabelHel.type,
          version = _rollupPluginBabelHel.version;

      if (type !== TRANSACTION_TYPE.REGISTER) {
        throw new InvalidFormatError();
      }

      var assetType = assertAssetType(reader.readUInt8());
      var name = reader.readVarString(1024);
      var amount = reader.readFixed8();
      var precision = reader.readUInt8();
      var owner = reader.readECPoint();
      var admin = reader.readUInt160();

      var _rollupPluginBabelHel2 = _get(RegisterTransaction.__proto__ || Object.getPrototypeOf(RegisterTransaction), "deserializeTransactionBaseEndWireBase", this).call(this, options),
          attributes = _rollupPluginBabelHel2.attributes,
          inputs = _rollupPluginBabelHel2.inputs,
          outputs = _rollupPluginBabelHel2.outputs,
          scripts = _rollupPluginBabelHel2.scripts;

      return new this({
        version: version,
        attributes: attributes,
        inputs: inputs,
        outputs: outputs,
        scripts: scripts,
        asset: {
          type: assetType,
          name: name,
          amount: amount,
          precision: precision,
          owner: owner,
          admin: admin
        }
      });
    }
  }]);
  return RegisterTransaction;
}(TransactionBase);

var ContractTransaction =
/*#__PURE__*/
function (_TransactionBase) {
  _inherits(ContractTransaction, _TransactionBase);

  function ContractTransaction(_ref) {
    var _this;

    var version = _ref.version,
        attributes = _ref.attributes,
        inputs = _ref.inputs,
        outputs = _ref.outputs,
        scripts = _ref.scripts,
        hash = _ref.hash;
    _classCallCheck(this, ContractTransaction);
    _this = _possibleConstructorReturn(this, (ContractTransaction.__proto__ || Object.getPrototypeOf(ContractTransaction)).call(this, {
      version: version,
      type: TRANSACTION_TYPE.CONTRACT,
      attributes: attributes,
      inputs: inputs,
      outputs: outputs,
      scripts: scripts,
      hash: hash
    }));

    if (_this.version !== 0) {
      throw new InvalidFormatError();
    }

    _this.__size = utils.lazy(function () {
      return _get(ContractTransaction.prototype.__proto__ || Object.getPrototypeOf(ContractTransaction.prototype), "size", _this) + IOHelper.sizeOfUInt8;
    });
    return _this;
  }

  _createClass(ContractTransaction, [{
    key: "clone",
    value: function clone(scripts) {
      return new this.constructor({
        version: this.version,
        attributes: this.attributes,
        inputs: this.inputs,
        outputs: this.outputs,
        scripts: scripts
      });
    } // eslint-disable-next-line

  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function () {
      var _ref2 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee(context) {
        var transactionBaseJSON;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return _get(ContractTransaction.prototype.__proto__ || Object.getPrototypeOf(ContractTransaction.prototype), "serializeTransactionBaseJSON", this).call(this, context);

              case 2:
                transactionBaseJSON = _context.sent;
                return _context.abrupt("return", {
                  type: 'ContractTransaction',
                  txid: transactionBaseJSON.txid,
                  size: transactionBaseJSON.size,
                  version: transactionBaseJSON.version,
                  attributes: transactionBaseJSON.attributes,
                  vin: transactionBaseJSON.vin,
                  vout: transactionBaseJSON.vout,
                  scripts: transactionBaseJSON.scripts,
                  sys_fee: transactionBaseJSON.sys_fee,
                  net_fee: transactionBaseJSON.net_fee
                });

              case 4:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function serializeJSON(_x) {
        return _ref2.apply(this, arguments);
      }

      return serializeJSON;
    }()
  }, {
    key: "size",
    get: function get() {
      return this.__size();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var _rollupPluginBabelHel = _get(ContractTransaction.__proto__ || Object.getPrototypeOf(ContractTransaction), "deserializeTransactionBaseStartWireBase", this).call(this, options),
          type = _rollupPluginBabelHel.type,
          version = _rollupPluginBabelHel.version;

      if (type !== TRANSACTION_TYPE.CONTRACT) {
        throw new InvalidFormatError();
      }

      var _rollupPluginBabelHel2 = _get(ContractTransaction.__proto__ || Object.getPrototypeOf(ContractTransaction), "deserializeTransactionBaseEndWireBase", this).call(this, options),
          attributes = _rollupPluginBabelHel2.attributes,
          inputs = _rollupPluginBabelHel2.inputs,
          outputs = _rollupPluginBabelHel2.outputs,
          scripts = _rollupPluginBabelHel2.scripts;

      return new this({
        version: version,
        attributes: attributes,
        inputs: inputs,
        outputs: outputs,
        scripts: scripts
      });
    }
  }]);
  return ContractTransaction;
}(TransactionBase);

var _fixBabelExtend$10 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var CONTRACT_PARAMETER_TYPE = {
  SIGNATURE: 0x00,
  BOOLEAN: 0x01,
  INTEGER: 0x02,
  HASH160: 0x03,
  HASH256: 0x04,
  BYTE_ARRAY: 0x05,
  PUBLIC_KEY: 0x06,
  STRING: 0x07,
  ARRAY: 0x10,
  INTEROP_INTERFACE: 0xf0,
  VOID: 0xff
};
var InvalidContractParameterTypeError = _fixBabelExtend$10(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidContractParameterTypeError, _Error);

  function InvalidContractParameterTypeError(contractParameterType) {
    var _this;

    _classCallCheck(this, InvalidContractParameterTypeError);
    _this = _possibleConstructorReturn(this, (InvalidContractParameterTypeError.__proto__ || Object.getPrototypeOf(InvalidContractParameterTypeError)).call(this, "Expected contract parameter type, " + "found: ".concat(contractParameterType.toString(16))));
    _this.contractParameterType = contractParameterType;
    return _this;
  }

  return InvalidContractParameterTypeError;
}(Error));
var assertContractParameterType = function assertContractParameterType(value) {
  switch (value) {
    case CONTRACT_PARAMETER_TYPE.SIGNATURE:
      return CONTRACT_PARAMETER_TYPE.SIGNATURE;

    case CONTRACT_PARAMETER_TYPE.BOOLEAN:
      return CONTRACT_PARAMETER_TYPE.BOOLEAN;

    case CONTRACT_PARAMETER_TYPE.INTEGER:
      return CONTRACT_PARAMETER_TYPE.INTEGER;

    case CONTRACT_PARAMETER_TYPE.HASH160:
      return CONTRACT_PARAMETER_TYPE.HASH160;

    case CONTRACT_PARAMETER_TYPE.HASH256:
      return CONTRACT_PARAMETER_TYPE.HASH256;

    case CONTRACT_PARAMETER_TYPE.BYTE_ARRAY:
      return CONTRACT_PARAMETER_TYPE.BYTE_ARRAY;

    case CONTRACT_PARAMETER_TYPE.PUBLIC_KEY:
      return CONTRACT_PARAMETER_TYPE.PUBLIC_KEY;

    case CONTRACT_PARAMETER_TYPE.STRING:
      return CONTRACT_PARAMETER_TYPE.STRING;

    case CONTRACT_PARAMETER_TYPE.ARRAY:
      return CONTRACT_PARAMETER_TYPE.ARRAY;

    case CONTRACT_PARAMETER_TYPE.INTEROP_INTERFACE:
      return CONTRACT_PARAMETER_TYPE.INTEROP_INTERFACE;
    // TODO: Seems to be a bug in the TestNet

    case 0x16:
      return CONTRACT_PARAMETER_TYPE.ARRAY;

    case CONTRACT_PARAMETER_TYPE.VOID:
      return CONTRACT_PARAMETER_TYPE.VOID;

    default:
      throw new InvalidContractParameterTypeError(value);
  }
};
var toJSONContractParameterType = function toJSONContractParameterType(type) {
  switch (type) {
    case 0x00:
      return 'Signature';

    case 0x01:
      return 'Boolean';

    case 0x02:
      return 'Integer';

    case 0x03:
      return 'Hash160';

    case 0x04:
      return 'Hash256';

    case 0x05:
      return 'ByteArray';

    case 0x06:
      return 'PublicKey';

    case 0x07:
      return 'String';

    case 0x10:
      return 'Array';

    case 0xf0:
      return 'InteropInterface';

    case 0xff:
      return 'Void';

    default:
      // eslint-disable-next-line
      
      throw new InvalidContractParameterTypeError(type);
  }
};

var ContractParameterBase =
/*#__PURE__*/
function () {
  function ContractParameterBase() {
    _classCallCheck(this, ContractParameterBase);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(this.serializeWireBase.bind(this))
    });
  }

  _createClass(ContractParameterBase, [{
    key: "asBuffer",
    // eslint-disable-next-line
    value: function asBuffer() {
      throw new Error('Unimplemented.');
    }
  }, {
    key: "asBoolean",
    value: function asBoolean() {
      return this.asBuffer().some(function (value) {
        return value !== 0;
      });
    } // TODO: Need to deal with circular references in arrays

  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt8(this.type);
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      throw new Error('Not Implemented');
    }
  }], [{
    key: "deserializeContractParameterBaseWireBase",
    // eslint-disable-next-line
    value: function deserializeContractParameterBaseWireBase(_ref) {
      var reader = _ref.reader;
      var type = reader.readUInt8();
      return {
        type: type
      };
    } // eslint-disable-next-line

  }, {
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      throw new Error('Not Implemented');
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return ContractParameterBase;
}();

// TODO: What is this? How is it used?
var SignatureContractParameter =
/*#__PURE__*/
function (_ContractParameterBas) {
  _inherits(SignatureContractParameter, _ContractParameterBas);

  function SignatureContractParameter(value) {
    var _this;

    _classCallCheck(this, SignatureContractParameter);
    _this = _possibleConstructorReturn(this, (SignatureContractParameter.__proto__ || Object.getPrototypeOf(SignatureContractParameter)).call(this));
    Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: CONTRACT_PARAMETER_TYPE.SIGNATURE
    });
    _this.value = value;
    return _this;
  }

  _createClass(SignatureContractParameter, [{
    key: "asBuffer",
    value: function asBuffer() {
      return this.value;
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(SignatureContractParameter.prototype.__proto__ || Object.getPrototypeOf(SignatureContractParameter.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeVarBytesLE(this.value);
    }
  }, {
    key: "serializeJSON",
    value: function serializeJSON( // eslint-disable-next-line
    context) {
      return {
        type: 'Signature',
        value: JSONHelper.writeBuffer(this.value)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      _get(SignatureContractParameter.__proto__ || Object.getPrototypeOf(SignatureContractParameter), "deserializeContractParameterBaseWireBase", this).call(this, options);
      var value = reader.readVarBytesLE();
      return new this(value);
    }
  }]);
  return SignatureContractParameter;
}(ContractParameterBase);

var BooleanContractParameter =
/*#__PURE__*/
function (_ContractParameterBas) {
  _inherits(BooleanContractParameter, _ContractParameterBas);

  function BooleanContractParameter(value) {
    var _this;

    _classCallCheck(this, BooleanContractParameter);
    _this = _possibleConstructorReturn(this, (BooleanContractParameter.__proto__ || Object.getPrototypeOf(BooleanContractParameter)).call(this));
    Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: CONTRACT_PARAMETER_TYPE.BOOLEAN
    });
    _this.value = value;
    return _this;
  }

  _createClass(BooleanContractParameter, [{
    key: "asBuffer",
    value: function asBuffer() {
      return this.value ? this.constructor.TRUE : this.constructor.FALSE;
    }
  }, {
    key: "asBoolean",
    value: function asBoolean() {
      return this.value;
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(BooleanContractParameter.prototype.__proto__ || Object.getPrototypeOf(BooleanContractParameter.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeBoolean(this.value);
    }
  }, {
    key: "serializeJSON",
    value: function serializeJSON( // eslint-disable-next-line
    context) {
      return {
        type: 'Boolean',
        value: this.value
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      _get(BooleanContractParameter.__proto__ || Object.getPrototypeOf(BooleanContractParameter), "deserializeContractParameterBaseWireBase", this).call(this, options);
      var value = reader.readBoolean();
      return new this(value);
    }
  }]);
  return BooleanContractParameter;
}(ContractParameterBase);

Object.defineProperty(BooleanContractParameter, "TRUE", {
  configurable: true,
  enumerable: true,
  writable: true,
  value: Buffer.from([1])
});
Object.defineProperty(BooleanContractParameter, "FALSE", {
  configurable: true,
  enumerable: true,
  writable: true,
  value: Buffer.from([0])
});

var IntegerContractParameter =
/*#__PURE__*/
function (_ContractParameterBas) {
  _inherits(IntegerContractParameter, _ContractParameterBas);

  function IntegerContractParameter(value) {
    var _this;

    _classCallCheck(this, IntegerContractParameter);
    _this = _possibleConstructorReturn(this, (IntegerContractParameter.__proto__ || Object.getPrototypeOf(IntegerContractParameter)).call(this));
    Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: CONTRACT_PARAMETER_TYPE.INTEGER
    });
    _this.value = value;
    return _this;
  }

  _createClass(IntegerContractParameter, [{
    key: "asBoolean",
    value: function asBoolean() {
      return !this.value.isZero();
    }
  }, {
    key: "asBuffer",
    value: function asBuffer() {
      return utils.toSignedBuffer(this.value);
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(IntegerContractParameter.prototype.__proto__ || Object.getPrototypeOf(IntegerContractParameter.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeVarBytesLE(utils.toSignedBuffer(this.value));
    }
  }, {
    key: "serializeJSON",
    value: function serializeJSON( // eslint-disable-next-line
    context) {
      return {
        type: 'Integer',
        value: this.value.toString(10)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      _get(IntegerContractParameter.__proto__ || Object.getPrototypeOf(IntegerContractParameter), "deserializeContractParameterBaseWireBase", this).call(this, options);
      var value = utils.fromSignedBuffer(reader.readVarBytesLE());
      return new this(value);
    }
  }]);
  return IntegerContractParameter;
}(ContractParameterBase);

var Hash160ContractParameter =
/*#__PURE__*/
function (_ContractParameterBas) {
  _inherits(Hash160ContractParameter, _ContractParameterBas);

  function Hash160ContractParameter(value) {
    var _this;

    _classCallCheck(this, Hash160ContractParameter);
    _this = _possibleConstructorReturn(this, (Hash160ContractParameter.__proto__ || Object.getPrototypeOf(Hash160ContractParameter)).call(this));
    Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: CONTRACT_PARAMETER_TYPE.HASH160
    });
    _this.value = value;
    return _this;
  }

  _createClass(Hash160ContractParameter, [{
    key: "asBuffer",
    value: function asBuffer() {
      return common.uInt160ToBuffer(this.value);
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(Hash160ContractParameter.prototype.__proto__ || Object.getPrototypeOf(Hash160ContractParameter.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeUInt160(this.value);
    }
  }, {
    key: "serializeJSON",
    value: function serializeJSON( // eslint-disable-next-line
    context) {
      return {
        type: 'Hash160',
        value: JSONHelper.writeUInt160(this.value)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      _get(Hash160ContractParameter.__proto__ || Object.getPrototypeOf(Hash160ContractParameter), "deserializeContractParameterBaseWireBase", this).call(this, options);
      var value = reader.readUInt160();
      return new this(value);
    }
  }]);
  return Hash160ContractParameter;
}(ContractParameterBase);

var Hash256ContractParameter =
/*#__PURE__*/
function (_ContractParameterBas) {
  _inherits(Hash256ContractParameter, _ContractParameterBas);

  function Hash256ContractParameter(value) {
    var _this;

    _classCallCheck(this, Hash256ContractParameter);
    _this = _possibleConstructorReturn(this, (Hash256ContractParameter.__proto__ || Object.getPrototypeOf(Hash256ContractParameter)).call(this));
    Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: CONTRACT_PARAMETER_TYPE.HASH256
    });
    _this.value = value;
    return _this;
  }

  _createClass(Hash256ContractParameter, [{
    key: "asBuffer",
    value: function asBuffer() {
      return common.uInt256ToBuffer(this.value);
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(Hash256ContractParameter.prototype.__proto__ || Object.getPrototypeOf(Hash256ContractParameter.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeUInt256(this.value);
    }
  }, {
    key: "serializeJSON",
    value: function serializeJSON( // eslint-disable-next-line
    context) {
      return {
        type: 'Hash256',
        value: JSONHelper.writeUInt256(this.value)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      _get(Hash256ContractParameter.__proto__ || Object.getPrototypeOf(Hash256ContractParameter), "deserializeContractParameterBaseWireBase", this).call(this, options);
      var value = reader.readUInt256();
      return new this(value);
    }
  }]);
  return Hash256ContractParameter;
}(ContractParameterBase);

var ByteArrayContractParameter =
/*#__PURE__*/
function (_ContractParameterBas) {
  _inherits(ByteArrayContractParameter, _ContractParameterBas);

  function ByteArrayContractParameter(value) {
    var _this;

    _classCallCheck(this, ByteArrayContractParameter);
    _this = _possibleConstructorReturn(this, (ByteArrayContractParameter.__proto__ || Object.getPrototypeOf(ByteArrayContractParameter)).call(this));
    Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: CONTRACT_PARAMETER_TYPE.BYTE_ARRAY
    });
    _this.value = value;
    return _this;
  }

  _createClass(ByteArrayContractParameter, [{
    key: "asBuffer",
    value: function asBuffer() {
      return this.value;
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(ByteArrayContractParameter.prototype.__proto__ || Object.getPrototypeOf(ByteArrayContractParameter.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeVarBytesLE(this.value);
    }
  }, {
    key: "serializeJSON",
    value: function serializeJSON( // eslint-disable-next-line
    context) {
      return {
        type: 'ByteArray',
        value: JSONHelper.writeBuffer(this.value)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      _get(ByteArrayContractParameter.__proto__ || Object.getPrototypeOf(ByteArrayContractParameter), "deserializeContractParameterBaseWireBase", this).call(this, options);
      var value = reader.readVarBytesLE();
      return new this(value);
    }
  }]);
  return ByteArrayContractParameter;
}(ContractParameterBase);

var PublicKeyContractParameter =
/*#__PURE__*/
function (_ContractParameterBas) {
  _inherits(PublicKeyContractParameter, _ContractParameterBas);

  function PublicKeyContractParameter(value) {
    var _this;

    _classCallCheck(this, PublicKeyContractParameter);
    _this = _possibleConstructorReturn(this, (PublicKeyContractParameter.__proto__ || Object.getPrototypeOf(PublicKeyContractParameter)).call(this));
    Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: CONTRACT_PARAMETER_TYPE.PUBLIC_KEY
    });
    _this.value = value;
    return _this;
  }

  _createClass(PublicKeyContractParameter, [{
    key: "asBuffer",
    value: function asBuffer() {
      return common.ecPointToBuffer(this.value);
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(PublicKeyContractParameter.prototype.__proto__ || Object.getPrototypeOf(PublicKeyContractParameter.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeECPoint(this.value);
    }
  }, {
    key: "serializeJSON",
    value: function serializeJSON( // eslint-disable-next-line
    context) {
      return {
        type: 'PublicKey',
        value: JSONHelper.writeECPoint(this.value)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      _get(PublicKeyContractParameter.__proto__ || Object.getPrototypeOf(PublicKeyContractParameter), "deserializeContractParameterBaseWireBase", this).call(this, options);
      var value = reader.readECPoint();
      return new this(value);
    }
  }]);
  return PublicKeyContractParameter;
}(ContractParameterBase);

var StringContractParameter =
/*#__PURE__*/
function (_ContractParameterBas) {
  _inherits(StringContractParameter, _ContractParameterBas);

  function StringContractParameter(value) {
    var _this;

    _classCallCheck(this, StringContractParameter);
    _this = _possibleConstructorReturn(this, (StringContractParameter.__proto__ || Object.getPrototypeOf(StringContractParameter)).call(this));
    Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: CONTRACT_PARAMETER_TYPE.STRING
    });
    _this.value = value;
    return _this;
  }

  _createClass(StringContractParameter, [{
    key: "asBuffer",
    value: function asBuffer() {
      return Buffer.from(this.value, 'utf8');
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(StringContractParameter.prototype.__proto__ || Object.getPrototypeOf(StringContractParameter.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeVarString(this.value);
    }
  }, {
    key: "serializeJSON",
    value: function serializeJSON( // eslint-disable-next-line
    context) {
      return {
        type: 'String',
        value: this.value
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      _get(StringContractParameter.__proto__ || Object.getPrototypeOf(StringContractParameter), "deserializeContractParameterBaseWireBase", this).call(this, options);
      var value = reader.readVarString();
      return new this(value);
    }
  }]);
  return StringContractParameter;
}(ContractParameterBase);

var ArrayContractParameter =
/*#__PURE__*/
function (_ContractParameterBas) {
  _inherits(ArrayContractParameter, _ContractParameterBas);

  function ArrayContractParameter(value) {
    var _this;

    _classCallCheck(this, ArrayContractParameter);
    _this = _possibleConstructorReturn(this, (ArrayContractParameter.__proto__ || Object.getPrototypeOf(ArrayContractParameter)).call(this));
    Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: CONTRACT_PARAMETER_TYPE.ARRAY
    });
    _this.value = value;
    return _this;
  }

  _createClass(ArrayContractParameter, [{
    key: "asBoolean",
    value: function asBoolean() {
      return this.value.length > 0;
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(ArrayContractParameter.prototype.__proto__ || Object.getPrototypeOf(ArrayContractParameter.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeArray(this.value, function (parameter) {
        return parameter.serializeWireBase(writer);
      });
    } // deserialize is monkey patched on later

  }, {
    key: "serializeJSON",
    value: function serializeJSON(context) {
      return {
        type: 'Array',
        value: this.value.map(function (val) {
          return val.serializeJSON(context);
        })
      };
    }
  }]);
  return ArrayContractParameter;
}(ContractParameterBase);

var InteropInterfaceContractParameter =
/*#__PURE__*/
function (_ContractParameterBas) {
  _inherits(InteropInterfaceContractParameter, _ContractParameterBas);

  function InteropInterfaceContractParameter() {
    var _ref;

    var _temp, _this;

    _classCallCheck(this, InteropInterfaceContractParameter);

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return _possibleConstructorReturn(_this, (_temp = _this = _possibleConstructorReturn(this, (_ref = InteropInterfaceContractParameter.__proto__ || Object.getPrototypeOf(InteropInterfaceContractParameter)).call.apply(_ref, [this].concat(args))), Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: CONTRACT_PARAMETER_TYPE.INTEROP_INTERFACE
    }), _temp));
  }

  _createClass(InteropInterfaceContractParameter, [{
    key: "serializeJSON",
    value: function serializeJSON( // eslint-disable-next-line
    context) {
      return {
        type: 'InteropInterface'
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      _get(InteropInterfaceContractParameter.__proto__ || Object.getPrototypeOf(InteropInterfaceContractParameter), "deserializeContractParameterBaseWireBase", this).call(this, options);
      return new this();
    }
  }]);
  return InteropInterfaceContractParameter;
}(ContractParameterBase);

var VoidContractParameter =
/*#__PURE__*/
function (_ContractParameterBas) {
  _inherits(VoidContractParameter, _ContractParameterBas);

  function VoidContractParameter() {
    var _ref;

    var _temp, _this;

    _classCallCheck(this, VoidContractParameter);

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return _possibleConstructorReturn(_this, (_temp = _this = _possibleConstructorReturn(this, (_ref = VoidContractParameter.__proto__ || Object.getPrototypeOf(VoidContractParameter)).call.apply(_ref, [this].concat(args))), Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: CONTRACT_PARAMETER_TYPE.VOID
    }), _temp));
  }

  _createClass(VoidContractParameter, [{
    key: "asBoolean",
    value: function asBoolean() {
      return false;
    }
  }, {
    key: "serializeJSON",
    value: function serializeJSON( // eslint-disable-next-line
    context) {
      return {
        type: 'Void'
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      _get(VoidContractParameter.__proto__ || Object.getPrototypeOf(VoidContractParameter), "deserializeContractParameterBaseWireBase", this).call(this, options);
      return new this();
    }
  }]);
  return VoidContractParameter;
}(ContractParameterBase);

var deserializeWireBase$1 = function deserializeWireBase(options) {
  var reader = options.reader;
  var type = assertContractParameterType(reader.clone().readUInt8());

  switch (type) {
    case 0x00:
      return SignatureContractParameter.deserializeWireBase(options);

    case 0x01:
      return BooleanContractParameter.deserializeWireBase(options);

    case 0x02:
      return IntegerContractParameter.deserializeWireBase(options);

    case 0x03:
      return Hash160ContractParameter.deserializeWireBase(options);

    case 0x04:
      return Hash256ContractParameter.deserializeWireBase(options);

    case 0x05:
      return ByteArrayContractParameter.deserializeWireBase(options);

    case 0x06:
      return PublicKeyContractParameter.deserializeWireBase(options);

    case 0x07:
      return StringContractParameter.deserializeWireBase(options);

    case 0x10:
      return ArrayContractParameter.deserializeWireBase(options);

    case 0xf0:
      return InteropInterfaceContractParameter.deserializeWireBase(options);

    case 0xff:
      return VoidContractParameter.deserializeWireBase(options);

    default:
      // eslint-disable-next-line
      
      throw new InvalidContractParameterTypeError(type);
  }
};
 // $FlowFixMe

ArrayContractParameter.deserializeWireBase = function (options) {
  var reader = options.reader;
  reader.readUInt8();
  var value = reader.readArray(function () {
    return deserializeWireBase$1(options);
  });
  return new ArrayContractParameter(value);
}; // $FlowFixMe


ArrayContractParameter.deserializeWire = createDeserializeWire(ArrayContractParameter.deserializeWireBase.bind(ArrayContractParameter));

var Contract =
/*#__PURE__*/
function (_BaseState) {
  _inherits(Contract, _BaseState);

  function Contract(_ref) {
    var _this;

    var version = _ref.version,
        script = _ref.script,
        parameterList = _ref.parameterList,
        returnType = _ref.returnType,
        hasStorage = _ref.hasStorage,
        name = _ref.name,
        codeVersion = _ref.codeVersion,
        author = _ref.author,
        email = _ref.email,
        description = _ref.description;
    _classCallCheck(this, Contract);
    _this = _possibleConstructorReturn(this, (Contract.__proto__ || Object.getPrototypeOf(Contract)).call(this, {
      version: version
    }));
    Object.defineProperty(_this, "__contractSize", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.lazy(function () {
        return sizeOfContract({
          script: _this.script,
          parameterList: _this.parameterList,
          name: _this.name,
          codeVersion: _this.codeVersion,
          author: _this.author,
          email: _this.email,
          description: _this.description
        });
      })
    });
    Object.defineProperty(_this, "_hash", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.lazy(function () {
        return crypto$2.toScriptHash(_this.script);
      })
    });
    Object.defineProperty(_this, "_hashHex", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.lazy(function () {
        return common.uInt160ToHex(_this.hash);
      })
    });
    Object.defineProperty(_this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(Contract, function (other) {
        return common.uInt160Equal(_this.hash, other.hash);
      })
    });
    Object.defineProperty(_this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(_this.serializeWireBase.bind(_this))
    });
    _this.script = script;
    _this.parameterList = parameterList;
    _this.returnType = returnType;
    _this.hasStorage = hasStorage;
    _this.name = name;
    _this.codeVersion = codeVersion;
    _this.author = author;
    _this.email = email;
    _this.description = description;
    return _this;
  } // eslint-disable-next-line


  _createClass(Contract, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      // eslint-disable-next-line
      serializeContractWireBase({
        writer: writer,
        contract: this
      });
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        version: this.version,
        code: {
          hash: JSONHelper.writeUInt160(this.hash),
          script: JSONHelper.writeBuffer(this.script),
          parameters: this.parameterList.map(function (parameter) {
            return toJSONContractParameterType(parameter);
          }),
          returntype: toJSONContractParameterType(this.returnType)
        },
        needstorage: this.hasStorage,
        name: this.name,
        code_version: this.codeVersion,
        author: this.author,
        email: this.email,
        description: this.description
      };
    }
  }, {
    key: "size",
    get: function get() {
      return this.__contractSize();
    }
  }, {
    key: "hash",
    get: function get() {
      return this._hash();
    }
  }, {
    key: "hashHex",
    get: function get() {
      return this._hashHex();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      // eslint-disable-next-line
      return deserializeContractWireBase({
        context: options.context,
        reader: options.reader
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return Contract;
}(BaseState);

var sizeOfContract = function sizeOfContract(_ref2) {
  var script = _ref2.script,
      parameterList = _ref2.parameterList,
      name = _ref2.name,
      codeVersion = _ref2.codeVersion,
      author = _ref2.author,
      email = _ref2.email,
      description = _ref2.description,
      publishVersion = _ref2.publishVersion;
  return IOHelper.sizeOfVarBytesLE(script) + IOHelper.sizeOfVarBytesLE(Buffer.from(parameterList)) + IOHelper.sizeOfUInt8 + ( // TODO: Doesn't take version into account?
  publishVersion == null ? IOHelper.sizeOfBoolean : 0) + IOHelper.sizeOfVarString(name) + IOHelper.sizeOfVarString(codeVersion) + IOHelper.sizeOfVarString(author) + IOHelper.sizeOfVarString(email) + IOHelper.sizeOfVarString(description);
};
var deserializeContractWireBase = function deserializeContractWireBase(_ref3) {
  var reader = _ref3.reader,
      publishVersion = _ref3.publishVersion;
  var script = reader.readVarBytesLE();
  var parameterList = [].concat(_toConsumableArray(reader.readVarBytesLE())).map(function (value) {
    return assertContractParameterType(value);
  });
  var returnType = assertContractParameterType(reader.readUInt8());
  var hasStorage = publishVersion == null || publishVersion >= 1 ? reader.readBoolean() : false;
  var name = reader.readVarString(252);
  var codeVersion = reader.readVarString(252);
  var author = reader.readVarString(252);
  var email = reader.readVarString(252);
  var description = reader.readVarString(65536);
  return new Contract({
    script: script,
    parameterList: parameterList,
    returnType: returnType,
    hasStorage: hasStorage,
    name: name,
    codeVersion: codeVersion,
    author: author,
    email: email,
    description: description
  });
};
var serializeContractWireBase = function serializeContractWireBase(_ref4) {
  var writer = _ref4.writer,
      contract = _ref4.contract,
      publishVersion = _ref4.publishVersion;
  writer.writeVarBytesLE(contract.script);
  writer.writeVarBytesLE(Buffer.from(contract.parameterList));
  writer.writeUInt8(contract.returnType);

  if (publishVersion == null || publishVersion >= 1) {
    writer.writeBoolean(contract.hasStorage);
  }

  writer.writeVarString(contract.name);
  writer.writeVarString(contract.codeVersion);
  writer.writeVarString(contract.author);
  writer.writeVarString(contract.email);
  writer.writeVarString(contract.description);
};

var PublishTransaction =
/*#__PURE__*/
function (_TransactionBase) {
  _inherits(PublishTransaction, _TransactionBase);

  // TODO: How is version set?
  // static VERSION = 1;
  function PublishTransaction(_ref) {
    var _this;

    var version = _ref.version,
        attributes = _ref.attributes,
        inputs = _ref.inputs,
        outputs = _ref.outputs,
        scripts = _ref.scripts,
        hash = _ref.hash,
        contract = _ref.contract;
    _classCallCheck(this, PublishTransaction);
    _this = _possibleConstructorReturn(this, (PublishTransaction.__proto__ || Object.getPrototypeOf(PublishTransaction)).call(this, {
      version: version,
      type: TRANSACTION_TYPE.PUBLISH,
      attributes: attributes,
      inputs: inputs,
      outputs: outputs,
      scripts: scripts,
      hash: hash
    }));
    _this.contract = contract;

    if (_this.version > 1) {
      throw new InvalidFormatError();
    }

    _this.__size = utils.lazy(function () {
      return _get(PublishTransaction.prototype.__proto__ || Object.getPrototypeOf(PublishTransaction.prototype), "size", _this) + IOHelper.sizeOfUInt8 + sizeOfContract({
        script: _this.contract.script,
        parameterList: _this.contract.parameterList,
        name: _this.contract.name,
        codeVersion: _this.contract.codeVersion,
        author: _this.contract.author,
        email: _this.contract.email,
        description: _this.contract.description,
        publishVersion: _this.version
      });
    });
    return _this;
  }

  _createClass(PublishTransaction, [{
    key: "clone",
    value: function clone(scripts) {
      return new this.constructor({
        version: this.version,
        attributes: this.attributes,
        inputs: this.inputs,
        outputs: this.outputs,
        scripts: scripts,
        contract: this.contract
      });
    }
  }, {
    key: "serializeExclusiveBase",
    value: function serializeExclusiveBase(writer) {
      serializeContractWireBase({
        contract: this.contract,
        writer: writer,
        publishVersion: this.version
      });
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function () {
      var _ref2 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee(context) {
        var transactionBaseJSON;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return _get(PublishTransaction.prototype.__proto__ || Object.getPrototypeOf(PublishTransaction.prototype), "serializeTransactionBaseJSON", this).call(this, context);

              case 2:
                transactionBaseJSON = _context.sent;
                return _context.abrupt("return", {
                  type: 'PublishTransaction',
                  txid: transactionBaseJSON.txid,
                  size: transactionBaseJSON.size,
                  version: transactionBaseJSON.version,
                  attributes: transactionBaseJSON.attributes,
                  vin: transactionBaseJSON.vin,
                  vout: transactionBaseJSON.vout,
                  scripts: transactionBaseJSON.scripts,
                  sys_fee: transactionBaseJSON.sys_fee,
                  net_fee: transactionBaseJSON.net_fee,
                  contract: this.contract.serializeJSON(context)
                });

              case 4:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function serializeJSON(_x) {
        return _ref2.apply(this, arguments);
      }

      return serializeJSON;
    }() // eslint-disable-next-line

  }, {
    key: "verify",
    value: function () {
      var _ref3 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee2(options) {
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                throw new VerifyError('Enrollment transactions are obsolete');

              case 1:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function verify(_x2) {
        return _ref3.apply(this, arguments);
      }

      return verify;
    }()
  }, {
    key: "size",
    get: function get() {
      return this.__size();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var _rollupPluginBabelHel = _get(PublishTransaction.__proto__ || Object.getPrototypeOf(PublishTransaction), "deserializeTransactionBaseStartWireBase", this).call(this, options),
          type = _rollupPluginBabelHel.type,
          version = _rollupPluginBabelHel.version;

      if (type !== TRANSACTION_TYPE.PUBLISH) {
        throw new InvalidFormatError();
      }

      var contract = deserializeContractWireBase(_extends({}, options, {
        publishVersion: version
      }));

      var _rollupPluginBabelHel2 = _get(PublishTransaction.__proto__ || Object.getPrototypeOf(PublishTransaction), "deserializeTransactionBaseEndWireBase", this).call(this, options),
          attributes = _rollupPluginBabelHel2.attributes,
          inputs = _rollupPluginBabelHel2.inputs,
          outputs = _rollupPluginBabelHel2.outputs,
          scripts = _rollupPluginBabelHel2.scripts;

      return new this({
        version: version,
        attributes: attributes,
        inputs: inputs,
        outputs: outputs,
        scripts: scripts,
        contract: contract
      });
    }
  }]);
  return PublishTransaction;
}(TransactionBase);

var InvocationTransaction =
/*#__PURE__*/
function (_TransactionBase) {
  _inherits(InvocationTransaction, _TransactionBase);

  function InvocationTransaction(_ref) {
    var _this;

    var version = _ref.version,
        attributes = _ref.attributes,
        inputs = _ref.inputs,
        outputs = _ref.outputs,
        scripts = _ref.scripts,
        hash = _ref.hash,
        gas = _ref.gas,
        script = _ref.script;
    _classCallCheck(this, InvocationTransaction);
    _this = _possibleConstructorReturn(this, (InvocationTransaction.__proto__ || Object.getPrototypeOf(InvocationTransaction)).call(this, {
      version: version,
      type: TRANSACTION_TYPE.INVOCATION,
      attributes: attributes,
      inputs: inputs,
      outputs: outputs,
      scripts: scripts,
      hash: hash
    }));
    _this.gas = gas;
    _this.script = script;

    if (_this.version > 1) {
      throw new InvalidFormatError();
    }

    if (_this.script.length === 0) {
      throw new InvalidFormatError();
    }

    if (_this.gas.lt(utils.ZERO)) {
      throw new InvalidFormatError();
    }

    _this.__size = utils.lazy(function () {
      return _get(InvocationTransaction.prototype.__proto__ || Object.getPrototypeOf(InvocationTransaction.prototype), "size", _this) + IOHelper.sizeOfUInt8 + // TODO: Doesn't seem to add size of GAS
      IOHelper.sizeOfVarBytesLE(_this.script);
    });
    return _this;
  }

  _createClass(InvocationTransaction, [{
    key: "clone",
    value: function clone(scripts) {
      return new this.constructor({
        version: this.version,
        attributes: this.attributes,
        inputs: this.inputs,
        outputs: this.outputs,
        scripts: scripts,
        gas: this.gas,
        script: this.script
      });
    }
  }, {
    key: "serializeExclusiveBase",
    value: function serializeExclusiveBase(writer) {
      writer.writeVarBytesLE(this.script);

      if (this.version >= 1) {
        writer.writeFixed8(this.gas);
      }
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function () {
      var _ref2 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee(context) {
        var transactionBaseJSON, _ref3, asset, contracts, actions, result, json;

        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return _get(InvocationTransaction.prototype.__proto__ || Object.getPrototypeOf(InvocationTransaction.prototype), "serializeTransactionBaseJSON", this).call(this, context);

              case 2:
                transactionBaseJSON = _context.sent;
                _context.next = 5;
                return context.getInvocationData(this);

              case 5:
                _ref3 = _context.sent;
                asset = _ref3.asset;
                contracts = _ref3.contracts;
                actions = _ref3.actions;
                result = _ref3.result;
                json = {
                  type: 'InvocationTransaction',
                  txid: transactionBaseJSON.txid,
                  size: transactionBaseJSON.size,
                  version: transactionBaseJSON.version,
                  attributes: transactionBaseJSON.attributes,
                  vin: transactionBaseJSON.vin,
                  vout: transactionBaseJSON.vout,
                  scripts: transactionBaseJSON.scripts,
                  sys_fee: transactionBaseJSON.sys_fee,
                  net_fee: transactionBaseJSON.net_fee,
                  script: JSONHelper.writeBuffer(this.script),
                  gas: JSONHelper.writeFixed8(this.gas),
                  result: result.serializeJSON(context),
                  asset: asset == null ? undefined : asset.serializeJSON(context),
                  contracts: contracts.map(function (contract) {
                    return contract.serializeJSON(context);
                  }),
                  actions: actions.map(function (action) {
                    return action.serializeJSON(context);
                  })
                };

                if (asset == null) {
                  delete json.asset;
                }

                return _context.abrupt("return", json);

              case 13:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function serializeJSON(_x) {
        return _ref2.apply(this, arguments);
      }

      return serializeJSON;
    }() // eslint-disable-next-line

  }, {
    key: "getSystemFee",
    value: function getSystemFee(context) {
      return this.gas;
    } // eslint-disable-next-line

  }, {
    key: "verify",
    value: function () {
      var _ref4 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee2(options) {
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                if (this.gas.mod(utils.ONE_HUNDRED_MILLION).eq(utils.ZERO)) {
                  _context2.next = 2;
                  break;
                }

                throw new VerifyError('Invalid GAS amount');

              case 2:
                _context2.next = 4;
                return _get(InvocationTransaction.prototype.__proto__ || Object.getPrototypeOf(InvocationTransaction.prototype), "verify", this).call(this, options);

              case 4:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function verify(_x2) {
        return _ref4.apply(this, arguments);
      }

      return verify;
    }()
  }, {
    key: "size",
    get: function get() {
      return this.__size();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;

      var _rollupPluginBabelHel = _get(InvocationTransaction.__proto__ || Object.getPrototypeOf(InvocationTransaction), "deserializeTransactionBaseStartWireBase", this).call(this, options),
          type = _rollupPluginBabelHel.type,
          version = _rollupPluginBabelHel.version;

      if (type !== TRANSACTION_TYPE.INVOCATION) {
        throw new InvalidFormatError();
      }

      var script = reader.readVarBytesLE(65536);

      if (script.length === 0) {
        throw new InvalidFormatError();
      }

      var gas = utils.ZERO;

      if (version >= 1) {
        gas = reader.readFixed8();
      }

      var _rollupPluginBabelHel2 = _get(InvocationTransaction.__proto__ || Object.getPrototypeOf(InvocationTransaction), "deserializeTransactionBaseEndWireBase", this).call(this, options),
          attributes = _rollupPluginBabelHel2.attributes,
          inputs = _rollupPluginBabelHel2.inputs,
          outputs = _rollupPluginBabelHel2.outputs,
          scripts = _rollupPluginBabelHel2.scripts;

      return new this({
        version: version,
        attributes: attributes,
        inputs: inputs,
        outputs: outputs,
        scripts: scripts,
        script: script,
        gas: gas
      });
    }
  }]);
  return InvocationTransaction;
}(TransactionBase);

Object.defineProperty(InvocationTransaction, "VERSION", {
  configurable: true,
  enumerable: true,
  writable: true,
  value: 1
});

var deserializeWireBase$2 = function deserializeWireBase(options) {
  var reader = options.reader;
  var type = assertTransactionType(reader.clone().readUInt8());

  switch (type) {
    case 0x00:
      return MinerTransaction.deserializeWireBase(options);

    case 0x01:
      return IssueTransaction.deserializeWireBase(options);

    case 0x02:
      return ClaimTransaction.deserializeWireBase(options);

    case 0x20:
      return EnrollmentTransaction.deserializeWireBase(options);

    case 0x40:
      return RegisterTransaction.deserializeWireBase(options);

    case 0x80:
      return ContractTransaction.deserializeWireBase(options);

    case 0xd0:
      return PublishTransaction.deserializeWireBase(options);

    case 0xd1:
      return InvocationTransaction.deserializeWireBase(options);

    default:
      // eslint-disable-next-line
      
      throw new InvalidTransactionTypeError(type);
  }
};
var deserializeWire$2 = createDeserializeWire(deserializeWireBase$2);

/* @fl/*  */
var Block =
/*#__PURE__*/
function (_BlockBase) {
  _inherits(Block, _BlockBase);

  function Block(_ref) {
    var _this;

    var version = _ref.version,
        previousHash = _ref.previousHash,
        merkleRoot = _ref.merkleRoot,
        timestamp = _ref.timestamp,
        index = _ref.index,
        consensusData = _ref.consensusData,
        nextConsensus = _ref.nextConsensus,
        script = _ref.script,
        hash = _ref.hash,
        transactions = _ref.transactions;
    _classCallCheck(this, Block);
    _this = _possibleConstructorReturn(this, (Block.__proto__ || Object.getPrototypeOf(Block)).call(this, {
      version: version,
      previousHash: previousHash,
      merkleRoot: merkleRoot || MerkleTree.computeRoot(transactions.map(function (transaction) {
        return transaction.hash;
      })),
      timestamp: timestamp,
      index: index,
      consensusData: consensusData,
      nextConsensus: nextConsensus,
      script: script,
      hash: hash
    }));
    Object.defineProperty(_this, "_header", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.lazy(function () {
        return new Header({
          version: _this.version,
          previousHash: _this.previousHash,
          merkleRoot: _this.merkleRoot,
          timestamp: _this.timestamp,
          index: _this.index,
          consensusData: _this.consensusData,
          nextConsensus: _this.nextConsensus,
          script: _this.script
        });
      })
    });
    _this.transactions = transactions;
    _this.__blockSize = utils.lazy(function () {
      return _get(Block.prototype.__proto__ || Object.getPrototypeOf(Block.prototype), "size", _this) + IOHelper.sizeOfArray(_this.transactions, function (transaction) {
        return transaction.size;
      });
    });
    return _this;
  }

  _createClass(Block, [{
    key: "getNetworkFee",
    value: function () {
      var _ref2 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee(context) {
        var fees;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return Promise.all(this.transactions.map(function (transaction) {
                  return transaction.getNetworkFee(context);
                }));

              case 2:
                fees = _context.sent;
                return _context.abrupt("return", fees.reduce(function (acc, fee) {
                  return acc.add(fee);
                }, utils.ZERO));

              case 4:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function getNetworkFee(_x) {
        return _ref2.apply(this, arguments);
      }

      return getNetworkFee;
    }()
  }, {
    key: "getSystemFee",
    value: function getSystemFee(context) {
      return this.transactions.reduce(function (acc, transaction) {
        return acc.add(transaction.getSystemFee(context));
      }, utils.ZERO);
    }
  }, {
    key: "verify",
    value: function () {
      var _ref3 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee2(options) {
        var _options$completely, completely;

        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _options$completely = options.completely, completely = _options$completely === void 0 ? false : _options$completely;

                if (!(this.transactions.length === 0 || this.transactions[0].type !== TRANSACTION_TYPE.MINER || this.transactions.slice(1).some(function (transaction) {
                  return transaction.type === TRANSACTION_TYPE.MINER;
                }))) {
                  _context2.next = 3;
                  break;
                }

                throw new VerifyError('Invalid miner transaction in block.');

              case 3:
                _context2.next = 5;
                return Promise.all([this._verifyBase(options), completely ? this._verifyComplete(options) : Promise.resolve()]);

              case 5:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function verify(_x2) {
        return _ref3.apply(this, arguments);
      }

      return verify;
    }()
  }, {
    key: "_verifyBase",
    value: function () {
      var _ref5 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee3(_ref4) {
        var genesisBlock, tryGetBlock, tryGetHeader, verifyScript, existingBlock, previousHeader;
        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                genesisBlock = _ref4.genesisBlock, tryGetBlock = _ref4.tryGetBlock, tryGetHeader = _ref4.tryGetHeader, verifyScript = _ref4.verifyScript;

                if (!(this.hash === genesisBlock.hash)) {
                  _context3.next = 3;
                  break;
                }

                return _context3.abrupt("return");

              case 3:
                _context3.next = 5;
                return tryGetBlock({
                  hashOrIndex: this.hash
                });

              case 5:
                existingBlock = _context3.sent;

                if (!(existingBlock != null)) {
                  _context3.next = 8;
                  break;
                }

                return _context3.abrupt("return");

              case 8:
                _context3.next = 10;
                return tryGetHeader({
                  hashOrIndex: this.previousHash
                });

              case 10:
                previousHeader = _context3.sent;

                if (!(previousHeader == null)) {
                  _context3.next = 13;
                  break;
                }

                throw new VerifyError('Previous header does not exist.');

              case 13:
                if (!(previousHeader.index + 1 !== this.index)) {
                  _context3.next = 15;
                  break;
                }

                throw new VerifyError('Previous index + 1 does not match index.');

              case 15:
                if (!(previousHeader.timestamp >= this.timestamp)) {
                  _context3.next = 17;
                  break;
                }

                throw new VerifyError('Previous timestamp is greater than block.');

              case 17:
                _context3.next = 19;
                return verifyScript({
                  scriptContainer: {
                    type: SCRIPT_CONTAINER_TYPE.BLOCK,
                    value: this
                  },
                  hash: previousHeader.nextConsensus,
                  witness: this.script
                });

              case 19:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function _verifyBase(_x3) {
        return _ref5.apply(this, arguments);
      }

      return _verifyBase;
    }()
  }, {
    key: "_verifyComplete",
    value: function () {
      var _ref6 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee4(options) {
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                _context4.next = 2;
                return Promise.all([this._verifyConsensus(options), this._verifyTransactions(options), this._verifyNetworkFee(options)]);

              case 2:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function _verifyComplete(_x4) {
        return _ref6.apply(this, arguments);
      }

      return _verifyComplete;
    }()
  }, {
    key: "_verifyConsensus",
    value: function () {
      var _ref8 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee5(_ref7) {
        var getValidators, validators;
        return regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                getValidators = _ref7.getValidators;
                _context5.next = 3;
                return getValidators(this.transactions);

              case 3:
                validators = _context5.sent;

                if (common.uInt160Equal(this.nextConsensus, crypto$2.getConsensusAddress(validators))) {
                  _context5.next = 6;
                  break;
                }

                throw new VerifyError('Invalid next consensus address');

              case 6:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function _verifyConsensus(_x5) {
        return _ref8.apply(this, arguments);
      }

      return _verifyConsensus;
    }()
  }, {
    key: "_verifyTransactions",
    value: function () {
      var _ref9 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee6(options) {
        return regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                _context6.next = 2;
                return Promise.all(this.transactions.map(function (transaction) {
                  return transaction.verify({
                    isSpent: options.isSpent,
                    getAsset: options.getAsset,
                    getOutput: options.getOutput,
                    calculateClaimAmount: options.calculateClaimAmount,
                    verifyScript: options.verifyScript,
                    currentHeight: options.currentHeight,
                    governingToken: options.governingToken,
                    utilityToken: options.utilityToken,
                    fees: options.fees
                  });
                }));

              case 2:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function _verifyTransactions(_x6) {
        return _ref9.apply(this, arguments);
      }

      return _verifyTransactions;
    }()
  }, {
    key: "_verifyNetworkFee",
    value: function () {
      var _ref10 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee7(options) {
        var networkFee, minerTransaction, minerTransactionNetworkFee;
        return regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                _context7.next = 2;
                return this.getNetworkFee({
                  getOutput: options.getOutput,
                  governingToken: options.governingToken,
                  utilityToken: options.utilityToken,
                  fees: options.fees
                });

              case 2:
                networkFee = _context7.sent;
                minerTransaction = this.transactions.find(function (transaction) {
                  return transaction.type === TRANSACTION_TYPE.MINER;
                });

                if (!(minerTransaction == null)) {
                  _context7.next = 6;
                  break;
                }

                throw new VerifyError('Missing miner transaction');

              case 6:
                minerTransactionNetworkFee = minerTransaction.outputs.reduce(function (acc, output) {
                  return acc.add(output.value);
                }, utils.ZERO);

                if (networkFee.eq(minerTransactionNetworkFee)) {
                  _context7.next = 9;
                  break;
                }

                throw new VerifyError('Miner output does not equal network fee.');

              case 9:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function _verifyNetworkFee(_x7) {
        return _ref10.apply(this, arguments);
      }

      return _verifyNetworkFee;
    }()
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(Block.prototype.__proto__ || Object.getPrototypeOf(Block.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeArray(this.transactions, function (transaction) {
        return transaction.serializeWireBase(writer);
      });
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function () {
      var _ref11 = _asyncToGenerator(
      /*#__PURE__*/
      regeneratorRuntime.mark(function _callee8(context) {
        var blockBaseJSON;
        return regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                blockBaseJSON = _get(Block.prototype.__proto__ || Object.getPrototypeOf(Block.prototype), "serializeBlockBaseJSON", this).call(this, context);
                _context8.t0 = blockBaseJSON.version;
                _context8.t1 = blockBaseJSON.hash;
                _context8.t2 = blockBaseJSON.previousblockhash;
                _context8.t3 = blockBaseJSON.merkleroot;
                _context8.t4 = blockBaseJSON.time;
                _context8.t5 = blockBaseJSON.index;
                _context8.t6 = blockBaseJSON.nonce;
                _context8.t7 = blockBaseJSON.nextconsensus;
                _context8.t8 = blockBaseJSON.script;
                _context8.next = 12;
                return Promise.all(this.transactions.map(function (transaction) {
                  return transaction.serializeJSON(context);
                }));

              case 12:
                _context8.t9 = _context8.sent;
                _context8.t10 = blockBaseJSON.size;
                _context8.t11 = blockBaseJSON.confirmations;
                return _context8.abrupt("return", {
                  version: _context8.t0,
                  hash: _context8.t1,
                  previousblockhash: _context8.t2,
                  merkleroot: _context8.t3,
                  time: _context8.t4,
                  index: _context8.t5,
                  nonce: _context8.t6,
                  nextconsensus: _context8.t7,
                  script: _context8.t8,
                  tx: _context8.t9,
                  size: _context8.t10,
                  confirmations: _context8.t11
                });

              case 16:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee8, this);
      }));

      function serializeJSON(_x8) {
        return _ref11.apply(this, arguments);
      }

      return serializeJSON;
    }()
  }, {
    key: "size",
    get: function get() {
      return this.__blockSize();
    }
  }, {
    key: "header",
    get: function get() {
      return this._header();
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase$$1(options) {
      var reader = options.reader;
      var blockBase = _get(Block.__proto__ || Object.getPrototypeOf(Block), "deserializeBlockBaseWireBase", this).call(this, options);
      var transactions = reader.readArray(function () {
        return deserializeWireBase$2(options);
      }, 0x10000);

      if (transactions.length === 0) {
        throw new InvalidFormatError();
      }

      var merkleRoot = MerkleTree.computeRoot(transactions.map(function (transaction) {
        return transaction.hash;
      }));

      if (!common.uInt256Equal(merkleRoot, blockBase.merkleRoot)) {
        throw new InvalidFormatError();
      }

      return new this({
        version: blockBase.version,
        previousHash: blockBase.previousHash,
        merkleRoot: blockBase.merkleRoot,
        timestamp: blockBase.timestamp,
        index: blockBase.index,
        consensusData: blockBase.consensusData,
        nextConsensus: blockBase.nextConsensus,
        script: blockBase.script,
        transactions: transactions
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return Block;
}(BlockBase);

var _fixBabelExtend$11 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var INVOCATION_RESULT_TYPE = {
  ERROR: 0x00,
  SUCCESS: 0x01
};
var InvalidInvocationResultTypeError = _fixBabelExtend$11(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidInvocationResultTypeError, _Error);

  function InvalidInvocationResultTypeError(value) {
    var _this;

    _classCallCheck(this, InvalidInvocationResultTypeError);
    _this = _possibleConstructorReturn(this, (InvalidInvocationResultTypeError.__proto__ || Object.getPrototypeOf(InvalidInvocationResultTypeError)).call(this, "Expected invocation result type, found: ".concat(value.toString(16))));
    _this.value = value;
    return _this;
  }

  return InvalidInvocationResultTypeError;
}(Error));
var assertInvocationResultType = function assertInvocationResultType(value) {
  switch (value) {
    case INVOCATION_RESULT_TYPE.ERROR:
      return INVOCATION_RESULT_TYPE.ERROR;

    case INVOCATION_RESULT_TYPE.SUCCESS:
      return INVOCATION_RESULT_TYPE.SUCCESS;

    default:
      throw new InvalidInvocationResultTypeError(value);
  }
};

var InvocationResultBase =
/*#__PURE__*/
function () {
  function InvocationResultBase() {
    _classCallCheck(this, InvocationResultBase);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(this.serializeWireBase.bind(this))
    });
  }

  _createClass(InvocationResultBase, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt8(this.type);
    }
  }], [{
    key: "deserializeInvocationResultWireBase",
    // eslint-disable-next-line
    value: function deserializeInvocationResultWireBase(_ref) {
      var reader = _ref.reader;
      var type = reader.readUInt8();
      return {
        type: type
      };
    } // eslint-disable-next-line

  }, {
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      throw new Error('Not Implemented');
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return InvocationResultBase;
}();

var InvocationResultSuccess =
/*#__PURE__*/
function (_InvocationResultBase) {
  _inherits(InvocationResultSuccess, _InvocationResultBase);

  function InvocationResultSuccess(_ref) {
    var _this;

    var stack = _ref.stack,
        stackAlt = _ref.stackAlt;
    _classCallCheck(this, InvocationResultSuccess);
    _this = _possibleConstructorReturn(this, (InvocationResultSuccess.__proto__ || Object.getPrototypeOf(InvocationResultSuccess)).call(this));
    Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: INVOCATION_RESULT_TYPE.SUCCESS
    });
    _this.stack = stack;
    _this.stackAlt = stackAlt;
    return _this;
  }

  _createClass(InvocationResultSuccess, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(InvocationResultSuccess.prototype.__proto__ || Object.getPrototypeOf(InvocationResultSuccess.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeArray(this.stack, function (contractParameter) {
        return contractParameter.serializeWireBase(writer);
      });
      writer.writeArray(this.stackAlt, function (contractParameter) {
        return contractParameter.serializeWireBase(writer);
      });
    } // eslint-disable-next-line

  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        type: 'Success',
        stack: this.stack.map(function (value) {
          return value.serializeJSON(context);
        }),
        stackAlt: this.stackAlt.map(function (value) {
          return value.serializeJSON(context);
        })
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase$$1(options) {
      var reader = options.reader;

      var _rollupPluginBabelHel = _get(InvocationResultSuccess.__proto__ || Object.getPrototypeOf(InvocationResultSuccess), "deserializeInvocationResultWireBase", this).call(this, options),
          type = _rollupPluginBabelHel.type;

      if (type !== INVOCATION_RESULT_TYPE.SUCCESS) {
        throw new InvalidFormatError();
      }

      var stack = reader.readArray(function () {
        return deserializeWireBase$1(options);
      });
      var stackAlt = reader.readArray(function () {
        return deserializeWireBase$1(options);
      });
      return new this({
        stack: stack,
        stackAlt: stackAlt
      });
    }
  }]);
  return InvocationResultSuccess;
}(InvocationResultBase);

var MAX_SIZE = 1024;

var InvocationResultError =
/*#__PURE__*/
function (_InvocationResultBase) {
  _inherits(InvocationResultError, _InvocationResultBase);

  function InvocationResultError(_ref) {
    var _this;

    var message = _ref.message;
    _classCallCheck(this, InvocationResultError);
    _this = _possibleConstructorReturn(this, (InvocationResultError.__proto__ || Object.getPrototypeOf(InvocationResultError)).call(this));
    Object.defineProperty(_this, "type", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: INVOCATION_RESULT_TYPE.ERROR
    });
    _this.message = message;
    return _this;
  }

  _createClass(InvocationResultError, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(InvocationResultError.prototype.__proto__ || Object.getPrototypeOf(InvocationResultError.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeVarString(this.message, MAX_SIZE);
    } // eslint-disable-next-line

  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        type: 'Error',
        message: this.message
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;

      var _rollupPluginBabelHel = _get(InvocationResultError.__proto__ || Object.getPrototypeOf(InvocationResultError), "deserializeInvocationResultWireBase", this).call(this, options),
          type = _rollupPluginBabelHel.type;

      if (type !== INVOCATION_RESULT_TYPE.ERROR) {
        throw new InvalidFormatError();
      }

      var message = reader.readVarString(MAX_SIZE);
      return new this({
        message: message
      });
    }
  }]);
  return InvocationResultError;
}(InvocationResultBase);

var deserializeWireBase$3 = function deserializeWireBase(options) {
  var reader = options.reader;
  var type = assertInvocationResultType(reader.clone().readUInt8());

  switch (type) {
    case 0x00:
      return InvocationResultError.deserializeWireBase(options);

    case 0x01:
      return InvocationResultSuccess.deserializeWireBase(options);

    default:
      // eslint-disable-next-line
      
      throw new InvalidInvocationResultTypeError(type);
  }
};
var deserializeWire$3 = createDeserializeWire(deserializeWireBase$3);

var InvocationData =
/*#__PURE__*/
function () {
  function InvocationData(_ref) {
    var hash = _ref.hash,
        assetHash = _ref.assetHash,
        contractHashes = _ref.contractHashes,
        blockIndex = _ref.blockIndex,
        transactionIndex = _ref.transactionIndex,
        result = _ref.result;
    _classCallCheck(this, InvocationData);
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.hash = hash;
    this.assetHash = assetHash;
    this.contractHashes = contractHashes;
    this.blockIndex = blockIndex;
    this.transactionIndex = transactionIndex;
    this.result = result;
  }

  _createClass(InvocationData, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt256(this.hash);
      writer.writeUInt256(this.assetHash || common.ZERO_UINT256);
      writer.writeArray(this.contractHashes, function (contractHash) {
        writer.writeUInt160(contractHash);
      });
      writer.writeUInt32LE(this.blockIndex);
      writer.writeUInt32LE(this.transactionIndex);
      this.result.serializeWireBase(writer);
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase$$1(options) {
      var reader = options.reader;
      var hash = reader.readUInt256();
      var assetHash = reader.readUInt256();
      var contractHashes = reader.readArray(function () {
        return reader.readUInt160();
      });
      var blockIndex = reader.readUInt32LE();
      var transactionIndex = reader.readUInt32LE();

      var result = deserializeWireBase$3(options);

      return new this({
        hash: hash,
        assetHash: common.uInt256Equal(assetHash, common.ZERO_UINT256) ? undefined : assetHash,
        contractHashes: contractHashes,
        blockIndex: blockIndex,
        transactionIndex: transactionIndex,
        result: result
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return InvocationData;
}();

var StorageItem =
/*#__PURE__*/
function () {
  function StorageItem(_ref) {
    var _this = this;

    var hash = _ref.hash,
        key = _ref.key,
        value = _ref.value;
    _classCallCheck(this, StorageItem);
    Object.defineProperty(this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(StorageItem, function (other) {
        return common.uInt160Equal(_this.hash, other.hash) && _this.key.equals(other.key) && _this.value.equals(other.value);
      })
    });
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(this.serializeWireBase.bind(this))
    });
    this.hash = hash;
    this.key = key;
    this.value = value;
  }

  _createClass(StorageItem, [{
    key: "update",
    value: function update(_ref2) {
      var value = _ref2.value;
      return new this.constructor({
        hash: this.hash,
        key: this.key,
        value: value
      });
    }
  }, {
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt160(this.hash);
      writer.writeVarBytesLE(this.key);
      writer.writeVarBytesLE(this.value);
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        hash: JSONHelper.writeUInt160(this.hash),
        key: JSONHelper.writeBuffer(this.key),
        value: JSONHelper.writeBuffer(this.value)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref3) {
      var reader = _ref3.reader;
      var hash = reader.readUInt160();
      var key = reader.readVarBytesLE();
      var value = reader.readVarBytesLE();
      return new this({
        hash: hash,
        key: key,
        value: value
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return StorageItem;
}();

var Validator =
/*#__PURE__*/
function (_BaseState) {
  _inherits(Validator, _BaseState);

  function Validator(_ref) {
    var _this;

    var version = _ref.version,
        publicKey = _ref.publicKey;
    _classCallCheck(this, Validator);
    _this = _possibleConstructorReturn(this, (Validator.__proto__ || Object.getPrototypeOf(Validator)).call(this, {
      version: version
    }));
    Object.defineProperty(_this, "equals", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: utils.equals(Validator, function (other) {
        return common.ecPointEqual(_this.publicKey, other.publicKey);
      })
    });
    Object.defineProperty(_this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: createSerializeWire(_this.serializeWireBase.bind(_this))
    });
    _this.publicKey = publicKey;
    return _this;
  }

  _createClass(Validator, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt8(this.version);
      writer.writeECPoint(this.publicKey);
    }
  }, {
    key: "serializeJSON",
    // eslint-disable-next-line
    value: function serializeJSON(context) {
      return {
        version: this.version,
        publicKey: JSONHelper.writeECPoint(this.publicKey)
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(_ref2) {
      var reader = _ref2.reader;
      var version = reader.readUInt8();
      var publicKey = reader.readECPoint();
      return new this({
        version: version,
        publicKey: publicKey
      });
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return Validator;
}(BaseState);

var _fixBabelExtend$12 = function (O) {
  var gOPD = O.getOwnPropertyDescriptor,
      gPO = O.getPrototypeOf || function (o) {
    return o.__proto__;
  },
      sPO = O.setPrototypeOf || function (o, p) {
    o.__proto__ = p;
    return o;
  },
      construct = _typeof(Reflect) === 'object' ? Reflect.construct : function (Parent, args, Class) {
    var Constructor,
        a = [null];
    a.push.apply(a, args);
    Constructor = Parent.bind.apply(Parent, a);
    return sPO(new Constructor(), Class.prototype);
  };

  return function fixBabelExtend(Class) {
    var Parent = gPO(Class);
    return sPO(Class, sPO(function Super() {
      return construct(Parent, arguments, gPO(this).constructor);
    }, Parent));
  };
}(Object);

var ACTION_TYPE = {
  LOG: 0x00,
  NOTIFICATION: 0x01
};
var InvalidActionTypeError = _fixBabelExtend$12(
/*#__PURE__*/
function (_Error) {
  _inherits(InvalidActionTypeError, _Error);

  function InvalidActionTypeError(type) {
    var _this;

    _classCallCheck(this, InvalidActionTypeError);
    _this = _possibleConstructorReturn(this, (InvalidActionTypeError.__proto__ || Object.getPrototypeOf(InvalidActionTypeError)).call(this, "Expected action type, found: ".concat(type)));
    _this.type = type;
    return _this;
  }

  return InvalidActionTypeError;
}(Error));
var assertActionType = function assertActionType(value) {
  switch (value) {
    case ACTION_TYPE.LOG:
      return ACTION_TYPE.LOG;

    case ACTION_TYPE.NOTIFICATION:
      return ACTION_TYPE.NOTIFICATION;

    default:
      throw new InvalidActionTypeError(value);
  }
};

/* @/*  */
var ActionBase =
/*#__PURE__*/
function () {
  function ActionBase(_ref) {
    var type = _ref.type,
        version = _ref.version,
        blockIndex = _ref.blockIndex,
        blockHash = _ref.blockHash,
        transactionIndex = _ref.transactionIndex,
        transactionHash = _ref.transactionHash,
        index = _ref.index,
        scriptHash = _ref.scriptHash;
    _classCallCheck(this, ActionBase);

    _initialiseProps$1.call(this);

    this.type = type;
    this.version = version == null ? this.constructor.VERSION : version;
    this.blockIndex = blockIndex;
    this.blockHash = blockHash;
    this.transactionIndex = transactionIndex;
    this.transactionHash = transactionHash;
    this.index = index;
    this.scriptHash = scriptHash;
  }

  _createClass(ActionBase, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      writer.writeUInt8(this.type);
      writer.writeUInt8(this.version);
      writer.writeUInt32LE(this.blockIndex);
      writer.writeUInt256(this.blockHash);
      writer.writeUInt32LE(this.transactionIndex);
      writer.writeUInt256(this.transactionHash);
      writer.writeUInt32LE(this.index);
      writer.writeUInt160(this.scriptHash);
    }
  }, {
    key: "serializeActionBaseJSON",
    // eslint-disable-next-line
    value: function serializeActionBaseJSON(context) {
      return {
        version: this.version,
        blockIndex: this.blockIndex,
        blockHash: JSONHelper.writeUInt256(this.blockHash),
        transactionIndex: this.transactionIndex,
        transactionHash: JSONHelper.writeUInt256(this.transactionHash),
        index: this.index,
        scriptHash: JSONHelper.writeUInt160(this.scriptHash)
      };
    }
  }], [{
    key: "deserializeWireBase",
    // eslint-disable-next-line
    value: function deserializeWireBase(options) {
      throw new Error('Not Implemented');
    }
  }, {
    key: "deserializeWire",
    value: function deserializeWire(options) {
      return this.deserializeWireBase({
        context: options.context,
        reader: new BinaryReader(options.buffer)
      });
    }
  }]);
  return ActionBase;
}();

Object.defineProperty(ActionBase, "VERSION", {
  configurable: true,
  enumerable: true,
  writable: true,
  value: 0
});
Object.defineProperty(ActionBase, "deserializeActionBaseWireBase", {
  configurable: true,
  enumerable: true,
  writable: true,
  value: function value(_ref2) {
    var reader = _ref2.reader;
    var type = reader.readUInt8();
    var version = reader.readUInt8();
    var blockIndex = reader.readUInt32LE();
    var blockHash = reader.readUInt256();
    var transactionIndex = reader.readUInt32LE();
    var transactionHash = reader.readUInt256();
    var index = reader.readUInt32LE();
    var scriptHash = reader.readUInt160();
    return {
      type: type,
      version: version,
      blockIndex: blockIndex,
      blockHash: blockHash,
      transactionIndex: transactionIndex,
      transactionHash: transactionHash,
      index: index,
      scriptHash: scriptHash
    };
  }
});

var _initialiseProps$1 = function _initialiseProps() {
  Object.defineProperty(this, "serializeWire", {
    configurable: true,
    enumerable: true,
    writable: true,
    value: createSerializeWire(this.serializeWireBase.bind(this))
  });
};

var LogAction =
/*#__PURE__*/
function (_ActionBase) {
  _inherits(LogAction, _ActionBase);

  function LogAction(_ref) {
    var _this;

    var version = _ref.version,
        blockIndex = _ref.blockIndex,
        blockHash = _ref.blockHash,
        transactionIndex = _ref.transactionIndex,
        transactionHash = _ref.transactionHash,
        index = _ref.index,
        scriptHash = _ref.scriptHash,
        message = _ref.message;
    _classCallCheck(this, LogAction);
    _this = _possibleConstructorReturn(this, (LogAction.__proto__ || Object.getPrototypeOf(LogAction)).call(this, {
      type: ACTION_TYPE.LOG,
      version: version,
      blockIndex: blockIndex,
      blockHash: blockHash,
      transactionIndex: transactionIndex,
      transactionHash: transactionHash,
      index: index,
      scriptHash: scriptHash
    }));
    _this.message = message;
    return _this;
  }

  _createClass(LogAction, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(LogAction.prototype.__proto__ || Object.getPrototypeOf(LogAction.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeVarString(this.message);
    }
  }, {
    key: "serializeJSON",
    value: function serializeJSON(context) {
      var action = _get(LogAction.prototype.__proto__ || Object.getPrototypeOf(LogAction.prototype), "serializeActionBaseJSON", this).call(this, context);
      return {
        type: 'Log',
        version: action.version,
        blockIndex: action.blockIndex,
        blockHash: action.blockHash,
        transactionIndex: action.transactionIndex,
        transactionHash: action.transactionHash,
        index: action.index,
        scriptHash: action.scriptHash,
        message: this.message
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase(options) {
      var reader = options.reader;
      var action = _get(LogAction.__proto__ || Object.getPrototypeOf(LogAction), "deserializeActionBaseWireBase", this).call(this, options);
      var message = reader.readVarString(1024);
      return new this({
        version: action.version,
        blockIndex: action.blockIndex,
        blockHash: action.blockHash,
        transactionIndex: action.transactionIndex,
        transactionHash: action.transactionHash,
        index: action.index,
        scriptHash: action.scriptHash,
        message: message
      });
    }
  }]);
  return LogAction;
}(ActionBase);

var NotificationAction =
/*#__PURE__*/
function (_ActionBase) {
  _inherits(NotificationAction, _ActionBase);

  function NotificationAction(_ref) {
    var _this;

    var version = _ref.version,
        blockIndex = _ref.blockIndex,
        blockHash = _ref.blockHash,
        transactionIndex = _ref.transactionIndex,
        transactionHash = _ref.transactionHash,
        index = _ref.index,
        scriptHash = _ref.scriptHash,
        args = _ref.args;
    _classCallCheck(this, NotificationAction);
    _this = _possibleConstructorReturn(this, (NotificationAction.__proto__ || Object.getPrototypeOf(NotificationAction)).call(this, {
      type: ACTION_TYPE.NOTIFICATION,
      version: version,
      blockIndex: blockIndex,
      blockHash: blockHash,
      transactionIndex: transactionIndex,
      transactionHash: transactionHash,
      index: index,
      scriptHash: scriptHash
    }));
    _this.args = args;
    return _this;
  }

  _createClass(NotificationAction, [{
    key: "serializeWireBase",
    value: function serializeWireBase(writer) {
      _get(NotificationAction.prototype.__proto__ || Object.getPrototypeOf(NotificationAction.prototype), "serializeWireBase", this).call(this, writer);
      writer.writeArray(this.args, function (arg) {
        return arg.serializeWireBase(writer);
      });
    }
  }, {
    key: "serializeJSON",
    value: function serializeJSON(context) {
      var action = _get(NotificationAction.prototype.__proto__ || Object.getPrototypeOf(NotificationAction.prototype), "serializeActionBaseJSON", this).call(this, context);
      return {
        type: 'Notification',
        version: action.version,
        blockIndex: action.blockIndex,
        blockHash: action.blockHash,
        transactionIndex: action.transactionIndex,
        transactionHash: action.transactionHash,
        index: action.index,
        scriptHash: action.scriptHash,
        args: this.args.map(function (arg) {
          return arg.serializeJSON(context);
        })
      };
    }
  }], [{
    key: "deserializeWireBase",
    value: function deserializeWireBase$$1(options) {
      var reader = options.reader;
      var action = _get(NotificationAction.__proto__ || Object.getPrototypeOf(NotificationAction), "deserializeActionBaseWireBase", this).call(this, options);
      var args = reader.readArray(function () {
        return deserializeWireBase$1(options);
      });
      return new this({
        version: action.version,
        blockIndex: action.blockIndex,
        blockHash: action.blockHash,
        transactionIndex: action.transactionIndex,
        transactionHash: action.transactionHash,
        index: action.index,
        scriptHash: action.scriptHash,
        args: args
      });
    }
  }]);
  return NotificationAction;
}(ActionBase);

var deserializeWireBase$4 = function deserializeWireBase(options) {
  var reader = options.reader;
  var type = assertActionType(reader.clone().readUInt8());

  switch (type) {
    case 0x00:
      return LogAction.deserializeWireBase(options);

    case 0x01:
      return NotificationAction.deserializeWireBase(options);

    default:
      // eslint-disable-next-line
      
      throw new InvalidActionTypeError(type);
  }
};
var deserializeWire$4 = createDeserializeWire(deserializeWireBase$4);

exports.Account = Account;
exports.Asset = Asset;
exports.BaseState = BaseState;
exports.Block = Block;
exports.BlockBase = BlockBase;
exports.Contract = Contract;
exports.Header = Header;
exports.InvocationData = InvocationData;
exports.StorageItem = StorageItem;
exports.Validator = Validator;
exports.Witness = Witness;
exports.LogAction = LogAction;
exports.NotificationAction = NotificationAction;
exports.common = common;
exports.SignatureContractParameter = SignatureContractParameter;
exports.BooleanContractParameter = BooleanContractParameter;
exports.IntegerContractParameter = IntegerContractParameter;
exports.Hash160ContractParameter = Hash160ContractParameter;
exports.Hash256ContractParameter = Hash256ContractParameter;
exports.ByteArrayContractParameter = ByteArrayContractParameter;
exports.PublicKeyContractParameter = PublicKeyContractParameter;
exports.StringContractParameter = StringContractParameter;
exports.ArrayContractParameter = ArrayContractParameter;
exports.InteropInterfaceContractParameter = InteropInterfaceContractParameter;
exports.VoidContractParameter = VoidContractParameter;
exports.crypto = crypto$2;
exports.MerkleTree = MerkleTree;
exports.InvalidFormatError = InvalidFormatError;
exports.InvocationResultSuccess = InvocationResultSuccess;
exports.InvocationResultError = InvocationResultError;
exports.deserializeInvocationResultWire = deserializeWire$3;
exports.ATTRIBUTE_USAGE = ATTRIBUTE_USAGE;
exports.TRANSACTION_TYPE = TRANSACTION_TYPE;
exports.BufferAttribute = BufferAttribute;
exports.ECPointAttribute = ECPointAttribute;
exports.UInt160Attribute = UInt160Attribute;
exports.UInt256Attribute = UInt256Attribute;
exports.MinerTransaction = MinerTransaction;
exports.IssueTransaction = IssueTransaction;
exports.ClaimTransaction = ClaimTransaction;
exports.EnrollmentTransaction = EnrollmentTransaction;
exports.RegisterTransaction = RegisterTransaction;
exports.ContractTransaction = ContractTransaction;
exports.PublishTransaction = PublishTransaction;
exports.InvocationTransaction = InvocationTransaction;
exports.Input = Input;
exports.Output = Output;
exports.deserializeTransactionWire = deserializeWire$2;
exports.utils = utils;
exports.BinaryReader = BinaryReader;
exports.BinaryWriter = BinaryWriter;
exports.IOHelper = IOHelper;
exports.JSONHelper = JSONHelper;
exports.ScriptBuilder = ScriptBuilder;
exports.deserializeActionWire = deserializeWire$4;
exports.ASSET_TYPE = ASSET_TYPE;
exports.InvalidAssetTypeError = InvalidAssetTypeError;
exports.assertAssetType = assertAssetType;
exports.CONTRACT_PARAMETER_TYPE = CONTRACT_PARAMETER_TYPE;
exports.InvalidContractParameterTypeError = InvalidContractParameterTypeError;
exports.assertContractParameterType = assertContractParameterType;
exports.createSerializeWire = createSerializeWire;
exports.SCRIPT_CONTAINER_TYPE = SCRIPT_CONTAINER_TYPE;
exports.BYTECODE_TO_OPCODE = BYTECODE_TO_OPCODE;
exports.OPCODE_TO_BYTECODE = OPCODE_TO_BYTECODE;
//# sourceMappingURL=index.js.map
